pro baselines, drifts_all, map_back, smallfield, size_deg, maprot_deg=maprot_deg, $
               tab_offsets=tab_offsets, file_scalars=file_scalars, ind_sc_coarse=ind_sc_coarse, $
               ind_aver=ind_aver, ind_scans=ind_scans, ind_subscans=ind_subscans, maxnoise=maxnoise, $
               cut_start=cut_start, galactic=galactic, minimap=minimap, nocross=nocross, $
               time_1=time_1, t_proc=t_proc


;;; determines one linear baseline per bolometer per subscan by robust fits
;;; (definition of subscan adapted to the multi-block case)
;;; linear baselines replaced with zero-order baselines when the "galactic" option is set

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

size_lim_nooff = 0.75


if (not keyword_set(smallfield)) then smallfield = 0
ratio_thres = 0
med_wn = median(maxnoise)

restore, file_signal


w = where(finite(signal) lt 1, cw)
mask_bolo_init = 1. + fltarr(nt, nb)
if cw ge 1 then mask_bolo_init(w) = 0.

norm_init = total(mask_bolo_init, 2)
w = where(norm_init lt nb / 5., cw)
if cw ge 1 then norm_init(w) = 0.


restore, file_pixcoords
indxy = floor(x_bolos) + dimx * floor(y_bolos)
x_bolos = 0
y_bolos = 0

restore, file_times

restore, file_time_coarse

flag_save = intarr(nt_save)

tab_aver = lonarr(nt)
tab_aver(0:ind_aver(0)-1) = 0L
for i = 1L, nt_save - 1 do tab_aver(ind_aver(i-1):ind_aver(i)-1) = i


map_back = fltarr(dimx, dimy)

wb = where(finite(total(maxnoise, 1)) gt 0)
med_offsets = median(tab_offsets(*, wb), dimension=2, /even)
offsets = total(tab_offsets - rebin(med_offsets, nscans, nb), 1, /nan) / total(finite(tab_offsets), 1)
w = where(finite(offsets) lt 1, cw)
if cw ge 1 then offsets(w) = 0.


restore, file_maps
w = where(map_weight gt 0., cw)
map_signal = 0
map_weight = 0
size_deg = sqrt(float(cw)) * pix_as / 3600.
print, '              size ~ ' + chain(size_deg, 2) + ' degree(s)'

if ((size_deg lt 1. and nscans le 20) or nt lt 50000L) then begin
   ns = 49
   if keyword_set(debug) then print, 'ns = 49'
endif else ns = 0


;;; loop to iteratively exclude sources from the fits:

if (keyword_set(galactic) and size_deg gt 1.) then nloop = 2 else nloop = 3
for k = 0, nloop - 1 do begin

   mask_bolo = mask_bolo_init

   ;;; mask out bright sources flagged at previous iteration:
   if k gt 0 then begin
      signal_simul = map_back(indxy)
      w = where(finite(signal_simul) lt 1, cw)
      signal_simul = 0
      if cw ge 1 then mask_bolo(w) = 0.
   endif

   norm = total(mask_bolo, 2)
   w = where(norm lt nb / 5., cw)
   if cw ge 1 then norm(w) = 0.
   norm_denom = norm
   w = where(norm lt 0.5, cw)
   if cw ge 1 then norm_denom(w) = 1.
   aver = total(signal * mask_bolo, 2, /nan) / norm_denom
   if cw ge 1 then aver(w) = !values.f_nan


   ;;; inspect average signal of successive scans
   ;;; to visualize the temperature drift in raw data

   ;;; if no temperature drift left, for nominal scan-a-b observations:
   ;;; average signal should have periodicity of 2 scans

   if (keyword_set(debug) and ((not keyword_set(galactic)) or size_deg gt 1.)) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      wset, 0

      w = where(finite(aver) gt 0.)
      otherplot, time_s(w), aver(w), xtitle='time (s)', ytitle='average signal (Jy/bolometer)', $
         title='all '+chain(nscans)+' scan(s)'
      yrange = [min(aver(w))-1., max(aver(w))+1.]
      for i = 0, nscans - 2 do begin
         otherplot, /over, lstyle=2., time_s(ind_scans(i+1)-1) + fltarr(2), yrange, color=200
         otherplot, /over, lstyle=2., time_s(ind_scans(i+1)) + fltarr(2), yrange, color=200
      endfor
      res = dialog_message('   Continue ?   ', /center, /information, title='Pause')

      if nscans ge 3 then begin
         off_time = 0.
         for i = 0, nscans - 1 do begin
            w = where(flag_nza(ind_scans(i):ind_scans(i+1)-1) eq 1)
            off_time = max([off_time, time_s(ind_scans(i)+w(0)) - time_s(ind_scans(i))])
         endfor

         w = where(finite(aver(0:ind_scans(1)-1)) gt 0.)
         wref = where(flag_nza(0:ind_scans(1)-1) eq 1)
         x = time_s(w) - time_s(wref(0)) + off_time
         otherplot, x, aver(w), xtitle='time (s)', ytitle='average signal', title='even scans', $
            xrange=[-10., max(x)+10.]
         for i = 2, nscans - 1, 2 do begin
            w = where(finite(aver(ind_scans(i):ind_scans(i+1)-1)) gt 0.)
            wref = where(flag_nza(ind_scans(i):ind_scans(i+1)-1) eq 1)
            otherplot, /over, color=50*i, time_s(ind_scans(i)+w) - time_s(ind_scans(i)+wref(0)) $
               + off_time, aver(ind_scans(i)+w)
         endfor
         res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
      endif

      if nscans ge 4 then begin
         w = where(finite(aver(ind_scans(1):ind_scans(2)-1)) gt 0.)
         wref = where(flag_nza(ind_scans(1):ind_scans(2)-1) eq 1)
         x = time_s(ind_scans(1)+w) - time_s(ind_scans(1)+wref(0)) + off_time
         otherplot, x, aver(ind_scans(1)+w), xtitle='time (s)', ytitle='average signal', $
            title='odd scans', xrange=[-10., max(x)+10.]
         for i = 3, nscans - 1, 2 do begin
            w = where(finite(aver(ind_scans(i):ind_scans(i+1)-1)) gt 0.)
            wref = where(flag_nza(ind_scans(i):ind_scans(i+1)-1) eq 1)
            otherplot, /over, color=50*(i-1), time_s(ind_scans(i)+w) - time_s(ind_scans(i)+wref(0)) $
               + off_time, aver(ind_scans(i)+w)
         endfor
         res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
      endif
      time_1 = systime(/seconds)
   endif


   ;;; baselines on scan length:

   for i = 0, nscans - 1 do begin

      med_noise = median(maxnoise(i, *))

      if (not keyword_set(galactic)) then begin
         wfit = where(norm(ind_scans(i):ind_scans(i+1)-1) gt 0., cw)
         time_ref = time_s(ind_scans(i))
         if (k eq 1 and keyword_set(minimap)) then order = 3 else order = 1
         if order gt 1 then begin
            coef = poly_fit(float(-time_ref + time_s(ind_scans(i)+wfit)), aver(ind_scans(i)+wfit), $
               order, yfit=fit_scan, chisq=chisq_poly)
            coef_lin = poly_fit(float(-time_ref + time_s(ind_scans(i)+wfit)), aver(ind_scans(i)+wfit), $
               1, chisq=chisq_lin)
            if chisq_poly gt chisq_lin / 5. then order = 1
         endif
         if order eq 1 then fit_scan = absdev_fit(float(-time_ref + time_s(ind_scans(i)+wfit)), aver(ind_scans(i)+wfit), $
            coef=coef, /dyn, noise_level=med_noise, ns=ns)

         if keyword_set(debug) then begin
            time_2 = systime(/seconds)
            t_proc += (time_2 - time_1)
            wset, 0
            otherplot, time_s(ind_scans(i)+wfit), aver(ind_scans(i)+wfit), xtitle='time (s)', $
               ytitle='average signal', title='linear fit to scan '+chain(i)
            otherplot, /over, color=10, time_s(ind_scans(i)+wfit), fit_scan
            res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
            time_1 = systime(/seconds)
         endif

         fit_scan = coef(0)
         for pow = 1, order do fit_scan += (coef(pow) * float((-time_ref + time_s(ind_scans(i):ind_scans(i+1)-1))^float(pow)))
         signal(ind_scans(i):ind_scans(i+1)-1, *) -= rebin(fit_scan, ind_scans(i+1)-ind_scans(i), nb)
         fit_coarse = coef(0)
         for pow = 1, order do fit_coarse += (coef(pow) * (-float(time_ref) + time_save(ind_sc_coarse(i):ind_sc_coarse(i+1)-1))^float(pow))
         drifts_all(ind_sc_coarse(i):ind_sc_coarse(i+1)-1, *) += rebin(fit_coarse, ind_sc_coarse(i+1)-ind_sc_coarse(i), nb)

         if k eq 0 then begin
            wb = where(finite(maxnoise(i, *)) gt 0)
            off = fltarr(nb)
            off(wb) = median(signal(ind_scans(i):ind_scans(i+1)-1, wb), dimension=1)
            signal(ind_scans(i):ind_scans(i+1)-1, *) -= (off ## (1. + fltarr(ind_scans(i+1)-ind_scans(i))))
            drifts_all(ind_sc_coarse(i):ind_sc_coarse(i+1)-1, *) += $
               (off ## (1. + fltarr(ind_sc_coarse(i+1)-ind_sc_coarse(i))))
         endif

      endif else begin
         if k eq 0 then begin
            signal(ind_scans(i):ind_scans(i+1)-1, *) -= med_offsets(i)

            ;;; do not subtract any ill-defined offsets for very bright fields:
            if (i eq 0 and inst eq 'spire') then begin
               nt_scan = ind_scans(i+1) - ind_scans(i)
               nbin = ceil(nt_scan / 1.e+4)
               nt_bin = nt_scan / nbin
               amp_tot = 0.
               for ibin = 0, nbin - 1 do begin
                  signal_bin = signal(ind_scans(i)+ibin*nt_bin:ind_scans(i)+(ibin+1)*nt_bin-1, *)
                  signal_bin -= (median(signal_bin, dimension=1) ## (1. + fltarr(nt_bin)))
                  w = where(finite(signal_bin) gt 0)
                  amp = mean(signal_bin(w)) - min(median(signal_bin, dimension=2), /nan)
                  amp_tot = max([amp_tot, amp])
               endfor
               if (amp_tot gt 5. and size_deg lt size_lim_nooff) then no_offsets = 1 else no_offsets = 0
               print, '              average signal amplitude ~ ' + chain(amp_tot, 2) + ' Jy/bolometer'
               if no_offsets eq 1 then print, 'no offsets (negligible wrt sky brightness)'
            endif
         endif

         wb = where(finite(maxnoise(i, *)) gt 0)
         if (not keyword_set(no_offsets)) then begin
            off = fltarr(nb)
            off(wb) = median(signal(ind_scans(i):ind_scans(i+1)-1, wb), dimension=1)
            signal(ind_scans(i):ind_scans(i+1)-1, *) -= (off ## (1. + fltarr(ind_scans(i+1)-ind_scans(i))))
            drifts_all(ind_sc_coarse(i):ind_sc_coarse(i+1)-1, *) += $
               (off ## (1. + fltarr(ind_sc_coarse(i+1)-ind_sc_coarse(i))))
         endif
         if k eq 0 then drifts_all(ind_sc_coarse(i):ind_sc_coarse(i+1)-1, wb) += med_offsets(i)
      endelse
   endfor



   ;;; baselines on subscan length:

   if ((not keyword_set(galactic)) and k gt 0) then begin
      drifts_base = fltarr(nt_save, nb)

      for i = 0, nscans - 1 do begin
         wsubs = where(ind_subscans ge ind_scans(i) and ind_subscans le ind_scans(i+1) - 1, csubs)

         for isubs = 0L, csubs - 1 do begin
            subs = wsubs(isubs)
            it0 = ind_subscans(subs)
            w = where(flag_nza(it0:ind_subscans(subs+1)-1) eq 1, cw)
            if (cw ge 100 and csubs ge 5) then begin
               it0 += w(0)
               it1 = it0
               w = where(flag_nza(it1:ind_subscans(subs+1)-1) eq 0, cw)
               if cw eq 0 then it1 = ind_subscans(subs+1) - 1 else it1 += (w(0) - 1)
               it0_nom = it0
               it1_nom = it1
               it0 = max([ind_subscans(subs), it0 - 200])
               it1 = min([ind_subscans(subs+1) - 1, it1 + 200])
            endif else begin
               it0 = ind_subscans(subs)
               it1 = ind_subscans(subs+1) - 1
               it0_nom = it0
               it1_nom = it1
            endelse

            nt_fit = it1 - it0 + 1
            if (size_deg gt 1. and ratio_thres lt 0.01 and nt_fit gt 1000) then begin
               wfit = [it0 + lindgen(nt_fit/3), it1 - reverse(lindgen(nt_fit/3))]
               nt_fit = 2 * (nt_fit / 3)
            endif else wfit = it0 + lindgen(nt_fit)


            ;;; zero-order baselines:

            flux = signal(wfit, *)
            wz = where(mask_bolo(wfit, *) lt 0.5, cz)
            if cz ge 1 then flux(wz) = !values.f_nan
            med = median(flux, dimension=1)
            wbz = where(finite(maxnoise(i, *)) lt 1 or finite(med) lt 1, cbz, complement=wb)
            if cbz ge 1 then med(wbz) = 0.
            drifts_base(tab_aver(it0):tab_aver(it1), *) += (med ## (1. + fltarr(tab_aver(it1)-tab_aver(it0)+1)))
            flux -= (med ## (1. + fltarr(nt_fit)))


            ;;; first-order baselines (except for very structured fields):

            if k eq nloop - 1 then begin
               coef_lin = fltarr(nb)
               for b = 0L, nb - 1 do if finite(maxnoise(i, b)) gt 0 then begin
                  wdef = where(mask_bolo(wfit, b) gt 0.5, cw)
                  if cw ge 100 then begin
                     time_ref = time_s(wfit(wdef(0)))
                     yfit = absdev_fit(float(-time_ref + time_s(wfit(wdef))), flux(wdef, b), coef=coef, $
                        ndev=ndev, noise_level=med_noise, /dyn, ns=ns)
                     drifts_base(tab_aver(it0):tab_aver(it1), b) += (coef(0) + coef(1) * $
                        (-float(time_ref) + time_save(tab_aver(it0):tab_aver(it1))))
                     coef_lin(b) = coef(1)
                  endif
               endif

               if (nscans ge 2 and keyword_set(smallfield)) then begin
                  time_slope = time_save(tab_aver(it0):tab_aver(it1))
                  time_slope -= mean(time_slope)
                  wb = where(finite(maxnoise(i, *)) gt 0., cb)
                  coef_lin = total(coef_lin(wb) / maxnoise(i, wb)^2.) / total(1. / maxnoise(i, wb)^2.)
                  drifts_subs = drifts_base(tab_aver(it0):tab_aver(it1), *)
                  drifts_subs(*, wb) -= rebin(coef_lin * time_slope, tab_aver(it1)-tab_aver(it0)+1, cb)
                  drifts_base(tab_aver(it0), 0) = drifts_subs
               endif
            endif


            ;;; eliminate discontinuities between successive scan legs:
            flag_save(tab_aver(it0_nom)+1:tab_aver(it1_nom)-1) = 1
         endfor


         wsave = where(flag_save eq 1, csave)
         time_ref = time_s(ind_scans(i))
         for b = 0L, nb - 1 do if finite(maxnoise(i, b)) gt 0 then begin
            drifts_base(ind_sc_coarse(i), b) = interpol([drifts_base(wsave(0), b), $
               drifts_base(wsave, b), drifts_base(wsave(csave-1), b)], $
               [-float(timestep), -float(time_ref) + time_save(wsave), float(time_s(ind_scans(i+1)-1) + timestep - time_ref)], $
               -float(time_ref) + time_save(ind_sc_coarse(i):ind_sc_coarse(i+1)-1))
            subs_drifts = interpol([drifts_base(ind_sc_coarse(i), b), drifts_base(ind_sc_coarse(i):ind_sc_coarse(i+1)-1, b), $
               drifts_base(ind_sc_coarse(i+1)-1, b)], $
               [-float(timestep), -float(time_ref) + time_save(ind_sc_coarse(i):ind_sc_coarse(i+1)-1), $
               float(time_s(ind_scans(i+1)-1) + timestep - time_ref)], $
               float(-time_ref + time_s(ind_scans(i):ind_scans(i+1)-1)))
            signal(ind_scans(i):ind_scans(i+1)-1, b) -= subs_drifts
         endif

         flag_save *= 0
      endfor

      drifts_all += drifts_base
   endif


   save, file=file_signal, signal, /xdr



   ;;; produce source mask for next iteration:

   if (k eq 0 or k lt nloop - 1) then begin
      if keyword_set(cut_start) then if max(cut_start) gt 0 then begin
         signal_save = signal
         wcut = where(cut_start gt 0, cw_cut)
         for c = 0, cw_cut - 1 do begin
            subs = where(ind_subscans ge ind_scans(wcut(c)))
            signal(ind_subscans(subs(0)):ind_subscans(subs(0)+1)-1, *) = median(signal)
         endfor
         save, file=file_signal, signal, /xdr
         proj_map, file_scalars=file_scalars, maxnoise=maxnoise, ind_scans=ind_scans
         signal = signal_save
         signal_save = 0
         save, file=file_signal, signal, /xdr
      endif else proj_map, file_scalars=file_scalars, maxnoise=maxnoise, ind_scans=ind_scans
      restore, file_maps

      map_back = map_signal
      map_back_rot = map_back
      w = where(map_weight gt 0.)
      med_weight = median(map_weight(w))
      w = where(map_weight le med_weight/2.)
      map_back_rot(w) = !values.f_nan
      if abs(maprot_deg) gt 1.e-4 then map_back_rot = rot(map_back_rot, maprot_deg, missing=!values.f_nan)
      dimx_rot = n_elements(map_back_rot(*, 0))
      dimy_rot = n_elements(map_back_rot(0, *))

      w = where(finite(map_back_rot) gt 0., np_tot)
      wx = w mod dimx_rot
      wy = w / dimx_rot
      wx_min = min(wx)
      wx_max = max(wx)
      wy_min = min(wy)
      wy_max = max(wy)

      stat_map, med, sig, map=map_back, weightmap=map_weight, noise_level=med_wn
      if nscans gt 20 then sig *= sqrt(float(nscans) / 20.)

      nsig = 2.5
      w = where(finite(map_back_rot) gt 0. and map_back_rot gt med + nsig * sig, cw)
      wx = w mod dimx_rot
      wy = w / dimx_rot
      wmid = where(wx ge (2*wx_min+wx_max)/3 and wx le (wx_min+2*wx_max)/3 $
         and wy ge (2*wy_min+wy_max)/3 and wy le (wy_min+2*wy_max)/3, cmid)
      if cw ge 1 then ratio_mid = float(cmid) / float(cw) else ratio_mid = 1.
      ratio_thres = float(cw) / float(np_tot)


      if (k eq nloop - 2 and keyword_set(debug)) then $
         print, '              cmid / cw = ' + chain(ratio_mid, 2) + ' and cw / np_tot = ' + chain(ratio_thres, 2)

      if k eq nloop - 2 then begin
         if ((ratio_mid lt 0.5 and ratio_thres gt 0.25) $
            or (ratio_mid / ratio_thres lt 2. and ratio_thres gt 0.15)) then begin
            smallfield = 1
            if (not keyword_set(galactic)) then begin
               print, '              constraining the average baseline derived'
               print, '              subscan by subscan to have a null slope'
            endif
         endif
      endif


      if inst eq 'spire' then rt_min = -1. else rt_min = 0.03
      while ((ratio_mid lt 0.3 and ratio_thres gt rt_min and (not keyword_set(galactic)) and (not keyword_set(smallfield))) $
         or (ratio_mid lt 0.7 and ratio_thres gt 0.2) or (ratio_mid / ratio_thres lt 3.5 and ratio_thres gt 0.1)) do begin
         nsig += 0.5
         w = where(finite(map_back_rot) gt 0. and map_back_rot gt med + nsig * sig, cw)
         wx = w mod dimx_rot
         wy = w / dimx_rot
         wmid = where(wx ge (2*wx_min+wx_max)/3 and wx le (wx_min+2*wx_max)/3 $
            and wy ge (2*wy_min+wy_max)/3 and wy le (wy_min+2*wy_max)/3, cmid)
         if cw ge 1 then ratio_mid = float(cmid) / float(cw) else ratio_mid = 1.
         ratio_thres = float(cw) / float(np_tot)
      endwhile

      if cw ge 1 then begin
         map_back_rot(w) = !values.f_nan
         w = where(map_weight gt 0. and map_back gt med + nsig * sig, cw)
         map_back(w) = !values.f_nan
      endif

      if keyword_set(debug) then begin
         time_2 = systime(/seconds)
         t_proc += (time_2 - time_1)
         print, 'nsig = ' + chain(nsig, 1)
         weight_back = fltarr(dimx, dimy)
         w = where(map_weight gt 0. and finite(map_back) gt 0.)
         weight_back(w) = 10.
         disp_ima, win=10, map_back, weightmap=weight_back, title='source mask for baseline fits'
         res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
         time_1 = systime(/seconds)
      endif
   endif

endfor


if keyword_set(visu) then begin
   time_2 = systime(/seconds)
   t_proc += (time_2 - time_1)
   wset, 0
   otherplot, time_save, drifts_all(*, bref), charsize=2., xtitle='coarse time (s)', $
      title='baseline for bolometer ' + chain(bref)
   for i = 0, nscans - 1 do begin
      otherplot, /over, lstyle=2, color=200, time_s(ind_scans(i))+[0., 0.], [-100., 100.]
      otherplot, /over, lstyle=2, color=200, time_s(ind_scans(i+1)-1)+[0., 0.], [-100., 100.]
   endfor
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   time_1 = systime(/seconds)
endif


;;; remove flux calibration offsets from drifts:
for b = 0L, nb - 1L do drifts_all(*, b) -= offsets(b)


if (not keyword_set(galactic)) then begin
   w = where(finite(signal) gt 0)
   med = median(signal(w))
   sig = sqrt(mean((signal(w) - med)^2.))
   for k = 0, 9 do begin
      wl = where(signal(w) - med lt 2.5 * sig)
      med = median(signal(w(wl)))
      sig = sqrt(mean((signal(w(wl)) - med)^2.))
   endfor
endif else begin
   nsamp = nb * (2L + ceil(nt / 1000))
   w = where(finite(signal) gt 0)
   ind_sort = sort(signal(w))
   med = median(signal(w(ind_sort(0:nsamp))))
endelse
signal -= med


save, file=file_signal, signal, /xdr


end
