pro compute_imprint, file_scalars=file_scalars, pixsize=pixsize, diameter=diameter, $
                     nstep_imprint_final=nstep_imprint_final, file_imprint_final=file_imprint_final


;;; projects an axisymmetric bolometer imprint on a fine grid with given pixel size
;;; and number of steps within a pixel
;;; currently, imprint defined as a uniform disk, but could be changed with minimal effort
;;; (module sufficiently flexible)

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

if keyword_set(nstep_imprint_final) then nstep_imprint = nstep_imprint_final
if keyword_set(file_imprint_final) then file_imprint = file_imprint_final


if not keyword_set(pixsize) then pixsize = pix_as
if not keyword_set(diameter) then diameter = diam_as_imprint
nstep_imprint = max([nstep_imprint, 2 * ceil(2.5 * pixsize / diameter)])
nc = strlen(file_imprint)
file_imprint = strmid(file_imprint, 0, nc-6) + chain(nstep_imprint) + '.xdr'

nstep_imprint_final = nstep_imprint
file_imprint_final = file_imprint

print, ''
print, 'projecting bolometer imprint on ' + chain(pixsize, 2) + ' arcsec pixels with ' $
   + chain(nstep_imprint) + ' steps per pixel'


model_r_as = findgen(10000L)/ 100.
model_imprint = fltarr(10000L)
w = where(model_r_as le diameter / 2.)
model_imprint(w) = 1.

pix_cen = dim_imprint / 2
np = nstep_imprint^2
imprint = fltarr(np, dim_imprint, dim_imprint)

xc_imprint = fltarr(np)
yc_imprint = fltarr(np)
for i = 0L, nstep_imprint - 1 do begin
   xc_imprint(i*nstep_imprint:(i+1)*nstep_imprint-1) = findgen(nstep_imprint)
   yc_imprint(i*nstep_imprint:(i+1)*nstep_imprint-1) = i
endfor
xc_imprint = xc_imprint / float(nstep_imprint) + pix_cen
yc_imprint = yc_imprint / float(nstep_imprint) + pix_cen

dim_fine = dim_imprint * nstep_imprint
tab_x_fine = 0.5 + (fltarr(dim_fine) + 1.) ## findgen(dim_fine)
tab_y_fine = 0.5 + findgen(dim_fine) ## (fltarr(dim_fine) + 1.)
xc_fine = xc_imprint * nstep_imprint
yc_fine = yc_imprint * nstep_imprint

scale = pixsize / float(nstep_imprint)
for i = 0L, np - 1 do begin
   ray_as = sqrt((tab_x_fine - xc_fine(i))^2. + (tab_y_fine - yc_fine(i))^2.) * scale
   beam_fine = interpol(model_imprint, model_r_as, ray_as)
   beam = rebin(beam_fine, dim_imprint, dim_imprint)
   imprint(i, *, *) = beam / total(beam)
endfor


;;; useful only if the PSF is projected, rather than the pixel imprint
;;; in this implementation: imprint area, not beam area despite the name
beam_sqas = total(beam_fine) * scale^2.


save, file=file_imprint, imprint, xc_imprint, yc_imprint, beam_sqas, /xdr

save, file=file_scalars, file_times, file_time_coarse, file_signal, file_pixcoords, $
   file_projcoarse, file_projfine, file_imprint, file_maps, $
   file_vscan, file_ntmin_pass, $
   inst, nt, nt_save, nt_aver, timestep, maxtimediff, nb, bref, nscans, nsubscans, $
   logf_inf, logf_sup, logf_high_inf, logf_high_sup, $
   vscan_as_s, dimx, dimy, dimx_iter, dimy_iter, fact, $
   pix_as, diam_as_imprint, dim_imprint, nstep_imprint, xcmin_imprint, $
   npass_crit, fact_npass_cen, visu, debug, /xdr


end
