pro compute_indiv_drifts, array_drifts, array_npass, w_glitches, nb_glitch, $
                          maxnoise=maxnoise, highfnoise=highfnoise, ind_scans=ind_scans, ind_aver=ind_aver, $
                          file_scalars=file_scalars, min_nsamp=min_nsamp, detect_gl=detect_gl, $
                          t_cross_fwhm=t_cross_fwhm, f_fwhm=f_fwhm, time_1=time_1, t_proc=t_proc


;;; computes the individual drift of each bolometer by using the redundancy,
;;; after the average drift has been subtracted
;;; switches to a finer grid or drops the computation whenever significant flux variations are detected
;;; (thus avoids contamination from compact sources, strong brightness gradients, and glitches)
;;; performs second-pass deglitching

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

restore, file_pixcoords

restore, file_signal


;;; to correct for slight pointing differences of the samples within the coarse pixel,
;;; using the map produced at the previous iteration

restore, file_maps

map_error = 2.5 / sqrt(map_weight)
med_noise = median(maxnoise)

fact_bin = 5L
map_rebin = rebin(map_signal, fact_bin*dimx, fact_bin*dimy)
map_weight = 0
diff_signals = signal
for b = 0L, nb - 1 do diff_signals(*, b) -= $
   map_rebin(floor(fact_bin*x_bolos(*, b)) + fact_bin*dimx * floor(fact_bin*y_bolos(*, b)))
map_rebin = 0


restore, file_times
flag_nza = 0

tab_aver = lonarr(nt)
tab_aver(0:ind_aver(0)-1) = 0L
for i = 1L, nt_save - 1 do tab_aver(ind_aver(i-1):ind_aver(i)-1) = i


restore, file_projcoarse

;;; coarse map to visualize the drift correction type:
map_proj = fltarr(dimx_iter, dimy_iter)
dimxy_iter = dimx_iter * dimy_iter
map_proj(wproj mod dimxy_iter) = 2
if inst eq 'pacs' then begin
   map_pointcorr = fltarr(dimx_iter, dimy_iter)
   map_pointcorr(wproj mod dimxy_iter) = 2
endif


restore, file_ntmin_pass

w = where(ntmin_pass lt 3. * median(ntmin_pass))
med_ntmin_pass = median(ntmin_pass(w))
if med_ntmin_pass ge min_nsamp * fact then begin
   flag_fine = 1
   ntmin_pass_fine = ntmin_pass / fact
   w = where(ntmin_pass_fine lt min_nsamp, cw)
   if cw ge 1 then ntmin_pass_fine(w) = min_nsamp
   nproj_zoom = fact^2
   tab_dxpf = indgen(nproj_zoom) mod fact
   tab_dypf = indgen(nproj_zoom) / fact
endif else begin
   flag_fine = 0
   x_bolos = 0
   y_bolos = 0
endelse
w = where(ntmin_pass lt min_nsamp, cw)
if cw ge 1 then ntmin_pass(w) = min_nsamp



array_drifts = fltarr(nt_save, nb)
array_weight = fltarr(nt_save, nb)
array_npass = fltarr(nt_save, nb)



w_glitches = [-9L]
ncrit_gl = max([3, npass_crit])
fact_cen_gl = 0.75
if f_fwhm gt 2. then fact_cen_gl += (0.1 * (f_fwhm - 2.))
if inst eq 'spire' then thres_oscill = 3. else thres_oscill = 5.
thres_gl = 5.
thres_around = 5.
if inst eq 'spire' then thres_det = 4. else thres_det = 5.
dt_sig_gl = 4. * t_cross_fwhm
n_visu_gl = 0L


save_npass_save = -9L + lonarr(nproj)


ind = sort(signal)
med_all = median(signal(ind(0:2*nt*nb/3)))
ind = 0


sq_fact = fact^2.



off = floor(0.5 * fact) * (wproj / dimxy_iter)
x0 = off + ((wproj mod dimxy_iter) mod dimx_iter) * fact
y0 = off + ((wproj mod dimxy_iter) / dimx_iter) * fact
off = 0

for k = 0L, nproj - 1 do begin

   ;;; quicker way of evaluating: wp = where(indxy_beam_coarse eq wproj(k), cp):
   revind_0 = revind(wproj(k))
   revind_1 = revind(wproj(k)+1)
   cp = revind_1 - revind_0
   wp = revind(revind_0:revind_1-1)
   wt = wp mod nt

   if cp ge 3 * median(ntmin_pass(wt)) then begin
      wbolo = wp / nt
      distrib_time = time_s(wt)

      ;;; find the start indices of each distinct bolometer passage through the pixel:
      diff_bolo = [1L, wbolo(1:*) - wbolo(0:cp-2)]
      diff_time = [0.D0, distrib_time(1:*) - distrib_time(0:cp-2)]
      ind_startpass = where(diff_bolo ge 1L or diff_time gt maxtimediff, npass)
      if npass ge 3 then begin
         ind_endpass = [ind_startpass(1:*) - 1, cp-1]

         ;;; correct for pointing differences using the current map:
         if inst eq 'pacs' then begin
            x1 = min([x0(k) + fact - 1, dimx - 1])
            y1 = min([y0(k) + fact - 1, dimy - 1])
            box = map_signal(x0(k):x1, y0(k):y1)
            box_error = map_error(x0(k):x1, y0(k):y1)
            w = where(finite(box_error) gt 0, cw)
            if cw ge 3 then ratio = max(abs(box(w) - mean(box(w)))) / med_noise / mean(box_error(w)) else ratio = 2.
            if ratio lt 1. then begin
               distrib_flux = signal(wp)
               map_pointcorr(wproj(k) mod dimxy_iter) = 1
            endif else distrib_flux = diff_signals(wp)
         endif else distrib_flux = diff_signals(wp)
         distrib_true_flux = signal(wp)

         ;;; to detect whether the flux variations within the pixel require the finer grid:
         ns_pass = ind_endpass - ind_startpass + 1
         flag_ntmin = intarr(npass)

         wcen = where(ns_pass ge (ntmin_pass(wt(ind_startpass)) + ntmin_pass(wt(ind_endpass))) / 2, npass_cen)
         if npass_cen ge 1 then flag_ntmin(wcen) = 1

         ws = intarr(npass)
         for scan = 1, nscans - 1 do begin
            flag_scan = wt(ind_startpass) / ind_scans(scan)
            w = where(flag_scan gt 0, cw)
            if cw ge 1 then ws(w) = scan
         endfor
         hfnoise_pass = highfnoise(ws, wbolo(ind_startpass))
         maxnoise_pass = maxnoise(ws, wbolo(ind_startpass))


         ;;; reduce execution time by looping on maximum number of samples per crossing
         ;;; instead of looping on number of crossings (which scales with number of bolometers):

         ns_max = max(ns_pass)
         if ns_max ge 3 then begin
            tab_flux = fltarr(ns_max, npass) + !values.f_nan
            tab_true_flux = fltarr(ns_max, npass) + !values.f_nan
            for ind_s = 0, ns_max - 1 do begin
               w = where(ns_pass gt ind_s)
               tab_flux(ind_s, w) = distrib_flux(ind_startpass(w)+ind_s)
               tab_true_flux(ind_s, w) = distrib_true_flux(ind_startpass(w)+ind_s)
            endfor
            med_pass = total(tab_flux, 1, /nan) / float(ns_pass)
            true_med_pass = total(tab_true_flux, 1, /nan) / float(ns_pass)
            var_pass = total(abs(tab_flux - (med_pass ## (fltarr(ns_max) + 1.))), 1, /nan) / float(ns_pass)
            var_oscill_pass = sqrt(total((tab_flux(1:ns_max-1, *) - tab_flux(0:ns_max-2, *))^2., 1, /nan) / float(ns_pass - 1))
            sq_maxnoise_pass = maxnoise_pass^2.


            wsave = where(flag_ntmin eq 1 and var_pass lt hfnoise_pass, npass_save)


            ;;; glitch detection:
            if keyword_set(detect_gl) then if (npass_save ge ncrit_gl and npass_save ge $
               floor(fact_cen_gl * npass_cen)) then begin
               med_ref = median(true_med_pass(wsave), /even) - med_all
               wgl = where(flag_ntmin eq 1 and var_oscill_pass gt med_ref $
                  and var_oscill_pass gt thres_oscill * hfnoise_pass, npass_gl)

               for i = 0L, npass_gl - 1L do begin
                  ib = wbolo(ind_startpass(wgl(i)))
                  it = wt((ind_startpass(wgl(i)) + ind_endpass(wgl(i))) / 2)
                  delta_t = abs(time_s - time_s(it))
                  dt_det_gl = ns_pass(wgl(i)) * timestep
                  w = where(delta_t le dt_sig_gl)
                  med_gl = median(signal(w, ib))
                  sig_gl = sqrt(median((signal(w, ib) - med_gl)^2.))
                  dev = max([hfnoise_pass(wgl(i)), sig_gl])
                  w = wt(ind_startpass(wgl(i)):ind_endpass(wgl(i)))
                  peak = max(abs(signal(w, ib) - med_gl), it)
                  it = w(it)
                  delta_t = abs(time_s - time_s(it))
                  w = where(delta_t le dt_det_gl and delta_t gt dt_det_gl / 2.)
                  sig_around = sqrt(median((signal(w, ib) - med_gl)^2.))
                  if (peak gt thres_gl * dev and sig_around lt thres_around * dev) then begin
                     w = where(delta_t le dt_det_gl / 2.)
                     wz = where(abs(signal(w, ib) - med_gl) gt thres_det * hfnoise_pass(wgl(i)), cz)
                     if cz ge 1 then begin
                        wz = w(min(wz):max(wz))
                        w_glitches = [w_glitches, wz + nt * ib]

                        n_visu_gl += 1L
                        if keyword_set(debug) then if (n_visu_gl ge 100 and n_visu_gl le 120) then begin
                           time_2 = systime(/seconds)
                           t_proc += (time_2 - time_1)
                           wset, 0
                           w = where(time_s ge time_s(it) - 5. and time_s le time_s(it) + 5.)
                           wdef = where(finite(signal(w, ib)) gt 0)
                           otherplot, psym=-4, time_s(w(wdef)), signal(w(wdef), ib), xtitle='time (s)', $
                              ytitle='signal (Jy/bol.)', title='bolometer ' + chain(ib)
                           otherplot, /over, color=10, psym=4, symbsize=2., time_s(wz), signal(wz, ib)
                           res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
                           time_1 = systime(/seconds)
                        endif

                     endif
                  endif
               endfor
            endif

            if keyword_set(debug) then save_npass_save(k) = npass_save



            ;;; quasi-uniform flux -> coarse grid:
            if npass_save ge max([npass_crit, floor(fact_npass_cen * npass_cen)]) then begin

               map_proj(wproj(k) mod dimxy_iter) = 1

               itmean_pass = wt((ind_startpass(wsave) + ind_endpass(wsave)) / 2)
               it_aver = tab_aver(itmean_pass)

               weight = 1. / sq_maxnoise_pass(wsave)
               val = total(med_pass(wsave) * weight) / total(weight)

               drift_pass = med_pass(wsave) - val
               weight = 1. / (sq_maxnoise_pass(wsave) + mean(sq_maxnoise_pass(wsave)) / float(npass_save))

               bolo_pass = wbolo(ind_startpass(wsave))
               ind_array = it_aver + nt_save * bolo_pass
               array_drifts(ind_array) += (drift_pass * weight)
               array_weight(ind_array) += weight
               array_npass(ind_array) += (npass_save * weight)



            ;;; varying flux -> finer grid:
            endif else if flag_fine eq 1 then begin
               xpf0 = min(floor(x_bolos(wp)))
               ypf0 = min(floor(y_bolos(wp)))
               for kf = 0L, nproj_zoom - 1 do begin

                  xpf = xpf0 + tab_dxpf(kf)
                  ypf = ypf0 + tab_dypf(kf)
                  wpf = where(floor(x_bolos(wp)) eq xpf and floor(y_bolos(wp)) eq ypf, cpf)
                  if cpf ge 1 then wtf = wt(wpf) else wtf = 0
                  if cpf ge 3 * mean(ntmin_pass_fine(wtf)) then begin
                     wpsub = wp(wpf)
                     wfbolo = wbolo(wpf)
	             distrib_time_f = distrib_time(wpf)
                     diff_bolo = [1L, wfbolo(1:*) - wfbolo(0:cpf-2)]
                     diff_time = [0.D0, distrib_time_f(1:*) - distrib_time_f(0:cpf-2)]
                     ind_startpass = where(diff_bolo ge 1L or diff_time gt maxtimediff, npass)
                     if npass ge 2 then ind_endpass = [ind_startpass(1:*) - 1, cpf-1] else ind_endpass = [cpf-1]
                     distrib_flux_f = diff_signals(wpsub)

                     ns_pass = ind_endpass - ind_startpass + 1
                     flag_ntmin = intarr(npass)


                     wcen = where(ns_pass ge (ntmin_pass_fine(wtf(ind_startpass)) $
                        + ntmin_pass_fine(wtf(ind_endpass))) / 2, npass_cen)
                     if npass_cen ge 1 then flag_ntmin(wcen) = 1

                     ws = intarr(npass)
                     for scan = 1, nscans - 1 do begin
                        flag_scan = wtf(ind_startpass) / ind_scans(scan)
                        w = where(flag_scan gt 0, cw)
                        if cw ge 1 then ws(w) = scan
                     endfor
                     hfnoise_pass = highfnoise(ws, wfbolo(ind_startpass))
                     maxnoise_pass = maxnoise(ws, wfbolo(ind_startpass))


                     ;;; reduce execution time by looping on maximum number of samples per crossing
                     ;;; instead of looping on number of crossings (which scales with number of bolometers):

                     ns_max = max(ns_pass)
                     tab_flux = fltarr(ns_max, npass) + !values.f_nan
                     for ind_s = 0, ns_max - 1 do begin
                        w = where(ns_pass gt ind_s)
                        tab_flux(ind_s, w) = distrib_flux_f(ind_startpass(w)+ind_s)
                     endfor
                     med_pass = total(tab_flux, 1, /nan) / float(ns_pass)
                     var_pass = total(abs(tab_flux - (med_pass ## (fltarr(ns_max) + 1.))), 1, /nan) / float(ns_pass)
                     sq_maxnoise_pass = maxnoise_pass^2.

                     wsave = where(flag_ntmin eq 1 and var_pass lt hfnoise_pass, npass_save)


                     if npass_save ge max([npass_crit, floor(fact_npass_cen * npass_cen)]) then begin

                        map_proj(wproj(k) mod dimxy_iter) = 3

                        itmean_pass = wtf((ind_startpass(wsave) + ind_endpass(wsave)) / 2)
                        it_aver = tab_aver(itmean_pass)

                        weight = 1. / sq_maxnoise_pass(wsave)
                        val = total(med_pass(wsave) * weight) / total(weight)

                        drift_pass = med_pass(wsave) - val
                        weight = 1. / (sq_maxnoise_pass(wsave) + mean(sq_maxnoise_pass(wsave)) / float(npass_save)) $
                           / sq_fact

                        bolo_pass = wfbolo(ind_startpass(wsave))
                        ind_array = it_aver + nt_save * bolo_pass
                        array_drifts(ind_array) += (drift_pass * weight)
                        array_weight(ind_array) += weight
                        array_npass(ind_array) += (npass_save * weight)
                     endif

                  endif
               endfor
            endif
         endif
      endif
   endif
endfor

w = where(array_weight gt 0)
array_drifts(w) /= array_weight(w)
array_weight = 0


if keyword_set(detect_gl) then begin
   nb_glitch = n_elements(w_glitches) - 1
   if nb_glitch ge 1 then begin
      w_glitches = w_glitches(1:*)
      w_glitches = w_glitches(uniq(w_glitches, sort(w_glitches)))
      nb_glitch = n_elements(w_glitches)
   endif
endif else nb_glitch = 0


if keyword_set(visu) then begin
   time_2 = systime(/seconds)
   t_proc += (time_2 - time_1)
   if keyword_set(debug) then begin
      wset, 0
      histo = histogram(save_npass_save, locations=index, min=2)
      otherplot, index, histo, psym=10, charsize=1.5, $
         title='histogram of npass_save ; npass_crit = ' + chain(npass_crit)
      res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   endif

   disp_ima, map_proj, win=5, title='drift mode: 1 -> coarse grid ; 3 -> fine grid ; 2 -> none', $
      min_map=0.01, col_tab=15, /small
   if (inst eq 'pacs' and keyword_set(debug)) then disp_ima, map_pointcorr, win=6, $
      title='pointing correction: 1 -> not applied ; 2 -> applied or no drift', min_map=0.01, col_tab=15, /small
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   time_1 = systime(/seconds)
endif

end
