pro compute_spd_signal, freq_comp, spd_comp_aver, file_scalars=file_scalars, $
    maxnoise=maxnoise, ind_scans=ind_scans, ind_subscans=ind_subscans


;;; computes spectral density of signal, attempting to interpolate missing samples
;;; for visualization purposes only (not used for the processing)

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

restore, file_signal
;;; beware: signal is modified, but the modifications must NOT be saved !


restore, file_times


;;; replace missing samples:

w = where(finite(signal) lt 1, cw)
mask_bolo = 1. + fltarr(nt, nb)
if cw ge 1 then mask_bolo(w) = 0.

seed = 1111L
for b = 0L, nb - 1 do for i = 0, nscans - 1 do if finite(maxnoise(i, b)) gt 0 then begin
   wz = where(mask_bolo(ind_scans(i):ind_scans(i+1)-1, b) lt 1., cz, complement=wdef, ncomplement=cdef)
   if cz ge 1 then begin
      tab_flux = [signal(ind_scans(i)+wdef(0), b), signal(ind_scans(i)+wdef, b), $
         signal(ind_scans(i)+wdef(cdef-1), b)]
      tab_time = [time_s(ind_scans(i))-timestep, time_s(ind_scans(i)+wdef), $
         time_s(ind_scans(i+1)-1)+timestep]
      signal(ind_scans(i)+wz, b) = interpol(tab_flux, tab_time, time_s(ind_scans(i)+wz)) $
         + maxnoise(i, b) * randomn(seed, cz)
   endif
endif


nt_scan = min(ind_scans(1:nscans) - ind_scans(0:nscans-1))

power = 0
repeat begin
   nt_div = nt_scan / 2^power
   rest = nt_div mod 2
   if rest gt 0 then nt_scan -= 2^power
   power += 1
endrep until nt_scan / 2^power le 6
dt = timestep * nt_scan
logf0 = -alog10(dt) - 0.1
freq = [findgen(nt_scan/2+1) / dt, -reverse((findgen(nt_scan/2-1) + 1.) / dt)]
spd_aver = fltarr(nt_scan)

for scan = 0, nscans - 1 do begin
   power = abs(fft(signal(ind_scans(scan):ind_scans(scan)+nt_scan-1, *), dimension=1))
   wb = where(total(mask_bolo(ind_scans(scan):ind_scans(scan)+nt_scan-1, *), 1) gt nt_scan / 4 $
      and finite(maxnoise(scan, *)) gt 0, cwb)
      if cwb ge nb / 5 then spd_aver += total(power(*, wb) / rebin(maxnoise(scan, wb)^2., nt_scan, cwb), 2) $
         / total(1. / maxnoise(scan, wb)^2.) * sqrt(dt)
endfor
spd_aver /= float(nscans)


wset, 0
logf1 = -alog10(timestep) - 0.2
w = where(freq gt 0., cw)
yrange = [alog10(min(spd_aver(w(0:cw-2))))-0.2, alog10(max(spd_aver(w)))+0.2]
if keyword_set(spd_comp_aver) then begin
   yrange(0) = min([yrange(0), alog10(max(spd_comp_aver(w)))-0.2])
   yrange(1) = max([yrange(1), alog10(max(spd_comp_aver(w)))+0.2])
   otherplot, [alog10(freq(w(0)))], [alog10(spd_aver(w(0)))], psym=4, $
      charsize=2., xrange=[logf0, logf1], yrange=yrange, $
      xtitle='log f (Hz)', ytitle='log (signal SD) (Jy/bol. Hz!u-0.5!n)', $
      title='weighted average of all bolometers'
endif else otherplot, alog10(freq(w)), alog10(spd_aver(w)), psym=-4, $
   charsize=2., xrange=[logf0, logf1], yrange=yrange, $
   xtitle='log f (Hz)', ytitle='log (signal SD) (Jy/bol. Hz!u-0.5!n)', $
   title='weighted average of all bolometers'
wdef = where(finite(maxnoise) gt 0)
hfnoise = mean(maxnoise(wdef)) * sqrt(timestep)
otherplot, /over, alog10(freq(w)), w*0.+alog10(hfnoise)
if keyword_set(spd_comp_aver) then begin
   wcomp = where(freq_comp gt 0.)
   otherplot, /over, alog10(freq_comp(wcomp)), alog10(spd_comp_aver(wcomp)), color=10., psym=-4
   otherplot, /over, alog10(freq(w)), alog10(spd_aver(w)), psym=-4
   xyouts, [0.4*logf0+0.6*logf1], [0.1*yrange(0)+0.9*yrange(1)], 'initial spectral density', $
      charsize=2., charthick=2., color=10
   xyouts, [0.4*logf0+0.6*logf1], [0.2*yrange(0)+0.8*yrange(1)], 'spectral density after correction', $
      charsize=2., charthick=2., color=0
endif else begin
   freq_comp = freq
   spd_comp_aver = spd_aver
endelse

res = dialog_message('   Continue ?   ', /center, /information, title='Pause')

end
