pro compute_wnoise, file_scalars=file_scalars, maxnoise, highfnoise, update_mask=update_mask, $
                    update_proj=update_proj, ind_scans=ind_scans, ind_subscans=ind_subscans, $
                    initial=initial, mask_gl=mask_gl, f_fwhm=f_fwhm, cut_start=cut_start, $
                    wp_transient=wp_transient, time_1=time_1, t_proc=t_proc


;;; computes high-frequency noise estimates from spectral densities
;;; "white noise": used to define convergence criterion and compute weights
;;; "threshold noise": used to distinguish between uniform and non-uniform brightness

;;; also performs a first-pass deglitching

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

restore, file_signal
;;; beware: signal is modified, but the modifications must NOT be saved !

restore, file_times


if (keyword_set(debug) or keyword_set(mask_gl)) then begin
   signal_save = signal
   if keyword_set(mask_gl) then begin
      restore, file_maps
      map_signal_save = map_signal
      map_weight_save = map_weight
   endif
endif

n_gl = 0L
nexc = 0L


db = 0
w = where(finite(signal(*, bref)) gt 0, cw)
while (bref + db lt nb - 1 and cw lt nt / 3) do begin
   db += 1
   w = where(finite(signal(*, bref+db)) gt 0, cw)
endwhile
if cw lt nt / 3 then begin
   print, ''
   print, 'too many samples set to NaN !'
   print, 'Please check the mask settings in HIPE (especially the BADPIXELS mask).'
   print, 'Exit.'
   retall
endif
if (keyword_set(debug)) then signal_ref = signal(*, bref+db)


mask_bolo = 1. + fltarr(nt, nb)
w = where(finite(signal) lt 1, cw)
if cw ge 1 then mask_bolo(w) = 0.
w = 0

signal_interp = signal

restore, file_vscan
w = where(vscan lt 3., cw)
if cw ge 1 then signal_interp(w) = !values.f_nan



noise_interp = fltarr(nscans)
thres_interp = 12.
for s = 0, nscans - 1 do begin
   diff = abs(signal_interp(ind_scans(s)+1:ind_scans(s+1)-1, *) - signal_interp(ind_scans(s):ind_scans(s+1)-2, *))
   w = where(diff gt 0.)
   sig = sqrt(median(diff(w)^2.))
   if sig lt 1.e-3 then sig = sqrt(mean(diff(w)^2.))
   w = where(diff gt 0. and diff lt thres_interp * sig)
   noise_interp(s) = sqrt(mean(diff(w)^2.) / 2.)
endfor

if keyword_set(debug) then begin
   print, 'fiducial noise: '
   print, chain(noise_interp, 5)
endif


if inst eq 'spire' then thres = 6. * noise_interp else thres = 9. * noise_interp


if (not keyword_set(initial)) then maxnoise_save = maxnoise

maxnoise = fltarr(nscans, nb) + !values.f_nan
highfnoise = fltarr(nscans, nb) + !values.f_nan

logf_min = -1.05
dlogf = 0.1
logf_max = -alog10(2. * timestep)
nf = 1 + floor((logf_max - logf_min) / dlogf)
logf_bin = logf_min + dlogf * findgen(nf)


maxnoise_init = 1. + fltarr(nscans, nb)
hw = 3
if f_fwhm gt 3. then hw += round(0.5 * (f_fwhm - 3.))
for i = 0, nscans - 1 do begin
   nt_scan = ind_scans(i+1) - ind_scans(i)
   w = where(vscan(ind_scans(i):ind_scans(i+1)-1) ge 3., cw, ncomplement=clv)
   nt_thres = clv + (3L * cw) / 4L
   diff1 = fltarr(nt_scan, nb)
   diff1(3:nt_scan-1, *) = abs(signal_interp(ind_scans(i)+3:ind_scans(i+1)-1, *) $
      - signal_interp(ind_scans(i):ind_scans(i+1)-4, *))
   diff2 = fltarr(nt_scan, nb)
   diff2(2:nt_scan-1, *) = abs(signal_interp(ind_scans(i)+2:ind_scans(i+1)-1, *) $
      - signal_interp(ind_scans(i):ind_scans(i+1)-3, *))
   diff3 = fltarr(nt_scan, nb)
   diff3(1:nt_scan-1, *) = abs(signal_interp(ind_scans(i)+1:ind_scans(i+1)-1, *) $
      - signal_interp(ind_scans(i):ind_scans(i+1)-2, *))
   for b = 0L, nb - 1 do begin
      w = where(diff1(*, b) gt thres(i) or diff2(*, b) gt thres(i) or diff3(*, b) gt thres(i), cw)
      for k = 0L, cw - 1 do diff1(max([w(k)-hw, 0]):min([w(k)+hw-1, nt_scan-1]), b) = 2. * thres(i)
      wz = where(diff1(*, b) gt thres(i), cz)
      if (cz ge 1 and cz lt nt_thres) then signal_interp(ind_scans(i)+wz, b) = !values.f_nan
      w = where(finite(signal_interp(ind_scans(i):ind_scans(i+1)-1, b)) gt 0, cw)
      if cw ge 3 then signal_interp(ind_scans(i):ind_scans(i)+w(2), b) = !values.f_nan
   endfor

   signal_off = signal(ind_scans(i):ind_scans(i+1)-1, *) - (median(signal(ind_scans(i):ind_scans(i+1)-1, *), $
      dimension=1) ## (1. + fltarr(ind_scans(i+1)-ind_scans(i))))
   signal_off -= min(median(signal_off, dimension=2), /nan)
   w = where(signal_off gt 300. * median(noise_interp, /even), cw)
   if cw ge 1 then signal_interp(ind_scans(i)+w) = !values.f_nan
endfor
signal_off = 0


noise_interp = fltarr(nscans)
for i = 0, nscans - 1 do begin
   diff = abs(signal_interp(ind_scans(i)+1:ind_scans(i+1)-1, *) - signal_interp(ind_scans(i):ind_scans(i+1)-2, *))
   w = where(diff gt 0.)
   sig = sqrt(median(diff(w)^2.))
   w = where(diff gt 0. and diff lt thres_interp * sig)
   noise_interp(i) = sqrt(mean(diff(w)^2.) / 2.)
endfor

if keyword_set(debug) then begin
   print, 'replacement noise: '
   print, chain(noise_interp, 5)
endif


;;; initialize seed to a constant value so that we get reproducible results:
seed = 1111L

for s = 0, nscans - 1 do begin
   w = where(vscan(ind_scans(s):ind_scans(s+1)-1) ge 3., cw, ncomplement=clv)
   nt_thres = clv + (3L * cw) / 4L
   for b = 0L, nb - 1 do begin
      wz = where(finite(signal_interp(ind_scans(s):ind_scans(s+1)-1, b)) lt 1, cz, complement=wdef, ncomplement=cdef)
      if cz ge 1 then begin
         if cz - clv lt (3L * cw) / 4L then begin
            tab_sig = [signal_interp(ind_scans(s)+wdef(0), b), signal_interp(ind_scans(s)+wdef, b), $
               signal_interp(ind_scans(s)+wdef(cdef-1), b)]
            tab_time = [time_s(ind_scans(s))-timestep, time_s(ind_scans(s)+wdef), time_s(ind_scans(s+1)-1)+timestep]
            signal_interp(ind_scans(s)+wz, b) = interpol(tab_sig, tab_time, time_s(ind_scans(s)+wz)) $
               + noise_interp(s) * randomn(seed, cz)
         endif else maxnoise_init(s, b) = !values.f_nan
      endif
   endfor
endfor



if keyword_set(debug) then begin
   time_2 = systime(/seconds)
   t_proc += (time_2 - time_1)
   wset, 0
   w = where(finite(signal_ref) gt 0)
   otherplot, time_s(w), signal_ref(w), xtitle='time (s)', ytitle='bolometer '+chain(bref), $
      title='red: after replacing missing values and samples seeing compact sources'
   otherplot, /over, color=10, time_s(w), signal_interp(w, bref+db)

   signal = signal_save - signal_interp
   for i = 0, nscans - 1 do begin
      w = where(finite(maxnoise_init(i, *)) lt 1, cw)
      if cw ge 1 then signal(ind_scans(i):ind_scans(i+1)-1, w) = !values.f_nan
   endfor
   save, file=file_signal, signal, /xdr
   prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans
   proj_map, file_scalars=file_scalars, maxnoise=maxnoise_init, ind_scans=ind_scans
   restore, file_maps
   disp_ima, map_signal, win=7, title='compact sources replaced to compute white noise', $
      weightmap=map_weight, noise_level=10.*mean(noise_interp)

   res = dialog_message('Continue and delete window ?', /center, /information, title='Pause')
   map_signal = 0
   map_weight = 0
   wdelete, 7
   time_1 = systime(/seconds)
endif



spd_bin_bolo = fltarr(nf, nb)

for i = 0, nscans - 1 do begin
   w = where(total(mask_bolo(ind_scans(i):ind_scans(i+1)-1, *), 2) gt 0., nt_scan)
   flux_scan = signal_interp(ind_scans(i)+w, *)

   power = 0
   repeat begin
      nt_div = nt_scan / 2^power
      rest = nt_div mod 2
      if rest gt 0 then nt_scan -= 2^power
      power += 1
   endrep until nt_scan / 2^power le 6
   dt = timestep * nt_scan
   freq = [findgen(nt_scan/2+1) / dt, -reverse((findgen(nt_scan/2-1) + 1.) / dt)]

   power_bolo = abs(fft(flux_scan(0:nt_scan-1, *), dimension=1))

   spd_bolo = power_bolo * sqrt(dt)
   wb = where(total(mask_bolo(ind_scans(i):ind_scans(i)+nt_scan-1, *), 1) gt nt_scan / 4 $
      and finite(total(spd_bolo, 1)) gt 0, cwb)
   spd = total(spd_bolo(*, wb), 2) / float(cwb)

   w = where(freq gt 0.)
   logf = alog10(freq(w))
   for kf = 0, nf - 1 do begin
      wbin = where(abs(logf - logf_bin(kf)) le dlogf / 2., cbin)
      if cbin ge 1 then spd_bin_bolo(kf, *) = total(spd_bolo(w(wbin), *), 1) / float(cbin)
   endfor

   w = where(logf_bin gt logf_inf and logf_bin lt logf_sup, cw)
   if cw eq 0 then begin
      print, ''
      print, 'inadapted frequency windows !'
      print, 'Use the /parallel option if the data were acquired in parallel mode.'
      print, 'Exit.'
      retall
   endif
   maxnoise(i, *) = total(spd_bin_bolo(w, *), 1) / float(cw)

   w = where(logf_bin gt logf_high_inf and logf_bin lt logf_high_sup, cw)
   if cw eq 0 then begin
      print, ''
      print, 'inadapted frequency windows !'
      print, 'Use the /parallel option if the data were acquired in parallel mode.'
      print, 'Exit.'
      retall
   endif
   highfnoise(i, *) = total(spd_bin_bolo(w, *), 1) / float(cw)


   if (keyword_set(debug) and keyword_set(initial) and i lt 10) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      db = 0
      while finite(maxnoise(i, bref+db)) lt 1 do db += 1
      wset, 0
      w = where(freq gt 0. and spd_bolo(*, bref+db) gt 0.)
      subtitle = ', scan ' + chain(i)
      otherplot, alog10(freq(w)), alog10(spd_bolo(w, bref+db)), title='bolometer ' + chain(bref+db) + subtitle, $
         xtitle='log f (Hz)', ytitle='log signal SD (Jy/bol. Hz!u-0.5!n)'
      otherplot, /over, color=10, alog10(freq(w)), alog10(spd(w))
      otherplot, /over, psym=4, color=200, logf_bin, alog10(spd_bin_bolo(*, bref+db))
      otherplot, /over, color=200, alog10(freq(w)), w*0. + alog10(maxnoise(i, bref+db)), thick=3.
      otherplot, /over, color=100, alog10(freq(w)), w*0. + alog10(highfnoise(i, bref+db)), thick=3.
      otherplot, /over, lstyle=2, color=200, logf_inf+fltarr(2), [-10.,10.], thick=3.
      otherplot, /over, lstyle=2, color=200, logf_sup+fltarr(2), [-10.,10.], thick=3.
      otherplot, /over, lstyle=2, color=100, logf_high_inf+fltarr(2)+0.02, [-10.,10.], thick=3.
      otherplot, /over, lstyle=2, color=100, logf_high_sup+fltarr(2)+0.02, [-10.,10.], thick=3.
      res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
      time_1 = systime(/seconds)
   endif

endfor


;;; because of FFT normalization in IDL:
maxnoise /= sqrt(timestep)
highfnoise /= sqrt(timestep)


if (not keyword_set(initial)) then begin
   w = where(finite(maxnoise_save) lt 1, cw)
   if cw ge 1 then begin
      maxnoise(w) = !values.f_nan
      highfnoise(w) = !values.f_nan
   endif
   w = where((finite(maxnoise) lt 1 and finite(maxnoise_save) gt 0) or maxnoise lt 0.1 * median(maxnoise), cw)
   if cw ge 1 then begin
      update_mask = 1
      update_proj = 1
   endif
endif


;;; in case the high-frequency signal is stronger that the pipeline noise amplification (SPIRE):
wdef = where(finite(maxnoise) gt 0)
ratio = highfnoise(wdef) / maxnoise(wdef)
med = median(ratio)
if med lt 1. then begin
   highfnoise_save = highfnoise
   highfnoise = maxnoise
   maxnoise = highfnoise_save
   ratio = highfnoise(wdef) / maxnoise(wdef)
   med = median(ratio)
   highfnoise(wdef) = med * maxnoise(wdef)
   flag_highf = 1
endif else flag_highf = 0

;;; in case the "threshold noise" is smaller than the "white noise":
w = where(highfnoise lt maxnoise, cw)
if cw ge 1 then highfnoise(w) = maxnoise(w)



;;; glitch masking:

if keyword_set(mask_gl) then begin
   if inst eq 'spire' then begin
      thres_gl = 3. * median(maxnoise)
      fact_gl = 3.
   endif else begin
      thres_gl = 4. * median(maxnoise)
      fact_gl = 4.
   endelse
   signal_glitches = fltarr(nt, nb)

   map_signal_orig = map_signal_save
   w = where(map_weight_save gt 0., complement=wz)
   map_signal_orig(wz) = !values.f_nan
   thres_weight = 0.7 * median(map_weight_save(w))

   dr_min = round(3. * fact / float(f_fwhm))
   dr_max = round(6. * fact / float(f_fwhm))
   nr = 2 * dr_max + 1
   if f_fwhm gt 4. then ns_buf = 2L else if f_fwhm gt 2. then ns_buf = 6L else ns_buf = 12L


   signal = signal_save - signal_interp
   for i = 0, nscans - 1 do begin
      w = where(finite(maxnoise(i, *)) lt 1, cw)
      if cw ge 1 then signal(ind_scans(i):ind_scans(i+1)-1, w) = !values.f_nan
      if keyword_set(cut_start) then if cut_start(i) gt 0 then begin
         w = where(ind_subscans ge ind_scans(i))
         signal(ind_subscans(w(0)):min([(ind_subscans(w(0)+1)+ind_subscans(w(0)))/2, ind_subscans(w(0))+1000]), *) = !values.f_nan
      endif
   endfor
   save, file=file_signal, signal, /xdr
   prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans
   proj_map, file_scalars=file_scalars, maxnoise=maxnoise, ind_scans=ind_scans
   restore, file_maps

   restore, file_projfine

   wg = where(abs(map_signal) gt 0., cg)
   for ig = 0L, cg - 1 do begin
      revind_0 = revind_fine(wg(ig))
      revind_1 = revind_fine(wg(ig)+1)
      if revind_1 gt revind_0 then begin
         wp = revind_fine(revind_0:revind_1-1)
         wt = wp mod nt
         cp = revind_1 - revind_0
         ws = intarr(cp)
         for subs = 1, nsubscans - 1 do begin
            flag_subscan = wt / ind_subscans(subs)
            w = where(flag_subscan gt 0, cw)
            if cw ge 1 then ws(w) = subs
         endfor
         med_subs = !values.f_nan + fltarr(nsubscans)
         ns_subs = !values.f_nan + fltarr(nsubscans)
         dev_subs = !values.f_nan + fltarr(nsubscans)
         for subs = 0, nsubscans - 1 do begin
            w = where(ws eq subs, cw)
            if cw ge 1 then begin
               ns_subs(subs) = cw
               med_subs(subs) = median(signal_save(wp(w)), /even)
               dev_subs(subs) = sqrt(mean((signal_save(wp(w)) - med_subs(subs))^2.))
            endif
         endfor

         med_signal = map_signal_orig(wg(ig))
         x = wg(ig) mod dimx
         y = wg(ig) / dimx
         xtot = rebin(x - dr_max + lindgen(nr), nr, nr)
         ytot = (y - dr_max + lindgen(nr)) ## (1L + lonarr(nr))
         w = where(xtot ge 0 and xtot lt dimx and ytot ge 0 and ytot lt dimy)
         wtot = xtot(w) + dimx * ytot(w)
         dr = sqrt((xtot - x)^2. + (ytot - y)^2.)
         w = where(dr ge dr_min and dr le dr_max)
         med_local = median(map_signal_orig(wtot(w)), /even)
         med_source = max(map_signal_orig(max([x-1, 0]):min([x+1, dimx-1]), max([y-1, 0]):min([y+1, dimy-1])), /nan)

         wdef = where(ns_subs ge max([3, median(ns_subs, /even) / 2.]), cdef, complement=wz, ncomplement=cz)

         ;;; transient detection:
         if (keyword_set(wp_transient) and map_weight_save(wg(ig)) ge thres_weight) then if cdef ge 2 then begin
            winf_med = where(med_subs(wdef) le median(med_subs(wdef), /even))
            winf_dev = where(dev_subs(wdef) le median(dev_subs(wdef), /even))
            dev_ref = median(dev_subs(wdef(winf_dev)), /even)
            wsubs = where(abs(med_subs(wdef) - med_signal) gt $
               max([thres_gl, fact_gl * abs(median(med_subs(wdef(winf_med)), /even) - med_local), 9. * dev_ref]) $
               and abs(med_subs(wdef) - med_local) gt 4. * max([dev_ref, median(maxnoise)]), cw)
            for k = 0, cw - 1 do begin
               w = where(ws eq wdef(wsubs(k)))
               if (min(wt(w)) gt 10 and max(wt(w)) lt nt - 11) then for it = -ns_buf, ns_buf do $
                  wp_transient = [wp_transient, wp(w) + it]
            endfor
         endif

         ;;; glitch detection:
         if cdef ge 1 then begin
            wsup = where(dev_subs(wdef) ge median(dev_subs(wdef), /even))
            dev_ref = median(dev_subs(wdef(wsup)), /even)
         endif else dev_ref = median(maxnoise)
         if cdef lt 1 then med_subs(wz) = med_signal $
            else if cz ge 1 then med_subs(wz) = median(med_subs(wdef), /even)
         diff_comp = max([thres_gl, fact_gl * abs(med_source - med_local), 3. * dev_ref])
         w = where(abs(signal_save(wp) - med_signal) gt diff_comp $
            and abs(signal_save(wp) - med_subs(ws)) gt diff_comp, cw)
         if cw ge 1 then begin
            signal_glitches(wp(w)) = signal_save(wp(w))
            signal_save(wp(w)) = !values.f_nan
            n_gl += cw
         endif
      endif
   endfor

   print, 'glitches found by sample differences: ' + chain(n_gl) + ' sample(s) masked'

   if (keyword_set(visu) and n_gl gt 50) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      signal = signal_glitches
      for i = 0, nscans - 1 do begin
         w = where(finite(maxnoise_init(i, *)) lt 1, cw)
         if cw ge 1 then signal(ind_scans(i):ind_scans(i+1)-1, w) = !values.f_nan
      endfor
      save, file=file_signal, signal, /xdr
      prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans
      proj_map, file_scalars=file_scalars, maxnoise=maxnoise_init, ind_scans=ind_scans
      restore, file_maps
      disp_ima, map_signal, weightmap=map_weight, noise_level=median(maxnoise), /trunc_max, $
         win=8, title='glitches'
      res = dialog_message('Continue and delete window ?', /center, /information, title='Pause')
      wdelete, 8
      time_1 = systime(/seconds)
   endif
endif



if (keyword_set(debug) or keyword_set(mask_gl)) then begin
   signal = signal_save
   save, file=file_signal, signal, /xdr
   if n_gl ge 1 then prepare_proj_coarse, file_scalars=file_scalars
   prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans
endif

if keyword_set(debug) then begin
   time_2 = systime(/seconds)
   t_proc += (time_2 - time_1)
   wset, 0
   for s = 0, min([nscans - 1, 9]) do begin
      wdef = where(finite(maxnoise(s, *)) gt 0)
      otherplot, wdef, maxnoise(s, wdef), psym=4, xtitle='bolometer index', ytitle='white noise', $
         title='scan '+chain(s)
      res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   endfor
   time_1 = systime(/seconds)
endif



wdef = where(finite(maxnoise) gt 0)
med_wn = median(maxnoise(wdef))
sig_wn = sqrt(mean((maxnoise(wdef) - med_wn)^2.))
wdef = where(abs(maxnoise - med_wn) lt 6. * sig_wn)
med_wn = median(maxnoise(wdef))
sig_wn = sqrt(mean((maxnoise(wdef) - med_wn)^2.))
if keyword_set(debug) then begin
   print, 'med_wn = ' + chain(med_wn, 5)
   print, 'sig_wn = ' + chain(sig_wn, 5)
endif
if sig_wn gt 0.75 * med_wn then begin
   print, ''
   print, 'The field is too crowded to compute meaningful high-frequency noise values:'
   print, 'assigning the same high-frequency noise to all bolometers.'
   maxnoise(wdef) = med_wn
   highfnoise(wdef) = median(highfnoise(wdef))
endif


if keyword_set(update_mask) then begin
   restore, file_signal
   for s = 0, nscans - 1 do begin

      w = where(finite(maxnoise(s, *)) lt 1, cw)
      if cw ge 1 then begin
         nexc += cw
         if (keyword_set(debug) and s le 19) then $
            print, 'initially: ' + chain(cw) + ' invalid bolometer(s) in scan ' + chain(s)
         signal(ind_scans(s):ind_scans(s+1)-1, w) = !values.f_nan
      endif

      ;;; exclude bolometers with very deviant noise:
      wdef = where(finite(maxnoise(s, *)) gt 0)
      med_wn = median(maxnoise(s, wdef))
      sig_wn = sqrt(mean((maxnoise(s, wdef) - med_wn)^2.))
      thres = 6.
      w = where(maxnoise(s, *) gt min([3. * med_wn, med_wn + thres * sig_wn]) $
         or maxnoise(s, *) lt max([med_wn / 3., med_wn - thres * sig_wn]), cw)
      if cw ge 1 then begin
         nexc += cw
         print, 'excluding ' + chain(cw) + ' noisy or dead bolometer(s) in scan ' + chain(s) $
            + ':  ' + strjoin(chain(w) + '  ')
         if keyword_set(debug) then begin
            wset, 0
            otherplot, wdef, maxnoise(s, wdef), psym=4, xtitle='bolometer index', ytitle='white noise'
            otherplot, /over, color=10, w, maxnoise(s, w), psym=4
            res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
         endif
         signal(ind_scans(s):ind_scans(s+1)-1, w) = !values.f_nan
         maxnoise(s, w) = !values.f_nan
         highfnoise(s, w) = !values.f_nan
      endif

      ;;; exclude bolometers with very deviant spectral density shape:
      if flag_highf lt 1 then begin
         thres = 5.
         wdef = where(finite(maxnoise(s, *)) gt 0)
         ratio = highfnoise(s, wdef) / maxnoise(s, wdef)
         med = median(ratio)
         dev = sqrt(mean((ratio - med)^2.))
         wz = where(ratio gt max([1.3 * med, med + thres * dev]) $
            or ratio lt min([med / 1.3, med - thres * dev]), cw)
         if cw ge 1 then begin
            w = wdef(wz)
            nexc += cw
            print, 'excluding ' + chain(cw) + ' sick bolometer(s) in scan ' + chain(s) $
               + ':  ' + strjoin(chain(w) + '  ')
            if keyword_set(debug) then begin
               wset, 0
               otherplot, maxnoise(s, wdef), ratio, psym=4, xtitle='white noise', $
                  ytitle='threshold noise / white noise'
               otherplot, /over, color=10, maxnoise(s, wdef(wz)), ratio(wz), psym=4
               res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
            endif
            signal(ind_scans(s):ind_scans(s+1)-1, w) = !values.f_nan
            maxnoise(s, w) = !values.f_nan
            highfnoise(s, w) = !values.f_nan
         endif
      endif
   endfor

   ;;; exclude PACS array lines with deviant noise:
   if inst eq 'pacs' then begin
      nb_line = 16L
      nlines = nb / nb_line
      mn_lines = fltarr(nlines) + !values.f_nan
      for kl = 0L, nlines - 1L do begin
         wb = nb_line * kl + lindgen(nb_line)
         tab = maxnoise(*, wb)
         w = where(finite(tab) gt 0, cw)
         if cw ge 1 then mn_lines(kl) = median(tab(w), /even)
      endfor
      med = median(mn_lines, /even)
      w = where(finite(mn_lines) gt 0)
      sig = sqrt(mean((mn_lines(w) - med)^2.))
      w = where(abs(mn_lines - med) gt max([0.1 * med, 3. * sig]), cw)
      if cw ge 1 then print, 'excluding array row(s) ' + strjoin(chain(w) + '  ')
      for i = 0, cw - 1 do begin
         kl = w(i)
         wb = nb_line * kl + lindgen(nb_line)
         nexc += nb_line
         signal(*, wb) = !values.f_nan
         maxnoise(*, wb) = !values.f_nan
         highfnoise(*, wb) = !values.f_nan
      endfor
   endif


   db = 0
   while finite(total(maxnoise(*, bref+db))) lt 1 do db += 1
   if db gt 0 then begin
      bref += db
      save, file=file_scalars, file_times, file_time_coarse, file_signal, file_pixcoords, $
         file_projcoarse, file_projfine, file_imprint, file_maps, $
         file_vscan, file_ntmin_pass, $
         inst, nt, nt_save, nt_aver, timestep, maxtimediff, nb, bref, nscans, nsubscans, $
         logf_inf, logf_sup, logf_high_inf, logf_high_sup, $
         vscan_as_s, dimx, dimy, dimx_iter, dimy_iter, fact, $
         pix_as, diam_as_imprint, dim_imprint, nstep_imprint, xcmin_imprint, $
         npass_crit, fact_npass_cen, visu, debug, /xdr
   endif

   if nexc ge 1 then begin
      save, file=file_signal, signal, /xdr
      if keyword_set(update_proj) then begin
         prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans
         prepare_proj_coarse, file_scalars=file_scalars
      endif
   endif
endif


end
