pro corr_jumps_pacs, file_scalars=file_scalars, ind_scans=ind_scans, ind_subscans=ind_subscans, $
                     maxnoise=maxnoise, f_fwhm=f_fwhm, time_1=time_1, t_proc=t_proc


;;; detects and masks brightness discontinuities affecting whole array rows or single bolometers (PACS),
;;; using the current map as a reference

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; introduced December 2010: version 7

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



crs = string(13b) + string(10b)
restore, file_scalars

restore, file_pixcoords
restore, file_signal
restore, file_times


;;; simplified reference map on a fine grid:

restore, file_maps

mult_xy = 5
dimx_fine = dimx * mult_xy
dimy_fine = dimy * mult_xy
map_ref = rebin(map_signal, dimx_fine, dimy_fine)
map_signal = 0
map_weight = 0



off = 3
margin = 9
naver = ceil(2. / timestep)
nmask = ceil(3. / timestep)
nmask_max = 100L * nmask
njumps_aver = 0L
njumps_indiv = 0L


nb_line = 16L
nlines = nb / nb_line
dev_lines = !values.f_nan + fltarr(nsubscans, nlines)
dev_bolos = !values.f_nan + fltarr(nsubscans, nb)


;;; first pass: automatic masking of very unstable rows and bolometers (deviant ones)
;;; second pass: interactive masking of individual jumps

for loop = 0, 1 do begin

   for kl = 0L, nlines - 1L do begin
      wb = nb_line * kl + lindgen(nb_line)
      w = where(finite(signal(*, wb)) lt 1, cw)
      mask_bolo = 1 + lonarr(nt, nb_line)
      if cw ge 1 then mask_bolo(w) = 0L
      norm = total(mask_bolo, 2)
      norm_denom = norm
      w = where(norm eq 0., cw)
      if cw ge 1 then norm_denom(w) = 1.
      aver_line = total(signal(*, wb) * mask_bolo, 2, /nan) / norm_denom
      if cw ge 1 then aver_line(w) = !values.f_nan
      indiv_line = signal(*, wb)

      indxy_ref = floor(mult_xy * x_bolos(*, wb)) + dimx_fine * floor(mult_xy * y_bolos(*, wb))
      noise_level = 2. * median(maxnoise(*, wb), /even)
      noise_resi = noise_level / 4.


      for subs = 0L, nsubscans - 1L do begin
         ref_map = median(map_ref(indxy_ref(ind_subscans(subs):ind_subscans(subs+1)-1, *)), dimension=2)
         ref_map -= min(ref_map, /nan)
         wdef = where(finite(aver_line(ind_subscans(subs):ind_subscans(subs+1)-1)) gt 0, cdef)
         if cdef ge 4 * naver then begin
            aver_scan = aver_line(ind_subscans(subs)+wdef)
            indiv_scan = indiv_line(ind_subscans(subs)+wdef, *)
            ref = total(map_ref(indxy_ref(ind_subscans(subs)+wdef, *)) * mask_bolo(ind_subscans(subs)+wdef, *), 2, /nan) $
               / norm_denom(ind_subscans(subs)+wdef)
            ref -= median(ref)
            aver_scan -= ref
            ref = map_ref(indxy_ref(ind_subscans(subs)+wdef, *))
            ref -= median(ref)
            indiv_scan -= ref
            wnom = where(flag_nza(ind_subscans(subs)+wdef) gt 0, cnom)
            if cnom gt cdef / 3 then dev_lines(subs, kl) = total(abs(aver_scan(wnom) - mean(aver_scan(wnom)))) / float(cnom)

            if loop eq 1 then begin
               diff = abs(aver_scan(off:cdef-1) - aver_scan(0:cdef-1-off))
               diff(0:2*naver+off) = 0.
               diff(cdef-1-off-2*naver-margin:cdef-1-off) = 0.
               indiv_diff = abs(indiv_scan(off:cdef-1, *) - indiv_scan(0:cdef-1-off, *))
               indiv_diff(0:2*naver+off, *) = 0.
               indiv_diff(cdef-1-off-2*naver-margin:cdef-1-off, *) = 0.
               diff_time = time_s(ind_subscans(subs)+wdef(off:cdef-1)) - time_s(ind_subscans(subs)+wdef(0:cdef-1-off))
               w = where(diff_time gt 10. * maxtimediff, cw)
               if cw ge 1 then begin
                  diff(w) = 0.
                  indiv_diff(w, *) = 0.
               endif

               wj = where(diff gt noise_level and median(indiv_diff, dimension=2) gt noise_level, cj)
               while cj ge 1 do begin
                  kj_init = wj(0)
                  kj = kj_init
                  w = where(abs(aver_scan(kj:kj+off-1) - aver_scan(kj)) lt noise_level)
                  kj += max(w)
                  tab0 = aver_scan(kj-2*naver:kj-naver)
                  tab1 = aver_scan(kj-naver+1:kj)
                  tab2 = aver_scan(kj+margin:kj+margin+naver-1)
                  tab3 = aver_scan(kj+margin+naver:kj+margin+2*naver-1)
                  b0 = median(tab0, /even)
                  b1 = median(tab1, /even)
                  b2 = median(tab2, /even)
                  b3 = median(tab3, /even)
                  delta = min([abs(b2 - b1), abs(b2 - b0)])
                  sig1 = sqrt(mean((tab1 - b1)^2.))
                  sig2 = sqrt(mean((tab2 - b2)^2.))

                  if (delta gt max([noise_level, 3. * sig1, 3. * sig2, 2. * abs(b3 - b2), 2. * abs(b1 - b0)]) $
                   and flag_nza(ind_subscans(subs)+wdef(kj)) gt 0 $
                   and max(abs([b0, b1, b2, b3])) gt max(ref_map(wdef(kj-naver+1:kj+margin+naver-1)))) then begin
                     nmask_tot = 0L
                     delta_resi = 2. * noise_resi
                     while(delta_resi gt noise_resi and kj+nmask_tot+nmask lt cdef - 1 and nmask_tot lt nmask_max) do begin
                        nmask_tot += nmask
                        tab4 = aver_scan(kj+nmask_tot:min([kj+nmask_tot+nmask, cdef-1]))
                        b4 = median(tab4, /even)
                        delta_resi = min([abs(b4 - b0), abs(b4 - b1)])
                     endwhile
                     if (delta_resi gt noise_resi and nmask_tot lt nmask_max) then nmask_tot += nmask
                     nmask_tot += (nmask / 2)

                     ;;; beta-release module: interactive mode enforced to check detections
                     time_2 = systime(/seconds)
                     t_proc += (time_2 - time_1)
                     wset, 0
                     ind1 = max([0, kj - 1500])
                     ind2 = min([cdef - 1, max([kj + 1500, kj+nmask_tot+500])])
                     otherplot, time_s(ind_subscans(subs)+wdef(ind1:ind2)), aver_scan(ind1:ind2), $
                        xtitle='time (s)', ytitle='brightness (Jy/bol.)', $
                        title='average of array row ' + chain(kl) + ' in subscan ' + chain(subs) + ' (masked values in red)'
                     otherplot, /over, color=200, time_s, flag_nza * max(aver_scan(ind1:ind2))
                     otherplot, /over, color=10, time_s(ind_subscans(subs)+wdef(kj:min([kj+nmask_tot, cdef-1]))), $
                        aver_scan(kj:min([kj+nmask_tot, cdef-1]))
                     res = dialog_message('Confirm and mask discontinuity ?', /center, /question)
                     time_1 = systime(/seconds)

                     if res eq 'Yes' then begin
                        diff(kj_init:min([kj+nmask_tot, cdef-1-off])) = 0.
                        indiv_diff(kj_init:min([kj+nmask_tot, cdef-1-off]), *) = 0.
                        nm = min([kj+nmask_tot, cdef-1]) - kj + 1
                        if nm gt 0.75 * cdef then signal(ind_subscans(subs):ind_subscans(subs+1)-1, wb) = !values.f_nan $
                           else signal(ind_subscans(subs)+wdef(kj):ind_subscans(subs)+wdef(min([kj+nmask_tot, cdef-1])), wb) = !values.f_nan
                        njumps_aver += 1
                     endif
                  endif
                  diff(kj_init:kj+off) = 0.
                  indiv_diff(kj_init:kj+off, *) = 0.
                  wj = where(diff gt noise_level and median(indiv_diff, dimension=2) gt noise_level, cj)
               endwhile
            endif



            for kb = 0L, nb_line - 1L do begin
               bol_noise_level = 12. * median(maxnoise(*, wb(0)+kb), /even)
               if cnom gt cdef / 3 then begin
                  wbol = where(finite(indiv_scan(wnom, kb)) gt 0, cbol)
                  if cbol gt cdef / 3 then dev_bolos(subs, wb(0)+kb) = $
                     total(abs(indiv_scan(wnom(wbol), kb) - mean(indiv_scan(wnom(wbol), kb)))) / float(cbol)
               endif

               if loop eq 1 then begin
                  wj = where(indiv_diff(*, kb) gt bol_noise_level, cj)
                  while cj ge 1 do begin
                     kj_init = wj(0)
                     kj = kj_init
                     w = where(abs(indiv_scan(kj:kj+off-1, kb) - indiv_scan(kj, kb)) lt 3. * median(maxnoise(*, wb(0)+kb)))
                     kj += max(w)
                     tab0 = indiv_scan(kj-2*naver:kj-naver, kb)
                     tab1 = indiv_scan(kj-naver+1:kj, kb)
                     tab2 = indiv_scan(kj+margin:kj+margin+naver-1, kb)
                     tab3 = indiv_scan(kj+margin+naver:kj+margin+2*naver-1, kb)
                     b0 = median(tab0, /even)
                     b1 = median(tab1, /even)
                     b2 = median(tab2, /even)
                     b3 = median(tab3, /even)
                     delta = min([abs(b2 - b1), abs(b2 - b0)])
                     sig1 = sqrt(mean((tab1 - b1)^2.))
                     sig2 = sqrt(mean((tab2 - b2)^2.))

                     if (delta gt max([bol_noise_level, 3. * sig1, 3. * sig2, 2. * abs(b3 - b2), 2. * abs(b1 - b0)]) $
                      and flag_nza(ind_subscans(subs)+wdef(kj)) gt 0 $
                      and max(abs([b0, b1, b2, b3])) gt max(ref_map(wdef(kj-naver+1:kj+margin+naver-1)))) then begin
                        nmask_tot = 0L
                        delta_resi = 4. * noise_resi
                        while(delta_resi gt 2. * noise_resi and kj+nmask_tot+nmask lt cdef - 1 and nmask_tot lt nmask_max) do begin
                           nmask_tot += nmask
                           tab4 = indiv_scan(kj+nmask_tot:min([kj+nmask_tot+nmask, cdef-1]), kb)
                           b4 = median(tab4, /even)
                           delta_resi = min([abs(b4 - b0), abs(b4 - b1)])
                        endwhile
                        if (delta_resi gt 2. * noise_resi and nmask_tot lt nmask_max) then nmask_tot += nmask
                        nmask_tot += (nmask / 2)

                        ;;; beta-release module: interactive mode enforced to check detections
                        time_2 = systime(/seconds)
                        t_proc += (time_2 - time_1)
                        wset, 0
                        ind1 = max([0, kj - 1500])
                        ind2 = min([cdef - 1, max([kj + 1500, kj+nmask_tot+500])])
                        w = where(finite(indiv_scan(ind1:ind2, kb)) gt 0)
                        otherplot, time_s(ind_subscans(subs)+wdef(ind1+w)), indiv_scan(ind1+w, kb), $
                           xtitle='time (s)', ytitle='brightness (Jy/bol.)', $
                           title='bolometer '+chain(kb)+' of array row ' + chain(kl) + ' in subscan ' + chain(subs) + ' (masked values in red)'
                        otherplot, /over, color=200, time_s, flag_nza * max(indiv_scan(ind1+w, kb))
                        w = where(finite(indiv_scan(kj:min([kj+nmask_tot, cdef-1]), kb)) gt 0)
                        otherplot, /over, color=10, time_s(ind_subscans(subs)+wdef(kj+w)), indiv_scan(kj+w, kb)
                        res = dialog_message('Confirm and mask discontinuity ?', /center, /question)
                        time_1 = systime(/seconds)

                        if res eq 'Yes' then begin
                           indiv_diff(kj_init:min([kj+nmask_tot, cdef-1-off]), kb) = 0.
                           nm = min([kj+nmask_tot, cdef-1]) - kj + 1
                           if nm gt 0.75 * cdef then signal(ind_subscans(subs):ind_subscans(subs+1)-1, wb(0)+kb) = !values.f_nan $
                              else signal(ind_subscans(subs)+wdef(kj):ind_subscans(subs)+wdef(min([kj+nmask_tot, cdef-1])), wb(0)+kb) = !values.f_nan
                           njumps_indiv += 1
                        endif
                     endif
                     indiv_diff(kj_init:kj+off, kb) = 0.
                     wj = where(indiv_diff(*, kb) gt bol_noise_level, cj)
                  endwhile
               endif
            endfor

         endif
      endfor
   endfor



   if loop eq 0 then begin
      normdev_lines = fltarr(nsubscans, nlines)
      for s = 0, nscans - 1 do begin
         wsubs = where(ind_subscans ge ind_scans(s) and ind_subscans lt ind_scans(s+1), csubs)
         for k = 0L, nlines - 1L do normdev_lines(wsubs, k) = dev_lines(wsubs, k) / median(maxnoise(s, k*nb_line+lindgen(nb_line)), /even)
      endfor
      med_dev = total(normdev_lines, 1, /nan) / total(finite(normdev_lines), 1)

      if keyword_set(debug) then begin
         wset, 0
         w = where(finite(normdev_lines(0, *)) gt 0.)
         otherplot, w, normdev_lines(0, w), psym=4, yrange=[0., 2.]
         for subs = 1, nsubscans - 1 do begin
            w = where(finite(normdev_lines(subs, *)) gt 0)
            otherplot, /over, color=subs*10, w, normdev_lines(subs, w), psym=4
         endfor
         w = where(finite(med_dev) gt 0.)
         otherplot, /over, psym=4, symbsize=5., w, med_dev(w)
      endif

      if nlines gt 100 then begin
         nx_mat = 4
         nmat = 8
      endif else begin
         nx_mat = 2
         nmat = 2
      endelse
      tab_lines = lindgen(nlines)
      ndrop_row = 0
      w = where(finite(med_dev) gt 0.)
      m_all = median(med_dev(w), /even)
      sig_all = sqrt(mean((med_dev(w) - m_all)^2.))
      for k = 0, nmat - 1 do begin
         wmat = where((tab_lines mod nx_mat) eq (k mod nx_mat) and (tab_lines / nx_mat / nb_line) eq (k / nx_mat) $
            and finite(med_dev) gt 0.)
         m_lines = median(med_dev(wmat), /even)
         sig_lines = sqrt(mean((med_dev(wmat) - m_lines)^2.))
         w = where(med_dev(wmat) ge max([m_lines + 2.5 * sig_lines, m_all + 2. * sig_all]), cw)
         ndrop_row += cw
         for j = 0, cw - 1 do begin
            row = wmat(w(j))
            signal(*, row*nb_line+lindgen(nb_line)) = !values.f_nan
            dev_bolos(*, row*nb_line+lindgen(nb_line)) = !values.f_nan
            if keyword_set(debug) then print, 'dropping row ' + chain(row)
         endfor
      endfor
      if keyword_set(debug) then res = dialog_message('   Continue ?   ', /center, /information, title='Pause')


      normdev_bolos = fltarr(nsubscans, nb)
      for s = 0, nscans - 1 do begin
         wsubs = where(ind_subscans ge ind_scans(s) and ind_subscans lt ind_scans(s+1), csubs)
         normdev_bolos(wsubs, *) = dev_bolos(wsubs, *) / rebin(maxnoise(s, *), csubs, nb)
      endfor
      med_dev = total(normdev_bolos, 1, /nan) / total(finite(normdev_bolos), 1)

      if keyword_set(debug) then begin
         wset, 0
         w = where(finite(normdev_bolos(0, *)) gt 0.)
         otherplot, w, normdev_bolos(0, w), psym=4, yrange=[0., 5.5]
         for subs = 1, nsubscans - 1 do begin
            w = where(finite(normdev_bolos(subs, *)) gt 0.)
            otherplot, /over, color=subs*10, w, normdev_bolos(subs, w), psym=4
         endfor
         w = where(finite(med_dev) gt 0.)
         otherplot, /over, psym=4, symbsize=5., w, med_dev(w)
      endif

      w = where(med_dev gt 3., ndrop)
      if ndrop ge 1 then begin
         signal(*, w) = !values.f_nan
         if keyword_set(debug) then begin
            print, 'dropping bolometer(s): '
            print, w
         endif
      endif
      if keyword_set(debug) then res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   endif

endfor



if (njumps_aver gt 0 or njumps_indiv gt 0 or ndrop_row gt 0 or ndrop gt 0) then begin
   print, crs + 'masked ' + chain(njumps_aver) + ' brightness jump(s) affecting whole array row(s)'
   print, crs + 'masked ' + chain(njumps_indiv) + ' brightness jump(s) affecting single bolometer(s)'
   if ndrop_row gt 0 then print, crs + 'masked entirely ' + chain(ndrop_row) + ' unstable row(s)'
   if ndrop gt 0 then print, crs + 'masked entirely ' + chain(ndrop) + ' unstable bolometer(s)'
   save, file=file_signal, signal, /xdr
   prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans
   prepare_proj_coarse, file_scalars=file_scalars
endif

end
