pro disp_ima, map_disp, win=win, title=title, weightmap=weightmap, trunc_max=trunc_max, $
              min_map=min_map, max_map=max_map, noise_level=noise_level, color_scale=color_scale, $
              col_tab=col_tab, pow_int=pow_int, small=small


;;; displays a map with brightness ticks, guessing at reasonable brightness range and scale

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



if (not keyword_set(col_tab)) then begin
   col_tab = 50
   load_mapcol
endif else if col_tab ne 15 then loadct, col_tab

map = map_disp
w = where(finite(map) lt 1, cw, complement=wdef)
if keyword_set(weightmap) then w = where(finite(map) lt 1 or weightmap le 0., cw, complement=wdef)
if cw ge 1 then if (not keyword_set(min_map)) then map(w) = min(map(wdef), /nan) else map(w) = min_map

if not keyword_set(title) then title = ''
if not keyword_set(noise_level) then begin
   cw = 0
   if keyword_set(weightmap) then w = where(weightmap gt 0. and map ne 0., cw)
   if cw gt 0.1 * n_elements(map) then begin
      stat_map, med, sig, map=map, weightmap=weightmap, noise_level=0.03
      noise_level = sig
   endif else begin
      w = where(finite(map) gt 0 and map ne 0., cw)
      if cw ge 9 then begin
         med = median(map(w))
         noise_level = sqrt(mean((map(w) - med)^2.))
      endif else noise_level = 1.
   endelse
endif


nx = n_elements(map(*, 0))
ny = n_elements(map(0, *))
if keyword_set(small) then fact_zoom_guess = min([1000. / nx, 600. / ny]) $
   else fact_zoom_guess = min([1400. / nx, 800. / ny])

n = 0.
repeat begin
   n += 1.
   fact_zoom = floor(fact_zoom_guess * n) / n
endrep until fact_zoom gt 0.
new_nx = round(floor(nx * fact_zoom) / fact_zoom)
new_ny = round(floor(ny * fact_zoom) / fact_zoom)
xs = round(new_nx * fact_zoom)
ys = round(new_ny * fact_zoom)
if fact_zoom ge 1. then sample = 1 else sample = 0

if keyword_set(weightmap) then begin
   w = where(weightmap gt 0.)
   med = median(weightmap(w))
   w = where(weightmap gt med / 2. and map ne 0., cw)
   if cw eq 0 then w = where(weightmap gt med / 2., cw)
   med = median(map(w))
   sig = 3. * sqrt(median((map(w) - med)^2.))
   tab = map(w)
   tab = tab(sort(tab))
   if (not keyword_set(min_map)) then begin
      if cw ge 30 then min_map = max([tab(cw/20) - sig, tab(5)]) else min_map = min(map, /nan)
   endif
   if (not keyword_set(max_map)) then begin
      if cw ge 30 then begin
         max_map = tab(cw-4) + sig
         if (max_map - min_map) / noise_level gt 100. then max_map = tab(cw-17) + sig
      endif else max_map = max(map, /nan)
      if keyword_set(trunc_max) then max_map = max([tab(95*cw/100), noise_level / 2.])
   endif
endif else begin
   if ((not keyword_set(min_map)) or (not keyword_set(max_map))) then begin
      w = where(map ne 0., cw)
      if (cw eq 0 or min([nx, ny]) lt 100) then w = where(finite(map) gt 0, cw)
   endif
   if (not keyword_set(min_map)) then min_map = min(map(w))
   if (not keyword_set(max_map)) then max_map = max(map(w))
endelse
w = 0

min_map_orig = min_map
max_map_orig = max_map


if keyword_set(pow_int) then begin
   off_map = min(map, /nan)
   index_map = 1. / pow_int
   map = (map - off_map)^pow_int
   min_map = max([min_map - off_map, 0.])^pow_int
   max_map = (max_map - off_map)^pow_int
endif else if (keyword_set(weightmap) and (max_map - min_map) / noise_level gt 200.) then begin
   off_map = min(map, /nan)
   index_map = 10.
   map = (map - off_map)^0.1
   min_map = max([min_map - off_map, 0.])^0.1
   max_map = (max_map - off_map)^0.1
endif else if (keyword_set(weightmap) and (max_map - min_map) / noise_level gt 60.) then begin
   off_map = min(map, /nan)
   index_map = 4.
   map = (map - off_map)^0.25
   min_map = max([min_map - off_map, 0.])^0.25
   max_map = (max_map - off_map)^0.25
endif else if (keyword_set(weightmap) and (max_map - min_map) / noise_level gt 20.) then begin
   off_map = min(map, /nan)
   index_map = 2.
   map = sqrt(map - off_map)
   min_map = sqrt(max([min_map - off_map, 0.]))
   max_map = sqrt(max_map - off_map)
endif else begin
   off_map = 0.
   index_map = 1.
endelse


xs_scale = xs + 100
ys_scale = max([ys, 400])

map_scale = fltarr(xs_scale, ys_scale) + max_map
map_scale(0:xs-1, 0:ys-1) = rebin(map(0:new_nx-1, 0:new_ny-1), xs, ys, sample=sample)
map_scale(xs+5:xs+24, *) = (min_map + (max_map - min_map) * findgen(ys_scale) / (ys_scale - 1.)) ## (1. + fltarr(20))
map_scale(0, 0) = min_map
map_scale(0, 1) = max_map

window, win, xs=xs_scale, ys=ys_scale, title=title
tvscl, map_scale >min_map <max_map
if not keyword_set(color_scale) then color_scale = 0.
step_map = 2. * (max_map - min_map) / 8.
pow = 0
if step_map gt 1.e-5 then begin
   while round(step_map) lt 1 do begin
      pow += 1
      step_map *= 10.
   endwhile
   step = round(step_map) / 2.
   if step gt 2. then step_map = float(round(step_map / 2.)) else step_map = float(step)
   step_map /= 10.^pow
endif else step_map = 1.

for k = ceil(min_map / step_map), floor(max_map / step_map) do xyouts, color=color_scale, $
   /device, [xs+30.], [(k*step_map - min_map) / (max_map - min_map) * ys_scale], $
   chain(off_map+(k*step_map)^index_map, pow+2), charsize=1.5, charthick=2.


if col_tab ne 15 then loadct, 15, /silent
min_map = min_map_orig
max_map = max_map_orig


end
