pro gnomonic, xra_tan, ydec_tan, refx=refx, refy=refy


;;; computes map coordinates using the gnomonic projection type
;;; ref: http://lambda.gsfc.nasa.gov/product/iras/coordproj.cfm
;;;      http://mathworld.wolfram.com/GnomonicProjection.html

;;; convention xra = -RA, ydec = DEC

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



if (not keyword_set(refx)) then refx = median(xra_tan)
if (not keyword_set(refy)) then refy = median(ydec_tan)

rad_deg = !dpi / 180.D0
cos_dec = cos(ydec_tan * rad_deg)
sin_dec = sin(ydec_tan * rad_deg)
cos_mid_dec = cos(refy * rad_deg)
sin_mid_dec = sin(refy * rad_deg)

a = cos_dec * cos((xra_tan - refx) * rad_deg)
f = rad_deg * (sin_mid_dec * sin_dec + a * cos_mid_dec)
ydec_tan = (cos_mid_dec * sin_dec - a * sin_mid_dec) / f
xra_tan = cos_dec * sin((xra_tan - refx) * rad_deg) / f


end
