pro ingest_data, file_param_blocks, ind_scans, ind_subscans, nsubscans, vscan_as_s, timestep, $
                 narr=narr, nscans=nscans, nt=nt, nblocks=nblocks, block=block, $
                 inst=inst, nbolo=nbolo, boloref=boloref, ind_arrays_eff=ind_arrays_eff, $
                 arrays_default=arrays_default, xdeg_blocks=xdeg_blocks, ydeg_blocks=ydeg_blocks, $
                 orient=orient, timegap=timegap, dir_in=dir_in, files_scans=files_scans, $
                 hdr_ref=hdr_ref, file_XYscan=file_XYscan, file_times=file_times, file_vscan=file_vscan, $
                 root_signal=root_signal, root_xyas=root_xyas, maxtimediff=maxtimediff, $
                 visu=visu, vis_traject=vis_traject, frame_fromallscans=frame_fromallscans, $
                 galactic=galactic, minimap=minimap, offset_ra_as=offset_ra_as, offset_dec_as=offset_dec_as


;;; creates signal, coordinates, scan speed and non-zero acceleration flag arrays
;;; for a given block, and all supplied detector arrays
;;; computes time-related quantities

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



crs = string(13b) + string(10b)

restore, file_param_blocks
restore, dir_in + files_scans(0)


if nblocks gt 1 then begin
   print, crs + 'building block ' + chain(block+1) + ' / ' + chain(nblocks)
   print, crs

   restore, file_XYscan
   wtblock = where(Xra_scan ge xdeg_blocks(block, 0) and Xra_scan le xdeg_blocks(block, 1) $
         and Ydec_scan ge ydeg_blocks(block, 0) and Ydec_scan le ydeg_blocks(block, 1), nt)

endif else begin
   nt = n_elements(scan.time_s)
   tags = tag_names(scan)
   w = where(ind_arrays_eff eq 1)
   ind = w(0)
   pos = stregex(tags, 'Xra_bolos_' + arrays_default(ind), /fold_case)
   w = where(pos ge 0)
   Xra = scan.(w(0))(*, boloref(ind))
   pos = stregex(tags, 'Ydec_bolos_' + arrays_default(ind), /fold_case)
   w = where(pos ge 0)
   Ydec = scan.(w(0))(*, boloref(ind))

   gnomonic, Xra, Ydec, refx=mid_Xra, refy=mid_Ydec
   if keyword_set(offset_ra_as) then begin
      Xra -= (offset_ra_as(0) / 3600.D0)
      Ydec += (offset_dec_as(0) / 3600.D0)
   endif


   ;;; attempt to eliminate initial slew:
   ntoff = min([100, nt / 20])
   dx = Xra(ntoff:nt-1) - Xra(0:nt-1-ntoff)
   dy = Ydec(ntoff:nt-1) - Ydec(0:nt-1-ntoff)
   vect_scanrot_deg = atan(dy, dx) * 180.D0 / !dpi
   dx = 0
   dy = 0
   med = (min(vect_scanrot_deg) + max(vect_scanrot_deg)) / 2.
   if nt gt 12 * 100 then begin
      if median(vect_scanrot_deg(100:6*100)) gt med then w = where(vect_scanrot_deg gt med) $
         else w = where(vect_scanrot_deg lt med)
   endif else begin
      if median(vect_scanrot_deg(0:nt/2)) gt med then w = where(vect_scanrot_deg gt med) $
         else w = where(vect_scanrot_deg lt med)
   endelse
   histo = histogram(vect_scanrot_deg(w), locations=ang, binsize=1.)
   m = max(histo, ind)
   w = where(abs(vect_scanrot_deg - ang(ind)) le 10.)
   scanrot_deg_cross = (median(vect_scanrot_deg(w)) mod 180.D0)

   vect_diff = (abs(vect_scanrot_deg - scanrot_deg_cross) mod 180.)
   w = where(vect_diff le 10. or vect_diff ge 170.)
   if w(0) gt 0 then w_exc_cross = lindgen(w(0)) else w_exc_cross = 0
   w = 0

   if keyword_set(w_exc_cross) then Xra(w_exc_cross) = !values.f_nan
   wtblock = where(finite(Xra) gt 0, nt)
endelse


;;; time vector in seconds:
time_s = scan.time_s(wtblock) - scan.time_s(wtblock(0))

if !version.memory_bits ge 64 then nt = long64(nt)



;;; create pointers to signals for all arrays (as dblarr(nt, nb)):

signals = ptrarr(narr, /allocate_heap)

tags = tag_names(scan)
for detarr = 0, narr - 1 do if ind_arrays_eff(detarr) eq 1 then begin
   pos = stregex(tags, 'signal_' + arrays_default(detarr), /fold_case)
   w = where(pos ge 0)
   *signals(detarr) = float(scan.(w(0))(wtblock, 0:nbolo(detarr)-1))
endif



;;; map coordinates in arcsec:

cos_scanrot = cos(scanrot_deg * !dpi / 180.D0)
sin_scanrot = sin(scanrot_deg * !dpi / 180.D0)

xas_bolos = ptrarr(narr, /allocate_heap)
yas_bolos = ptrarr(narr, /allocate_heap)

if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
   p_cos_scanrot = cos(p_scanrot_deg * !dpi / 180.D0)
   p_sin_scanrot = sin(p_scanrot_deg * !dpi / 180.D0)
   p_xas_bolos = ptrarr(narr, /allocate_heap)
   p_yas_bolos = ptrarr(narr, /allocate_heap)
endif


for detarr = 0, narr - 1 do if ind_arrays_eff(detarr) eq 1 then begin
   pos = stregex(tags, 'Xra_bolos_' + arrays_default(detarr), /fold_case)
   w = where(pos ge 0)
   Xra = scan.(w(0))(wtblock, 0:nbolo(detarr)-1)
   pos = stregex(tags, 'Ydec_bolos_' + arrays_default(detarr), /fold_case)
   w = where(pos ge 0)
   Ydec = scan.(w(0))(wtblock, 0:nbolo(detarr)-1)

   gnomonic, Xra, Ydec, refx=mid_Xra, refy=mid_Ydec

   Xra *= 3600.D0
   Ydec *= 3600.D0
   if keyword_set(offset_ra_as) then begin
      Xra -= offset_ra_as(0)
      Ydec += offset_dec_as(0)
   endif

   if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
      *p_xas_bolos(detarr) = Xra * p_cos_scanrot + Ydec * p_sin_scanrot
      *p_yas_bolos(detarr) = Ydec * p_cos_scanrot - Xra * p_sin_scanrot
   endif

   case orient of
      'astro': begin
         *xas_bolos(detarr) = Xra
         *yas_bolos(detarr) = Ydec
         end
      'scan': begin
         *xas_bolos(detarr) = Xra * cos_scanrot + Ydec * sin_scanrot
         *yas_bolos(detarr) = Ydec * cos_scanrot - Xra * sin_scanrot
         end
   endcase
endif



;;; read cross-scan data:

ind_scans = [0L, nt]
for i = 1, nscans - 1 do begin
   restore, dir_in + files_scans(i)
   nt_b = n_elements(scan.time_s)
   tags = tag_names(scan)

   w = where(ind_arrays_eff eq 1)
   ind = w(0)
   pos = stregex(tags, 'Xra_bolos_' + arrays_default(ind), /fold_case)
   w = where(pos ge 0)
   Xra = scan.(w(0))(*, boloref(ind))
   pos = stregex(tags, 'Ydec_bolos_' + arrays_default(ind), /fold_case)
   w = where(pos ge 0)
   Ydec = scan.(w(0))(*, boloref(ind))

   gnomonic, Xra, Ydec, refx=mid_Xra, refy=mid_Ydec
   if keyword_set(offset_ra_as) then begin
      Xra -= (offset_ra_as(i) / 3600.D0)
      Ydec += (offset_dec_as(i) / 3600.D0)
   endif


   ;;; attempt to eliminate initial slew:
   ntoff = min([100, nt_b / 20])
   dx = Xra(ntoff:nt_b-1) - Xra(0:nt_b-1-ntoff)
   dy = Ydec(ntoff:nt_b-1) - Ydec(0:nt_b-1-ntoff)
   vect_scanrot_deg = atan(dy, dx) * 180.D0 / !dpi
   dx = 0
   dy = 0
   med = (min(vect_scanrot_deg) + max(vect_scanrot_deg)) / 2.
   if nt_b gt 12 * 100 then begin
      if median(vect_scanrot_deg(100:6*100)) gt med then w = where(vect_scanrot_deg gt med) $
         else w = where(vect_scanrot_deg lt med)
   endif else begin
      if median(vect_scanrot_deg(0:nt_b/2)) gt med then w = where(vect_scanrot_deg gt med) $
         else w = where(vect_scanrot_deg lt med)
   endelse
   histo = histogram(vect_scanrot_deg(w), locations=ang, binsize=1.)
   m = max(histo, ind)
   w = where(abs(vect_scanrot_deg - ang(ind)) le 10.)
   scanrot_deg_cross = (median(vect_scanrot_deg(w)) mod 180.D0)

   vect_diff = (abs(vect_scanrot_deg - scanrot_deg_cross) mod 180.)
   w = where(vect_diff le 10. or vect_diff ge 170.)
   if w(0) gt 0 then w_exc_cross = lindgen(w(0)) else w_exc_cross = 0
   w = 0

   if keyword_set(w_exc_cross) then Xra(w_exc_cross) = !values.f_nan


   if nblocks gt 1 then begin
      Xra_rot = Xra * cos_scanrot + Ydec * sin_scanrot + off_cen_Xra
      Ydec_rot = Ydec * cos_scanrot - Xra * sin_scanrot + off_cen_Ydec
      wtblock = where(Xra_rot ge xdeg_blocks(block, 0) and Xra_rot le xdeg_blocks(block, 1) $
         and Ydec_rot ge ydeg_blocks(block, 0) and Ydec_rot le ydeg_blocks(block, 1), nt_b)
   endif else wtblock = where(finite(Xra) gt 0, nt_b)

   time_b = scan.time_s(wtblock) - scan.time_s(wtblock(0))
   time_s = [time_s, max(time_s) + timegap + time_b]
   nt = nt + nt_b
   ind_scans = [ind_scans, nt]


   for detarr = 0, narr - 1 do if ind_arrays_eff(detarr) eq 1 then begin
      pos = stregex(tags, 'Xra_bolos_' + arrays_default(detarr), /fold_case)
      w = where(pos ge 0)
      Xra = (scan.(w(0)))(wtblock, 0:nbolo(detarr)-1)
      pos = stregex(tags, 'Ydec_bolos_' + arrays_default(detarr), /fold_case)
      w = where(pos ge 0)
      Ydec = (scan.(w(0)))(wtblock, 0:nbolo(detarr)-1)

      gnomonic, Xra, Ydec, refx=mid_Xra, refy=mid_Ydec

      Xra *= 3600.D0
      Ydec *= 3600.D0
      if keyword_set(offset_ra_as) then begin
         Xra -= offset_ra_as(i)
         Ydec += offset_dec_as(i)
      endif

      pos = stregex(tags, 'signal_' + arrays_default(detarr), /fold_case)
      w = where(pos ge 0)
      *signals(detarr) = [*signals(detarr), float(scan.(w(0))(wtblock, 0:nbolo(detarr)-1))]

      if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
         *p_xas_bolos(detarr) = [*p_xas_bolos(detarr), Xra * p_cos_scanrot + Ydec * p_sin_scanrot]
         *p_yas_bolos(detarr) = [*p_yas_bolos(detarr), Ydec * p_cos_scanrot - Xra * p_sin_scanrot]
      endif

      case orient of
         'astro': begin
            *xas_bolos(detarr) = [*xas_bolos(detarr), Xra]
            *yas_bolos(detarr) = [*yas_bolos(detarr), Ydec]
            end
         'scan': begin
            *xas_bolos(detarr) = [*xas_bolos(detarr), Xra * cos_scanrot + Ydec * sin_scanrot]
            *yas_bolos(detarr) = [*yas_bolos(detarr), Ydec * cos_scanrot - Xra * sin_scanrot]
            end
      endcase
   endif
endfor

for detarr = 0, narr - 1 do if ind_arrays_eff(detarr) eq 1 then begin
   if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
      *p_xas_bolos(detarr) += (p_off_cen_Xra * 3600.D0)
      *p_yas_bolos(detarr) += (p_off_cen_Ydec * 3600.D0)
   endif
   *xas_bolos(detarr) += (off_cen_Xra * 3600.D0)
   *yas_bolos(detarr) += (off_cen_Ydec * 3600.D0)
endif



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; save signal and coordinates to disk
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;; mask PACS blue-array unstable rows/bolometers and red-array cross-talk columns:

if inst eq 'pacs' then begin
   nb_line = 16L
   for detarr = 0, 1 do if ind_arrays_eff(detarr) eq 1 then begin
      if keyword_set(galactic) then rows = [28L, 36L, 45L, 75L, 124L] $
         else rows = [36L, 45L, 75L]
      nrows = n_elements(rows)
      for k = 0, nrows - 1 do (*signals(detarr))(*, rows(k)*nb_line+lindgen(nb_line)) = !values.f_nan
      columns = [0L, 16L, 32L, 48L]
      for k = 0, 3 do (*signals(detarr))(*, columns(k)+4L*nb_line*lindgen(2L*nb_line)) = !values.f_nan
   endif
   if ind_arrays_eff(2) eq 1 then begin
      columns = [0L, 16L]
      for k = 0, 1 do (*signals(detarr))(*, columns(k)+2L*nb_line*lindgen(nb_line)) = !values.f_nan
      bols = [5L+0L*nb_line]
      (*signals(detarr))(*, bols) = !values.f_nan
   endif
endif

for detarr = 0, narr - 1 do if ind_arrays_eff(detarr) eq 1 then begin
   file_signal = root_signal + arrays_default(detarr) + '.xdr'
   signal = *signals(detarr)
   save, file=file_signal, signal, /xdr
   signal = 0
   file_xyas = root_xyas + arrays_default(detarr) + '.xdr'
   file_xyas_final = root_xyas + 'final_' + arrays_default(detarr) + '.xdr'
   xas = *xas_bolos(detarr)
   yas = *yas_bolos(detarr)
   save, file=file_xyas_final, xas, yas, /xdr
   if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
      xas = *p_xas_bolos(detarr)
      yas = *p_yas_bolos(detarr)
   endif
   save, file=file_xyas, xas, yas, /xdr
   xas = 0
   yas = 0
endif

if (orient eq 'astro' or keyword_set(hdr_ref)) then ptr_free, p_xas_bolos, p_yas_bolos



;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; time-related quantities
;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;; instantaneous scan speed:

w = where(ind_arrays_eff eq 1)
detarr = w(0)
for i = 0, nscans - 1 do begin
   dd = sqrt(((*xas_bolos(detarr))(ind_scans(i)+1:ind_scans(i+1)-1, boloref(detarr)) $
      - (*xas_bolos(detarr))(ind_scans(i):ind_scans(i+1)-2, boloref(detarr)))^2. + $
      ((*yas_bolos(detarr))(ind_scans(i)+1:ind_scans(i+1)-1, boloref(detarr)) $
      - (*yas_bolos(detarr))(ind_scans(i):ind_scans(i+1)-2, boloref(detarr)))^2.)
   dt = time_s(ind_scans(i)+1:ind_scans(i+1)-1) - time_s(ind_scans(i):ind_scans(i+1)-2)
   if i eq 0 then vscan = [0., float(dd / dt)] else vscan = [vscan, 0., float(dd / dt)]
endfor
dd = 0
dt = 0

save, file=file_vscan, vscan, /xdr



;;; visualization of all bolometer trajectories:

if keyword_set(vis_traject) then begin
   w = where(ind_arrays_eff eq 1)
   detarr = w(0)
   window, 0, xs=800, ys=800
   m = max([max(*xas_bolos(detarr)), max(*yas_bolos(detarr))])
   if nblocks gt 1 then m = (max([max_xdeg, max_ydeg]) + drdeg_max(detarr) / 2.) * 3600.
   otherplot, (*xas_bolos(detarr))(0, *), (*yas_bolos(detarr))(0, *), psym=4, charsize=1.5, $
      xrange=[0,m], yrange=[0,m], xtitle='x coordinate (arcsec)', ytitle='y coordinate (arcsec)'
   if nt lt 100000L then dnt = 10L else dnt = 50L * min([1L, (1L + nt / 200000L)])
   for k = 0L, nscans - 1 do begin
      for i = 0L, (ind_scans(k+1) - ind_scans(k) - 1) / dnt do begin
         t = ind_scans(k) + dnt * i
         otherplot, /over, color=i*5, (*xas_bolos(detarr))(t, *), (*yas_bolos(detarr))(t, *), psym=4
      endfor
      wait, 0.5
   endfor
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   window, 0, xs=1200
endif

ptr_free, xas_bolos, yas_bolos



;;; median scan speed:

histo = histogram(vscan, locations=v, binsize=0.1, min=5.)
w = where(histo gt max(histo) / 5.)
v = v(w)
w = where(vscan ge min(v) and vscan le max(v))
vscan_as_s = median(vscan(w))


;;; flag vector for scan turnarounds (= 0 when non-zero acceleration, = 1 otherwise):

if keyword_set(minimap) then ntoff = 50 else ntoff = 100
ntoff = min([ntoff, round(median(ind_scans(1:nscans) - ind_scans(0:nscans-1), /even)) / 20L])
flag_nza = 1 + intarr(nt)
w = where(abs(vscan / vscan_as_s - 1.) gt 0.2, cw)
if cw ge 1 then flag_nza(w) = 0

wup = where([0, flag_nza(1:*) - flag_nza(0:nt-2)] eq 1, cup)
for i = 0L, cup - 1 do begin
   wdown = where(flag_nza(wup(i):*) - flag_nza(wup(i)-1:nt-2) eq (-1), cdown)
   if (cdown ge 1 and wdown(0) lt ntoff) then flag_nza(wup(i):wup(i)+wdown(0)) = 0
endfor

if ntoff ge 100 then begin
   wup = where([0, flag_nza(1:*) - flag_nza(0:nt-2)] eq 1, cup)
   for i = 0L, cup - 1 do begin
      wdown = where(flag_nza(wup(i):*) - flag_nza(wup(i)-1:nt-2) eq (-1), cdown)
      if (cdown ge 1 and wdown(0) + wup(i) lt nt - 5) then if flag_nza(wup(i)+wdown(0)+5) gt 0 $
         then flag_nza(wup(i)+wdown(0):wup(i)+wdown(0)+4) = 1
   endfor
endif
wup = 0
wdown = 0

save, file=file_times, time_s, flag_nza, /xdr



;;; nominal scan speed in arcsec/s and time step in s:

w = where(flag_nza eq 1, cw)
vscan_as_s = median(vscan(w))
print, 'nominal scan speed = ' + chain(vscan_as_s, 2) + ' arcsec/s'

diff_t = time_s(w(1:cw-1)) - time_s(w(0:cw-2))
w = 0
diff_t = diff_t(sort(diff_t))
timestep = float(median(diff_t(cw/10:9*cw/10-1)))
diff_t = 0



;;; subscan indices:

flag_drop = intarr(nt)

ind_subscans = [0L]
for s = 0, nscans - 1 do begin
   it0 = ind_scans(s)
   repeat begin
      w = where(flag_nza(it0:ind_scans(s+1)-1) eq 1)
      it0 += max([w(0), 0])
      it1 = it0 + 1
      if it1 lt ind_scans(s+1) - 1 then w = where(flag_nza(it1:ind_scans(s+1)-1) eq 0 $
         or [time_s(it1+1:ind_scans(s+1)-1) - time_s(it1:ind_scans(s+1)-2), 0.] gt maxtimediff, cw) $
       else cw = 0
      if cw eq 0 then it1 = ind_scans(s+1) - 1 else begin
         it1 += w(0)
         if flag_nza(it1) eq 0 then begin
            w = where(flag_nza(it1:ind_scans(s+1)-1) eq 1, cw)
            if cw eq 0 then it1 = ind_scans(s+1) - 1 else begin
               if inst eq 'spire' then it1 += (w(0) / 2) else begin
                  wls = where(vscan(it1:it1+w(0)) lt vscan_as_s / 4.)
                  it1 += ((min(wls) + max(wls)) / 2)
               endelse
            endelse
         endif
      endelse
      if (it1 ge ind_scans(s+1) - 31 and it1 lt ind_scans(s+1) - 1) then begin
         flag_drop(it1+1:ind_scans(s+1)-1) = 1
         it1 = ind_scans(s+1) - 1
      endif
      it0 = it1 + 1
      ind_subscans = [ind_subscans, it0]
   endrep until it0 ge ind_scans(s+1)
endfor
nsubscans = n_elements(ind_subscans) - 1



;;; save signal again after masking very low or very high speeds:

for s = 0L, nscans - 1L do begin
   wnom = where(flag_nza(ind_scans(s):ind_scans(s+1)-1) eq 1)
   wini = max(wnom) + 1
   if ind_scans(s) + wini lt ind_scans(s+1) - 10 then begin
      do_stop = 0
      repeat begin
         wstart = where(vscan(ind_scans(s)+wini:ind_scans(s+1)-1) lt 3., cw)
         if cw ge 1 then begin
            wstart = wini + wstart(0)
            wend = where(vscan(ind_scans(s)+wstart:ind_scans(s+1)-1) ge 3., ce)
            if (ce lt 1 or wend(0) ge 100) then begin
               flag_drop(ind_scans(s)+wstart:ind_scans(s+1)-1) = 1
               do_stop = 1
            endif else wini = wstart + wend(0)
         endif else do_stop = 1
      endrep until do_stop eq 1
   endif
endfor
w = where(vscan lt max([3., 0.2 * vscan_as_s]) or vscan gt 1.5 * vscan_as_s or flag_drop gt 0, cw)
if cw ge 1 then for detarr = 0, narr - 1 do if ind_arrays_eff(detarr) eq 1 then begin
   file_signal = root_signal + arrays_default(detarr) + '.xdr'
   signal = *signals(detarr)
   signal(w, *) = !values.f_nan
   save, file=file_signal, signal, /xdr
   signal = 0
endif

ptr_free, signals



if keyword_set(visu) then begin
   window, 0, xs=1200
   scan = 0
   otherplot, time_s(ind_scans(scan):ind_scans(scan+1)-1), vscan(ind_scans(scan):ind_scans(scan+1)-1), $
      xtitle='t (s)', ytitle='V!lscan!n (arcsec/s)', $
      title='scan speed of scan '+chain(scan), yrange=[0.8, 1.25]*vscan_as_s
   otherplot, /over, time_s(ind_scans(scan):ind_scans(scan+1)-1), $
      flag_nza(ind_scans(scan):ind_scans(scan+1)-1) * 1.2*vscan_as_s, color=200.
   if nsubscans le 40 then for subs = 0L, nsubscans - 1 do $
      otherplot, /over, time_s(ind_subscans(subs))+[0.,0.], [0., 1.25*vscan_as_s], color=30, lstyle=2.
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
endif


end
