pro prepare_blocks, file_param_blocks, ind_arrays_eff, orient, nt, nblocks, xdeg_blocks, ydeg_blocks, $
                    narr=narr, nscans=nscans, nbolo=nbolo, boloref=boloref, arrays_default=arrays_default, $
                    max_n_samples=max_n_samples, ind_arrays=ind_arrays, debug=debug, $
                    dir_in=dir_in, files_scans_frame=files_scans_frame, hdr_ref=hdr_ref, $
                    file_XYscan=file_XYscan


;;; prepares coordinates frame
;;; prepares several blocks in case the data volume is too large to be processed all at once

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



crs = string(13b) + string(10b)

nf = n_elements(files_scans_frame)
restore, dir_in + files_scans_frame(0)

nt = n_elements(scan.time_s)
if !version.memory_bits ge 64 then nt = long64(nt)


;;; check whether data have to be sliced into blocks:
;;; (in which case only one detector is processed, with a scan-parallel orientation)

ind_arrays_eff = ind_arrays
w = where(ind_arrays eq 1, cw)
ntb = nt * nbolo(w(0)) * nscans
print, crs + 'nt * nb ~ ' + chain(ntb)

if ((ntb gt max_n_samples or nblocks gt 1) and (not keyword_set(hdr_ref))) then begin
   print, crs + '*******************************************************************************'
   print, crs + 'The ' + arrays_default(w(0)) + ' map has to be sliced into overlapping sub-maps'
   print, ' to keep the processed data volume within reasonable bounds.'
   if orient eq 'astro' then begin
      print, crs + 'Changing the map orientation to scan-parallel.'
      orient = 'scan'
   endif
   if cw gt 1 then begin
      print, crs + 'Only the ' + arrays_default(w(0)) + ' map will be created.'
      print, 'To produce the other map(s) with the same spatial grid,'
      print, 're-run the program with the same parameters, except'
      print, 'with ' + arrays_default(w(0)) + ' deselected and "map orientation"="scan".'
      ind_arrays_eff(w(0)+1:*) = 0
   endif
      print, crs + '*******************************************************************************'
      print, crs
endif



;;; coordinate vectors of all bolometers in degrees (Xra = -RA, Ydec = DEC):

Xra_bolos = ptrarr(narr, /allocate_heap)
Ydec_bolos = ptrarr(narr, /allocate_heap)
mid_Xra = dblarr(narr)
mid_Ydec = dblarr(narr)
off_cen_Xra = dblarr(narr, nf)
off_cen_Ydec = dblarr(narr, nf)
drdeg_max = dblarr(narr)
max_xdeg = dblarr(narr, nf)
max_ydeg = dblarr(narr, nf)
tags = tag_names(scan)
scanrot_deg = 0

if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
   p_Xra_bolos = ptrarr(narr, /allocate_heap)
   p_Ydec_bolos = ptrarr(narr, /allocate_heap)
   p_mid_Xra = dblarr(narr)
   p_mid_Ydec = dblarr(narr)
   p_off_cen_Xra = dblarr(narr, nf)
   p_off_cen_Ydec = dblarr(narr, nf)
   p_max_xdeg = dblarr(narr, nf)
   p_max_ydeg = dblarr(narr, nf)
endif


for detarr = 0, narr - 1 do if ind_arrays(detarr) eq 1 then begin
   pos = stregex(tags, 'Xra_bolos_' + arrays_default(detarr), /fold_case)
   w = where(pos ge 0)
   Xra_bolos_scan = scan.(w(0))
   pos = stregex(tags, 'Ydec_bolos_' + arrays_default(detarr), /fold_case)
   w = where(pos ge 0)
   Ydec_bolos_scan = scan.(w(0))

   mid_Xra(detarr) = median(Xra_bolos_scan)
   mid_Ydec(detarr) = median(Ydec_bolos_scan)
   refx = mid_Xra(detarr)
   refy = mid_Ydec(detarr)

   gnomonic, Xra_bolos_scan, Ydec_bolos_scan, refx=refx, refy=refy
   *Xra_bolos(detarr) = Xra_bolos_scan
   *Ydec_bolos(detarr) = Ydec_bolos_scan


   ;;; scan rotation angle with respect to Xra axis:
   if (not keyword_set(scanrot_deg)) then begin
      ntoff = min([100, nt / 20])
      dx = (*Xra_bolos(detarr))(ntoff:nt-1, boloref(detarr)) - (*Xra_bolos(detarr))(0:nt-1-ntoff, boloref(detarr))
      dy = (*Ydec_bolos(detarr))(ntoff:nt-1, boloref(detarr)) - (*Ydec_bolos(detarr))(0:nt-1-ntoff, boloref(detarr))
      vect_scanrot_deg = atan(dy, dx) * 180.D0 / !dpi
      dx = 0
      dy = 0
      if keyword_set(debug) then begin
         window, 0, xs=1200
         otherplot, lindgen(nt-ntoff), vect_scanrot_deg, xtitle='sample index', $
            ytitle='scanning angle wrt X axis (degrees)'
      endif
      med = (min(vect_scanrot_deg) + max(vect_scanrot_deg)) / 2.
      if nt gt 12 * 100 then begin
         if median(vect_scanrot_deg(100:6*100)) gt med then w = where(vect_scanrot_deg gt med) $
            else w = where(vect_scanrot_deg lt med)
      endif else begin
         if median(vect_scanrot_deg(0:nt/2)) gt med then w = where(vect_scanrot_deg gt med) $
            else w = where(vect_scanrot_deg lt med)
      endelse
      histo = histogram(vect_scanrot_deg(w), locations=ang, binsize=1.)
      m = max(histo, ind)
      w = where(abs(vect_scanrot_deg - ang(ind)) le 10.)
      scanrot_deg = (median(vect_scanrot_deg(w)) mod 180.D0)
      print, 'scan rotation angle wrt EW axis = ' + chain(scanrot_deg, 2) + ' degrees'
      if keyword_set(debug) then begin
         otherplot, /over, color=10, lstyle=2, lindgen(nt-ntoff), scanrot_deg + fltarr(nt-ntoff)
         res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
      endif
      cos_scanrot = cos(scanrot_deg * !dpi / 180.D0)
      sin_scanrot = sin(scanrot_deg * !dpi / 180.D0)

      p_scanrot_deg = scanrot_deg
      p_cos_scanrot = cos_scanrot
      p_sin_scanrot = sin_scanrot
      p_maprot_deg = 0.D0


      ;;; determine map rotation angle with respect to x axis:
      if orient eq 'scan' then maprot_deg = 0.D0 else maprot_deg = scanrot_deg mod 90.D0

      ;;; attempt to eliminate initial slew:
      vect_diff = (abs(vect_scanrot_deg - scanrot_deg) mod 180.)
      w = where(vect_diff le 10. or vect_diff ge 170.)
      if w(0) gt 0 then w_exc = lindgen(w(0)) else w_exc = 0
      w = 0

      ;;; optionally take rotation in reference header:
      if keyword_set(hdr_ref) then begin
         crota2 = sxpar(hdr_ref, 'CROTA2', count=cc)
         if cc lt 1 then begin
            print, 'missing CROTA2 keyword in reference header: assuming 0.'
            crota2 = 0.D0
         endif
         if abs(crota2) lt 1.e-6 then orient = 'astro' else orient = 'scan'
         maprot_deg = (crota2 + scanrot_deg) mod 90.D0

         scanrot_deg = -crota2
         cos_scanrot = cos(scanrot_deg * !dpi / 180.D0)
         sin_scanrot = sin(scanrot_deg * !dpi / 180.D0)
      endif

      print, 'map rotation angle wrt x axis = ' + chain(maprot_deg, 2) + ' degrees'
   endif

   if keyword_set(w_exc) then begin
      (*Xra_bolos(detarr))(w_exc, *) = !values.f_nan
      (*Ydec_bolos(detarr))(w_exc, *) = !values.f_nan
   endif


   if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
      *p_Xra_bolos(detarr) = *Xra_bolos(detarr) * p_cos_scanrot + *Ydec_bolos(detarr) * p_sin_scanrot
      *p_Ydec_bolos(detarr) = *Ydec_bolos(detarr) * p_cos_scanrot - *Xra_bolos(detarr) * p_sin_scanrot
      p_off_cen_Xra(detarr, 0) = -min(*p_Xra_bolos(detarr), /nan)
      p_off_cen_Ydec(detarr, 0) = -min(*p_Ydec_bolos(detarr), /nan)
      p_max_xdeg(detarr, 0) = max(*p_Xra_bolos(detarr), /nan)
      p_max_ydeg(detarr, 0) = max(*p_Ydec_bolos(detarr), /nan)
      *p_Xra_bolos(detarr) += p_off_cen_Xra(detarr, 0)
      *p_Ydec_bolos(detarr) += p_off_cen_Ydec(detarr, 0)
   endif

   case orient of
      'astro': begin
         off_cen_Xra(detarr, 0) = -min(*Xra_bolos(detarr), /nan)
         off_cen_Ydec(detarr, 0) = -min(*Ydec_bolos(detarr), /nan)
         end
      'scan': begin
         Xra_tempo = *Xra_bolos(detarr)
         *Xra_bolos(detarr) = *Xra_bolos(detarr) * cos_scanrot + *Ydec_bolos(detarr) * sin_scanrot
         *Ydec_bolos(detarr) = *Ydec_bolos(detarr) * cos_scanrot - Xra_tempo * sin_scanrot
         Xra_tempo = 0
         off_cen_Xra(detarr, 0) = -min(*Xra_bolos(detarr), /nan)
         off_cen_Ydec(detarr, 0) = -min(*Ydec_bolos(detarr), /nan)
         end
   endcase

   drdeg_max(detarr) = sqrt(max((*Xra_bolos(detarr) $
      - rebin((*Xra_bolos(detarr))(*, boloref(detarr)), nt, nbolo(detarr)))^2. $
      + (*Ydec_bolos(detarr) - rebin((*Ydec_bolos(detarr))(*, boloref(detarr)), nt, nbolo(detarr)))^2., /nan))

   max_xdeg(detarr, 0) = max(*Xra_bolos(detarr), /nan)
   max_ydeg(detarr, 0) = max(*Ydec_bolos(detarr), /nan)

   *Xra_bolos(detarr) += off_cen_Xra(detarr, 0)
   *Ydec_bolos(detarr) += off_cen_Ydec(detarr, 0)
endif

if (orient eq 'astro' or keyword_set(hdr_ref)) then ptr_free, p_Xra_bolos, p_Ydec_bolos



if nscans gt 1 then begin
   w = where(ind_arrays eq 1)
   arr_1 = w(0)
   for kf = 1, n_elements(files_scans_frame) - 1 do begin
      restore, dir_in + files_scans_frame(kf)
      nt_b = n_elements(scan.time_s)

      for detarr = 0, narr - 1 do if ind_arrays(detarr) eq 1 then begin
         pos = stregex(tags, 'Xra_bolos_' + arrays_default(detarr), /fold_case)
         w = where(pos ge 0)
         Xra_bolos_scan = scan.(w(0))
         pos = stregex(tags, 'Ydec_bolos_' + arrays_default(detarr), /fold_case)
         w = where(pos ge 0)
         Ydec_bolos_scan = scan.(w(0))

         refx = mid_Xra(detarr)
         refy = mid_Ydec(detarr)
         gnomonic, Xra_bolos_scan, Ydec_bolos_scan, refx=refx, refy=refy


         ;;; attempt to eliminate initial slew:
         if detarr eq arr_1 then begin
            ntoff = min([100, nt_b / 20])
            dx = Xra_bolos_scan(ntoff:nt_b-1, boloref(detarr)) - Xra_bolos_scan(0:nt_b-1-ntoff, boloref(detarr))
            dy = Ydec_bolos_scan(ntoff:nt_b-1, boloref(detarr)) - Ydec_bolos_scan(0:nt_b-1-ntoff, boloref(detarr))
            vect_scanrot_deg = atan(dy, dx) * 180.D0 / !dpi
            dx = 0
            dy = 0
            med = (min(vect_scanrot_deg) + max(vect_scanrot_deg)) / 2.
            if nt_b gt 12 * 100 then begin
               if median(vect_scanrot_deg(100:6*100)) gt med then w = where(vect_scanrot_deg gt med) $
                  else w = where(vect_scanrot_deg lt med)
            endif else begin
               if median(vect_scanrot_deg(0:nt_b/2)) gt med then w = where(vect_scanrot_deg gt med) $
                  else w = where(vect_scanrot_deg lt med)
            endelse
            histo = histogram(vect_scanrot_deg(w), locations=ang, binsize=1.)
            m = max(histo, ind)
            w = where(abs(vect_scanrot_deg - ang(ind)) le 10.)
            scanrot_deg_cross = (median(vect_scanrot_deg(w)) mod 180.D0)

            vect_diff = (abs(vect_scanrot_deg - scanrot_deg_cross) mod 180.)
            w = where(vect_diff le 10. or vect_diff ge 170.)
            if w(0) gt 0 then w_exc_cross = lindgen(w(0)) else w_exc_cross = 0
            w = 0
         endif

         if keyword_set(w_exc_cross) then begin
            Xra_bolos_scan(w_exc_cross, *) = !values.f_nan
            Ydec_bolos_scan(w_exc_cross, *) = !values.f_nan
         endif


         if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
            p_Xra_bolos_scan = Xra_bolos_scan * p_cos_scanrot + Ydec_bolos_scan * p_sin_scanrot
            p_Ydec_bolos_scan = Ydec_bolos_scan * p_cos_scanrot - Xra_bolos_scan * p_sin_scanrot
            p_off_cen_Xra(detarr, kf) = -min(p_Xra_bolos_scan, /nan)
            p_off_cen_Ydec(detarr, kf) = -min(p_Ydec_bolos_scan, /nan)
            p_max_xdeg(detarr, kf) = max(p_Xra_bolos_scan, /nan)
            p_max_ydeg(detarr, kf) = max(p_Ydec_bolos_scan, /nan)
         endif

         case orient of
            'astro': begin
               off_cen_Xra(detarr, kf) = -min(Xra_bolos_scan, /nan)
               off_cen_Ydec(detarr, kf) = -min(Ydec_bolos_scan, /nan)
               end
           'scan': begin
               Xra_tempo = Xra_bolos_scan
               Xra_bolos_scan = Xra_bolos_scan * cos_scanrot + Ydec_bolos_scan * sin_scanrot
               Ydec_bolos_scan = Ydec_bolos_scan * cos_scanrot - Xra_tempo * sin_scanrot
               Xra_tempo = 0
               off_cen_Xra(detarr, kf) = -min(Xra_bolos_scan, /nan)
               off_cen_Ydec(detarr, kf) = -min(Ydec_bolos_scan, /nan)
               end
         endcase

         drdeg_max(detarr) = max([drdeg_max(detarr), $
            sqrt(max((Xra_bolos_scan - rebin(Xra_bolos_scan(*, boloref(detarr)), nt_b, nbolo(detarr)))^2. $
            + (Ydec_bolos_scan - rebin(Ydec_bolos_scan(*, boloref(detarr)), nt_b, nbolo(detarr)))^2., /nan))])

         max_xdeg(detarr, kf) = max(Xra_bolos_scan, /nan)
         max_ydeg(detarr, kf) = max(Ydec_bolos_scan, /nan)
      endif

   endfor
   Xra_bolos_scan = 0
   Ydec_bolos_scan = 0
   restore, dir_in + files_scans_frame(0)
endif




;;; take into account possible spatial shifts between reference scans:

if nf gt 1 then begin
   off_cen_Xra = max(off_cen_Xra, dimension=2)
   off_cen_Ydec = max(off_cen_Ydec, dimension=2)
   max_xdeg = max(max_xdeg, dimension=2)
   max_ydeg = max(max_ydeg, dimension=2)
endif
max_xdeg += off_cen_Xra
max_ydeg += off_cen_Ydec
if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
   if nf gt 1 then begin
      p_off_cen_Xra = max(p_off_cen_Xra, dimension=2)
      p_off_cen_Ydec = max(p_off_cen_Ydec, dimension=2)
      p_max_xdeg = max(p_max_xdeg, dimension=2)
      p_max_ydeg = max(p_max_ydeg, dimension=2)
   endif
   p_max_xdeg += p_off_cen_Xra
   p_max_ydeg += p_off_cen_Ydec
endif


;;; reference = middle of larger map:

if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
   w = where(ind_arrays eq 1)
   m = max(p_max_xdeg(w)^2. + p_max_ydeg(w)^2., ind)
   w = w(ind)
   p_mid_Xra = mid_Xra(w)
   p_mid_Ydec = mid_Ydec(w)
   p_off_cen_Xra = p_off_cen_Xra(w)
   p_off_cen_Ydec = p_off_cen_Ydec(w)
endif

if keyword_set(hdr_ref) then begin
   crval1 = sxpar(hdr_ref, 'CRVAL1')
   crval2 = sxpar(hdr_ref, 'CRVAL2')
   crpix1 = sxpar(hdr_ref, 'CRPIX1') - 0.5
   crpix2 = sxpar(hdr_ref, 'CRPIX2') - 0.5
   cdelt1 = sxpar(hdr_ref, 'CDELT1', count=cc)
   if cc gt 0 then pix_deg = sqrt(abs(cdelt1) * sxpar(hdr_ref, 'CDELT2')) $
      else pix_deg = sqrt((sxpar(hdr_ref, 'CD1_1'))^2. + (sxpar(hdr_ref, 'CD1_2'))^2.)
   mid_Xra = -crval1
   mid_Ydec = crval2
   off_cen_Xra = crpix1 * pix_deg
   off_cen_Ydec = crpix2 * pix_deg
endif else begin
   w = where(ind_arrays eq 1)
   m = max(max_xdeg(w)^2. + max_ydeg(w)^2., ind)
   w = w(ind)
   mid_Xra = mid_Xra(w)
   mid_Ydec = mid_Ydec(w)
   off_cen_Xra = off_cen_Xra(w)
   off_cen_Ydec = off_cen_Ydec(w)
endelse



for detarr = 0, narr - 1 do if ind_arrays(detarr) eq 1 then begin
   pos = stregex(tags, 'Xra_bolos_' + arrays_default(detarr), /fold_case)
   w = where(pos ge 0)
   Xra_bolos_scan = scan.(w(0))
   pos = stregex(tags, 'Ydec_bolos_' + arrays_default(detarr), /fold_case)
   w = where(pos ge 0)
   Ydec_bolos_scan = scan.(w(0))

   gnomonic, Xra_bolos_scan, Ydec_bolos_scan, refx=mid_Xra, refy=mid_Ydec
   *Xra_bolos(detarr) = Xra_bolos_scan
   *Ydec_bolos(detarr) = Ydec_bolos_scan

   if keyword_set(w_exc) then begin
      (*Xra_bolos(detarr))(w_exc, *) = !values.f_nan
      (*Ydec_bolos(detarr))(w_exc, *) = !values.f_nan
   endif
   Xra_bolos_scan = 0
   Ydec_bolos_scan = 0
endif



if (not keyword_set(hdr_ref)) then begin
   crval1 = -mid_Xra
   crval2 = mid_Ydec
endif



;;; prepare to slice data into spatial blocks:

w = where(ind_arrays_eff eq 1)
w = w(0)
margin = 1.5 * drdeg_max(w)
if ((ntb gt max_n_samples or nblocks gt 1) and (not keyword_set(hdr_ref))) then begin
   if nblocks le 1 then nblocks = 1L + ntb / max_n_samples else nblocks = long(nblocks)
   Xra_scan = (*Xra_bolos(w))(*, boloref(w)) * cos_scanrot $
      + (*Ydec_bolos(w))(*, boloref(w)) * sin_scanrot + off_cen_Xra
   Ydec_scan = (*Ydec_bolos(w))(*, boloref(w)) * cos_scanrot $
      - (*Xra_bolos(w))(*, boloref(w)) * sin_scanrot + off_cen_Ydec

   dxdeg = max_xdeg(w)
   dydeg = max_ydeg(w)
   min_Xra_scan = drdeg_max(w)
   min_Ydec_scan = drdeg_max(w)

   dxydeg_block = sqrt(dxdeg * dydeg / double(nblocks))
   if (dxdeg le 1.2 * dxydeg_block and dxdeg le dydeg) then begin
      dxdeg_block = dxdeg
      dydeg_block = dydeg / double(nblocks)
      nxbl = 1
      nybl = nblocks
   endif else if (dydeg le 1.2 * dxydeg_block and dydeg lt dxdeg) then begin
      dydeg_block = dydeg
      dxdeg_block = dxdeg / double(nblocks)
      nybl = 1
      nxbl = nblocks
   endif else if dxdeg le dydeg then begin
      nxbl = floor(dxdeg / dxydeg_block) + 1
      dxdeg_block = dxdeg / double(nxbl)
      nybl = ceil(nblocks / float(nxbl))
      dydeg_block = dydeg / double(nybl)
      nblocks = nxbl * nybl
   endif else begin
      nybl = floor(dydeg / dxydeg_block) + 1
      dydeg_block = dydeg / double(nybl)
      nxbl = ceil(nblocks / float(nybl))
      dxdeg_block = dxdeg / double(nxbl)
      nblocks = nxbl * nybl
   endelse

   xdeg_blocks = dblarr(nblocks, 2)
   ydeg_blocks = dblarr(nblocks, 2)
   tab_nxbl = lindgen(nxbl)
   for ybl = 0, nybl - 1 do begin
      xdeg_blocks(tab_nxbl+nxbl*ybl, 0) = min_Xra_scan + dxdeg_block * tab_nxbl - margin
      ydeg_blocks(tab_nxbl+nxbl*ybl, 0) = min_Ydec_scan + dydeg_block * ybl - margin
   endfor
   xdeg_blocks(*, 1) = xdeg_blocks(*, 0) + dxdeg_block + 2.D0 * margin
   ydeg_blocks(*, 1) = ydeg_blocks(*, 0) + dydeg_block + 2.D0 * margin

   tab_nybl = lindgen(nybl)
   xdeg_blocks(nxbl*tab_nybl, 0) -= drdeg_max(w)
   xdeg_blocks(nxbl-1+nxbl*tab_nybl, 1) += drdeg_max(w)
   ydeg_blocks(tab_nxbl, 0) -= drdeg_max(w)
   ydeg_blocks(tab_nxbl+nxbl*(nybl-1), 1) += drdeg_max(w)
   tab_nxbl = 0
   tab_nybl = 0

   save, file=file_XYscan, Xra_scan, Ydec_scan, /xdr

endif else nblocks = 1


ptr_free, Xra_bolos, Ydec_bolos

if (orient eq 'astro' or keyword_set(hdr)) then save, file=file_param_blocks, $
   p_scanrot_deg, p_maprot_deg, p_off_cen_Xra, p_off_cen_Ydec, p_mid_Xra, p_mid_Ydec, p_max_xdeg, p_max_ydeg, $
   scanrot_deg, maprot_deg, off_cen_Xra, off_cen_Ydec, mid_Xra, mid_Ydec, drdeg_max, max_xdeg, max_ydeg, $
   crval1, crval2, /xdr $
 else save, file=file_param_blocks, scanrot_deg, maprot_deg, $
    off_cen_Xra, off_cen_Ydec, mid_Xra, mid_Ydec, drdeg_max, max_xdeg, max_ydeg, crval1, crval2, /xdr


end
