pro prepare_proj_coarse, file_scalars=file_scalars, onegrid=onegrid, exclude_nza=exclude_nza, $
                         cut_start=cut_start, ind_subscans=ind_subscans, ind_scans=ind_scans


;;; projection vectors to speed up mapping of time and bolometer indices onto coarse pixels

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

restore, file_pixcoords


off_coord = -0.5
foff = fact * off_coord
ntb = nt * nb

if keyword_set(onegrid) then begin
   indxy_beam_coarse = dimx_iter * floor(y_bolos / fact) + floor(x_bolos / fact)
   x_bolos = 0
   y_bolos = 0
endif else begin
   indxy_beam_coarse = reform(dimx_iter * floor(y_bolos / fact) + floor(x_bolos / fact), ntb)
   off_indxy_beam_coarse = reform(dimx_iter * floor((foff + y_bolos) / fact) $
      + floor((foff + x_bolos) / fact), ntb)
   wnan = where(x_bolos lt -foff or y_bolos lt -foff, cw_nan)
   x_bolos = 0
   y_bolos = 0
   if cw_nan ge 1 then off_indxy_beam_coarse(wnan) = -9L - dimx_iter * dimy_iter
   wnan = 0
endelse


;;; suppress invalid samples:

restore, file_signal

wnan = where(finite(signal) lt 1, cw_nan)
signal = 0
if cw_nan ge 1 then begin
   indxy_beam_coarse(wnan) = -9L
   if (not keyword_set(onegrid)) then off_indxy_beam_coarse(wnan) = -9L - dimx_iter * dimy_iter
endif
wnan = 0


;;; optionally suppress non-zero acceleration data and scan start after PCAL flashes:

if keyword_set(onegrid) then begin
   if keyword_set(exclude_nza) then begin
      restore, file_times
      time_s = 0
      wnza = where(flag_nza lt 1, cw_nza)
      if cw_nza ge 1 then indxy_beam_coarse(wnza, *) = -9L
      flag_nza = 0
   endif

   if keyword_set(cut_start) then begin
      wcut = where(cut_start gt 0, cw_cut)
      for c = 0, cw_cut - 1 do begin
         subs = where(ind_subscans ge ind_scans(wcut(c)))
         indxy_beam_coarse(ind_subscans(subs(0)):ind_subscans(subs(0)+1)-1, *) = -9L
      endfor
   endif

   histo_indxy = histogram(indxy_beam_coarse, reverse_indices=revind, min=0., max=dimx_iter*dimy_iter-1)
   indxy_beam_coarse = 0
   wproj = where(histo_indxy ge 1L, nproj)


endif else begin
   histo_indxy = histogram([indxy_beam_coarse, off_indxy_beam_coarse+dimx_iter*dimy_iter], $
      reverse_indices=revind, min=0., max=2*dimx_iter*dimy_iter-1)
   indxy_beam_coarse = 0
   off_indxy_beam_coarse = 0

   n1_revind = n_elements(histo_indxy) + 1
   revind(n1_revind:*) = temporary(revind(n1_revind:*)) mod ntb
   wproj = where(histo_indxy ge 1L, nproj)
endelse


;;; save projection arrays:
save, file=file_projcoarse, wproj, nproj, revind, /xdr

end
