pro prepare_proj_fine, file_scalars=file_scalars, xdim=xdim, ydim=ydim, $
                       scan_next=scan_next, subscan_next=subscan_next, $
                       ind_scans=ind_scans, ind_subscans=ind_subscans, $
                       nzadata=nzadata, cut_pcal=cut_pcal


;;; projection vectors to speed up mapping of time and bolometer indices onto fine pixels

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

if (not keyword_set(xdim)) then xdim = dimx
if (not keyword_set(ydim)) then ydim = dimy


restore, file_pixcoords

restore, file_signal



;;; optionally retain a single scan or a single subscan:

if keyword_set(scan_next) then begin
   if !version.memory_bits ge 64 then indxy_beam = -9L + lon64arr(nt, nb) else indxy_beam = -9L + lonarr(nt, nb)
   scan = scan_next - 1
   slice_indxy = floor(x_bolos(ind_scans(scan):ind_scans(scan+1)-1, *)) $
      + xdim * floor(y_bolos(ind_scans(scan):ind_scans(scan+1)-1, *))

   ;;; suppress invalid samples:
   wnan = where(finite(signal(ind_scans(scan):ind_scans(scan+1)-1, *)) lt 1 $
      or x_bolos(ind_scans(scan):ind_scans(scan+1)-1, *) lt 0. $
      or x_bolos(ind_scans(scan):ind_scans(scan+1)-1, *) ge xdim $
      or y_bolos(ind_scans(scan):ind_scans(scan+1)-1, *) lt 0. $
      or y_bolos(ind_scans(scan):ind_scans(scan+1)-1, *) ge ydim, cw_nan)
   signal = 0
   x_bolos = 0
   y_bolos = 0
   if cw_nan ge 1 then slice_indxy(wnan) = -9L
   wnan = 0

   ;;; optionally supress scan start after PCAL flashes:
   if keyword_set(cut_pcal) then if cut_pcal(scan) gt 0 then begin
      subs = where(ind_subscans ge ind_scans(scan))
      slice_indxy(0:ind_subscans(subs(0)+1)-1-ind_scans(scan), *) = -9L
   endif

   if keyword_set(nzadata) then if nzadata eq 'no' then begin
      restore, file_times
      time_s = 0
      wnza = where(flag_nza(ind_scans(scan):ind_scans(scan+1)-1) eq 0, cw_nza)
      if cw_nza ge 1 then slice_indxy(wnza, *) = -9L
      wnza = 0
      flag_nza = 0
   endif

   indxy_beam(ind_scans(scan), 0) = slice_indxy



endif else if keyword_set(subscan_next) then begin
   if !version.memory_bits ge 64 then indxy_beam = -9L + lon64arr(nt, nb) else indxy_beam = -9L + lonarr(nt, nb)
   subs = subscan_next - 1
   slice_indxy = floor(x_bolos(ind_subscans(subs):ind_subscans(subs+1)-1, *)) $
      + xdim * floor(y_bolos(ind_subscans(subs):ind_subscans(subs+1)-1, *))

   ;;; suppress invalid samples:
   wnan = where(finite(signal(ind_subscans(subs):ind_subscans(subs+1)-1, *)) lt 1 $
      or x_bolos(ind_subscans(subs):ind_subscans(subs+1)-1, *) lt 0. $
      or x_bolos(ind_subscans(subs):ind_subscans(subs+1)-1, *) ge xdim $
      or y_bolos(ind_subscans(subs):ind_subscans(subs+1)-1, *) lt 0. $
      or y_bolos(ind_subscans(subs):ind_subscans(subs+1)-1, *) ge ydim, cw_nan)
   signal = 0
   x_bolos = 0
   y_bolos = 0
   if cw_nan ge 1 then slice_indxy(wnan) = -9L
   wnan = 0

   if keyword_set(nzadata) then if nzadata eq 'no' then begin
      restore, file_times
      time_s = 0
      wnza = where(flag_nza(ind_subscans(subs):ind_subscans(subs+1)-1) eq 0, cw_nza)
      if cw_nza ge 1 then slice_indxy(wnza, *) = -9L
      wnza = 0
      flag_nza = 0
   endif

   indxy_beam(ind_subscans(subs), 0) = slice_indxy



endif else begin
   indxy_beam = floor(x_bolos) + xdim * floor(y_bolos)

   ;;; suppress invalid samples:
   wnan = where(finite(signal) lt 1 or x_bolos lt 0. or x_bolos ge xdim $
      or y_bolos lt 0. or y_bolos ge ydim, cw_nan)
   signal = 0
   x_bolos = 0
   y_bolos = 0
   if cw_nan ge 1 then indxy_beam(wnan) = -9L
   wnan = 0

   ;;; optionally supress scan start after PCAL flashes:
   if keyword_set(cut_pcal) then begin
      wcut = where(cut_pcal gt 0, cw_cut)
      for c = 0, cw_cut - 1 do begin
         subs = where(ind_subscans ge ind_scans(wcut(c)))
         indxy_beam(ind_subscans(subs(0)):ind_subscans(subs(0)+1)-1, *) = -9L
      endfor
   endif

   if keyword_set(nzadata) then if nzadata eq 'no' then begin
      restore, file_times
      time_s = 0
      wnza = where(flag_nza eq 0, cw_nza)
      if cw_nza ge 1 then indxy_beam(wnza, *) = -9L
      wnza = 0
      flag_nza = 0
   endif
endelse




histo_indxy = histogram(indxy_beam, reverse_indices=revind_fine, min=0., max=xdim*ydim-1)
indxy_beam = 0
wproj_fine = where(histo_indxy ge 1L, nproj_fine)
histo_indxy = 0


;;; 1D index array of imprint lower-left corner (repetitions eliminated by wproj_fine):
vectpos_startimprint = floor(-xcmin_imprint + (wproj_fine mod xdim)) $
      + xdim * floor(-xcmin_imprint + wproj_fine / xdim)

;;; 1D index array of entire imprint, relative to the lower-left corner:
tab_ind = lindgen(dim_imprint)
sq_dim_imprint = dim_imprint^2
vectpos_imprint = reform(rebin(tab_ind, dim_imprint, dim_imprint), sq_dim_imprint) $
   + xdim * rebin(tab_ind, sq_dim_imprint, /sample)



save, file=file_projfine, wproj_fine, nproj_fine, revind_fine, $
   vectpos_startimprint, vectpos_imprint, /xdr

end
