pro proj_map, map_signal, file_scalars=file_scalars, nstep_imprint_final=nstep_imprint_final, $
              file_imprint_final=file_imprint_final, dimx_proj=dimx_proj, dimy_proj=dimy_proj, $
              maxnoise=maxnoise, ind_scans=ind_scans, comp_errors=comp_errors, lowfnoise=lowfnoise, $
              todisk=todisk


;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

if keyword_set(nstep_imprint_final) then nstep_imprint = nstep_imprint_final
if keyword_set(file_imprint_final) then file_imprint = file_imprint_final


if not keyword_set(dimx_proj) then begin
   dimx_proj = dimx
   dimy_proj = dimy
endif

map_weight = fltarr(dimx_proj, dimy_proj)
map_signal = fltarr(dimx_proj, dimy_proj)
if keyword_set(comp_errors) then begin
   map_error = fltarr(dimx_proj, dimy_proj)
   map_sqw = fltarr(dimx_proj, dimy_proj)
   map_lowfnoise = fltarr(dimx_proj, dimy_proj)
endif else begin
   map_error = 0
   map_lowfnoise = 0
endelse

weight = 1. / maxnoise^2.
w = where(finite(weight) gt 0)
weight /= mean(weight(w))


;;; give less weight to data taken at non-nominal speed:

restore, file_vscan

weight_vscan = 1. + fltarr(nt)
dev_v = abs(1. - vscan / vscan_as_s)
w = where(dev_v gt 0.2, cw)
if cw ge 1 then weight_vscan(w) = 1. - dev_v(w)
w = where(vscan gt 2. * vscan_as_s, cw)
if cw ge 1 then weight_vscan(w) = 0.
vscan = 0


tab_scan = lonarr(ind_scans(1))
for i = 1, nscans - 1 do tab_scan = [temporary(tab_scan), i + lonarr(ind_scans(i+1) - ind_scans(i))]


restore, file_pixcoords

indimprint = nstep_imprint * floor(nstep_imprint * (y_bolos - floor(y_bolos))) $
   + floor(nstep_imprint * (x_bolos - floor(x_bolos)))
x_bolos = 0
y_bolos = 0


restore, file_projfine

restore, file_imprint

sq_dim_imprint = dim_imprint^2
imprint_reform = reform(imprint, nstep_imprint^2, sq_dim_imprint)
imprint = 0


restore, file_signal


for k = 0L, nproj_fine - 1 do begin

   ;;; quicker way of evaluating: wp = where(indxy_beam eq wproj_fine(k), cp):
   revind_0 = revind_fine(wproj_fine(k))
   revind_1 = revind_fine(wproj_fine(k)+1)
   cp = revind_1 - revind_0
   wp = revind_fine(revind_0:revind_1-1)
   wt = wp mod nt

   vectpos = vectpos_startimprint(k) + vectpos_imprint
   bin_weight = rebin(weight(tab_scan(wt) + nscans * (wp / nt)) * weight_vscan(wt), cp, sq_dim_imprint) $
      * imprint_reform(indimprint(wp), *)
   bin_signal = rebin(signal(wp), cp, sq_dim_imprint)
   map_weight(vectpos) += total(bin_weight, 1)
   map_signal(vectpos) += total(bin_signal * bin_weight, 1)

   if keyword_set(comp_errors) then begin
      map_sqw(vectpos) += total(bin_weight^2., 1)
      map_error(vectpos) += total(bin_signal^2. * bin_weight, 1)
      map_lowfnoise(vectpos) += total(rebin(lowfnoise(wp), cp, sq_dim_imprint) * bin_weight, 1)
   endif
endfor


wcov = where(map_weight gt 0.)
if keyword_set(comp_errors) then begin
   map_error(wcov) = (map_error(wcov) * map_weight(wcov) - map_signal(wcov)^2.) $
      / (map_weight(wcov)^2. - map_sqw(wcov))
   map_lowfnoise(wcov) /= map_weight(wcov)
endif
map_signal(wcov) /= map_weight(wcov)

if keyword_set(comp_errors) then map_error(wcov) = sqrt(map_error(wcov) / map_weight(wcov))


if (n_params() eq 0 or keyword_set(todisk)) then $
   save, file=file_maps, map_signal, map_weight, map_error, map_lowfnoise, /xdr

end
