pro proj_map_clean, map_signal, file_scalars=file_scalars, nstep_imprint_final=nstep_imprint_final, $
                    file_imprint_final=file_imprint_final, dimx_proj=dimx_proj, dimy_proj=dimy_proj, $
                    maxnoise=maxnoise, ind_scans=ind_scans, comp_errors=comp_errors, lowfnoise=lowfnoise, $
                    todisk=todisk


;;; builds weighted signal and error maps for each scan, and then weighs each scan
;;; by its inverse variance, to eliminate hot pixels and hot trails
;;; called only when nscans < 4

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; introduced November 2010: version 5

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

if keyword_set(nstep_imprint_final) then nstep_imprint = nstep_imprint_final
if keyword_set(file_imprint_final) then file_imprint = file_imprint_final


if (not keyword_set(dimx_proj)) then begin
   dimx_proj = dimx
   dimy_proj = dimy
endif
dimxy = dimx_proj * dimy_proj

cube_weight = fltarr(dimx_proj, dimy_proj, nscans)
cube_signal = fltarr(dimx_proj, dimy_proj, nscans)
cube_error = fltarr(dimx_proj, dimy_proj, nscans)
cube_sqw = fltarr(dimx_proj, dimy_proj, nscans)
if keyword_set(comp_errors) then begin
   map_error = fltarr(dimx_proj, dimy_proj)
   map_lowfnoise = fltarr(dimx_proj, dimy_proj)
endif else begin
   map_error = 0
   map_lowfnoise = 0
endelse

weight = 1. / maxnoise^2.
w = where(finite(weight) gt 0)
weight /= mean(weight(w))


;;; give less weight to data taken at non-nominal speed:

restore, file_vscan

weight_vscan = 1. + fltarr(nt)
dev_v = abs(1. - vscan / vscan_as_s)
w = where(dev_v gt 0.2, cw)
if cw ge 1 then weight_vscan(w) = 1. - dev_v(w)
w = where(vscan gt 2. * vscan_as_s, cw)
if cw ge 1 then weight_vscan(w) = 0.
vscan = 0


tab_scan = lonarr(ind_scans(1))
for i = 1, nscans - 1 do tab_scan = [temporary(tab_scan), i + lonarr(ind_scans(i+1) - ind_scans(i))]


restore, file_pixcoords

indimprint = nstep_imprint * floor(nstep_imprint * (y_bolos - floor(y_bolos))) $
   + floor(nstep_imprint * (x_bolos - floor(x_bolos)))
x_bolos = 0
y_bolos = 0


restore, file_projfine

restore, file_imprint

sq_dim_imprint = dim_imprint^2
imprint_reform = reform(imprint, nstep_imprint^2, sq_dim_imprint)
imprint = 0


restore, file_signal


for k = 0L, nproj_fine - 1 do begin

   ;;; quicker way of evaluating: wp = where(indxy_beam eq wproj_fine(k), cp):
   revind_0 = revind_fine(wproj_fine(k))
   revind_1 = revind_fine(wproj_fine(k)+1)
   cp = revind_1 - revind_0
   wp = revind_fine(revind_0:revind_1-1)
   wt = wp mod nt

   vectpos = vectpos_startimprint(k) + vectpos_imprint
   bin_weight = rebin(weight(tab_scan(wt) + nscans * (wp / nt)) * weight_vscan(wt), cp, sq_dim_imprint) $
      * imprint_reform(indimprint(wp), *)
   bin_signal = rebin(signal(wp), cp, sq_dim_imprint)

   for i = 0, nscans - 1 do begin
      ws = where(tab_scan(wt) eq i, cs)
      if cs ge 1 then begin
         ind = dimxy * i + vectpos
         cube_weight(ind) += total(bin_weight(ws, *), 1)
         cube_signal(ind) += total(bin_signal(ws, *) * bin_weight(ws, *), 1)
         cube_sqw(ind) += total(bin_weight(ws, *)^2., 1)
         cube_error(ind) += total(bin_signal(ws, *)^2. * bin_weight(ws, *), 1)
      endif
   endfor

   if keyword_set(comp_errors) then $
      map_lowfnoise(vectpos) += total(rebin(lowfnoise(wp), cp, sq_dim_imprint) * bin_weight, 1)
endfor



map_signal_uncl = total(cube_signal, 3)
map_weight = total(cube_weight, 3)
wcov = where(map_weight gt 0.)
if keyword_set(comp_errors) then begin
   map_error = total(cube_error, 3)
   map_sqw = total(cube_sqw, 3)
   map_error(wcov) = (map_error(wcov) * map_weight(wcov) - map_signal_uncl(wcov)^2.) $
      / (map_weight(wcov)^2. - map_sqw(wcov))
   map_lowfnoise(wcov) /= map_weight(wcov)
endif
map_signal_uncl(wcov) /= map_weight(wcov)

if keyword_set(comp_errors) then map_error(wcov) = sqrt(map_error(wcov) / map_weight(wcov))


wcub = where(cube_weight gt 0., complement=wnocub)
cube_error(wcub) = (cube_error(wcub) * cube_weight(wcub) - cube_signal(wcub)^2.) $
   / (cube_weight(wcub)^2. - cube_sqw(wcub))

;;; weight each scan map by its inverse variance map:
cube_weight_scan = cube_error
w = where(cube_error gt 0.)
med = median(cube_error(w))
w = where(finite(cube_error) lt 1 or cube_error lt med, cw)
if cw ge 1 then cube_weight_scan(w) = med
cube_weight_scan = 1. / temporary(cube_weight_scan)
cube_weight_scan(wnocub) = 0.

;;; do not weight pixels containing very bright sources:
thres = 600. * median(maxnoise)
w = where(min(cube_signal / cube_weight, dimension=3, /nan) gt thres $
   and min(cube_weight, dimension=3) gt 0., cw)
if cw ge 1 then for i = 1, nscans - 1 do cube_weight_scan(dimxy*i+w) = cube_weight_scan(w)

map_signal = total(cube_signal * cube_weight_scan, 3)
map_weight_sig = total(cube_weight * cube_weight_scan, 3)
map_signal(wcov) /= map_weight_sig(wcov)


if (n_params() eq 0 or keyword_set(todisk)) then $
   save, file=file_maps, map_signal, map_weight, map_error, map_lowfnoise, map_signal_uncl, /xdr

end
