function proj_map_init, tab_offsets, file_scalars=file_scalars, maxnoise=maxnoise, ind_scans=ind_scans


;;; project the initial map, after removing offsets (one per scan per bolometer)
;;; due to flux calibration constants

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

restore, file_signal


signal_orig = signal
tab_offsets = !values.f_nan + fltarr(nscans, nb)

for scan = 0, nscans - 1 do begin
   wb = where(finite(maxnoise(scan, *)) gt 0)
   tab_offsets(scan, wb) = median(signal(ind_scans(scan):ind_scans(scan+1)-1, wb), dimension=1)
endfor


wb = where(finite(total(maxnoise, 1)) gt 0)
med_offsets = median(tab_offsets(*, wb), dimension=2, /even)
offsets = total(tab_offsets - rebin(med_offsets, nscans, nb), 1, /nan) / total(finite(tab_offsets), 1)
w = where(finite(offsets) lt 1, cw)
if cw ge 1 then offsets(w) = 0.


for b = 0L, nb - 1 do signal(*, b) -= offsets(b)
save, file=file_signal, signal, /xdr

proj_map, map_init, file_scalars=file_scalars, maxnoise=maxnoise, ind_scans=ind_scans

signal = temporary(signal_orig)
save, file=file_signal, signal, /xdr


return, map_init

end
