pro scanamorphos, version=version, spire=spire, pacs=pacs, nobs=nobs, $
                  visu=visu, vis_traject=vis_traject, debug=debug, jumps_pacs=jumps_pacs, $
                  nothermal=nothermal, noglitch=noglitch, nogains=nogains, $
                  parallel=parallel, galactic=galactic, minimap=minimap, nocross=nocross, $
                  files_scans_frame=files_scans_frame, frame_fromallscans=frame_fromallscans, $
                  hdr_ref=hdr_ref, nblocks=nblocks, block_start=block_start, one_plane_fits=one_plane_fits, $
                  offset_ra_as=offset_ra_as, offset_dec_as=offset_dec_as, $
                  batch=batch, file_out=file_out, orient=orient, pixsize=pixsize, nzadata=nzadata


;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; with contributions from Marc Sauvage and Pierre Chanial
;;; for the formatting of the input data

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1
;;; October 2011: version 14

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; to be used with input structures formatted from level 1 HIPE files
;;; by  ../pro_format_input/convert_hcssfits_[inst].pro
;;;
;;; directory and file names of the input structures in the ascii file  ../scanlist_[inst]
;;;
;;; "[inst]": either "spire" or "pacs"
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; VERY IMPORTANT:
;;;
;;; both instruments: In the input file list, always interlace the scan directions (scan_a and scan_b).
;;;                   The astrometry frame is determined from the first two scans only,
;;;                   and the destriping assumes scan_a - scan_b - scan_a - scan_b - ....
;;;                   (default ordering for SPIRE).
;;;
;;; PACS: It is assumed that a PCAL flash took place just before each observation
;;;       (affecting the first subscan of the first scan of each scan_a or scan_b observation).
;;;
;;;
;;; PACS example: obs1 with 2 repetitions of scan_a, obs2 with 2 repetitions of scan_a,
;;;               obs3 with 2 repetitions of scan_b, obs4 with 2 repetitions of scan_b
;;;
;;; --> input file list in this order:
;;; obs1_scan1 - obs3_scan1 - obs1_scan2 - obs3_scan2 - obs2_scan1 - obs4_scan1 - obs2_scan2 - obs4_scan2
;;;
;;;
;;; In case of doubt, it might be a good idea to check that scan directions are
;;; alternating correctly by using the /vis_traject option.
;;;
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; automatically slices the data into overlapping spatial blocks when the data volume has to be cut;
;;; maximum block volume depends on the "max_n_samples" parameter
;;; (function of machine characteristics, has to be calibrated)
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



resolve_all, /continue_on_error, /quiet
loadct, 15, /silent


;;; parent directory of SCANAMORPHOS/:

root = ''
openr, 1, '$HOME/scanam.dir'
readf, 1, format='(A)', root
close, 1



;;;;;;;;;;;;;;;;
;;; definitions
;;;;;;;;;;;;;;;;


if keyword_set(version) then get_version, parentdir=root

if keyword_set(spire) then begin
   inst = 'spire'
   narr = 3
   arrays_names = ['250', '350', '500']
   nbolo = [139L, 88L, 43L]
   fwhm_as = [18., 25., 36.]
   boloref = [104L, 83L, 31L]
   fact_scales = [2., 2., 2.]
   tab_npass_crit = [3, 3, 3]
   fact_npass_cen = 0.8
   if keyword_set(parallel) then begin
      logf_inf = 0.2
      logf_sup = 0.5
      logf_high_inf = 0.5
      logf_high_sup = 0.7
   endif else begin
      logf_inf = 0.4
      logf_sup = 0.7
      logf_high_inf = 0.7
      logf_high_sup = 1.
   endelse
endif else if keyword_set(pacs) then begin
   inst = 'pacs'
   narr = 3
   arrays_names = ['70', '100', '160']
   nbolo = [64L*32L, 64L*32L, 32L*16L]
   fwhm_as = [5.6, 6.8, 11.4]
   pfov = [3.2, 3.2, 6.4]
   p_arrhw = [16.5, 16.5, 8.5]
   boloref = [31L+64L*15L, 31L+64L*15L, 15L+32L*7L]
   fact_scales = [2., 2., 2.]
   tab_npass_crit = [10, 10, 10]
   fact_npass_cen = 0.72
   logf_inf = 0.5
   logf_sup = 0.7
   logf_high_inf = 0.1
   logf_high_sup = 0.5
endif else begin
   print, 'no valid instrument selected: exit'
   goto, end_label
endelse




nc = strlen(root)
if strmid(root, nc-1) ne '/' then root = root + '/'
dir_scanmap = root + 'SCANAMORPHOS/'
file_params = 'scanmap_userparams_' + inst + '.xdr'
file_scanlist = 'scanlist_' + inst
dir_cal = dir_scanmap + 'calib/' + inst + '/'

if inst eq 'spire' then iter_aver_max = 3 else iter_aver_max = 1
iter_max = 10

fact_fwhm = [1., 1., 1.]   ;;; fact_fwhm * FWHM = stability length (automatically increased if needed)
min_nsamples = [6, 6, 6]   ;;; minimum number of samples per stability length crossing
tc_drift_s = 0.2           ;;; minimum timescale for drifts (automatically increased as needed)



;;; following parameter depends on machine characteristics
;;; note: if nothermal option checked, constraints can be relaxed

;;; if nt * nb > max_n_samples, then the field of view will be sliced into several blocks:
max_n_samples = long(6.5D+8)   ;;; maximum tested so far (with MemTotal = 48 Gb)



if not keyword_set(debug) then debug = 0
if not keyword_set(visu) then visu = 0
if keyword_set(debug) then visu = 1

if keyword_set(minimap) then nothermal = 1
if (keyword_set(galactic) or (not keyword_set(pacs))) then jumps_pacs = 0

if (not keyword_set(nblocks)) then nblocks = 1



;;;;;;;;;;;;;;;;;;;;
;;; user parameters
;;;;;;;;;;;;;;;;;;;;


dir_in = ''
files_scans = ''
file = ''
openr, 1, dir_scanmap + file_scanlist
readf, 1, format='(A)', dir_in
dir_in = strtrim(dir_in, 2)
nc = strlen(dir_in)
if strmid(dir_in, nc-1) ne '/' then dir_in = dir_in + '/'
readf, 1, format='(A)', files_scans
files_scans = strtrim(files_scans, 2)
while (not eof(1)) do begin
   readf, 1, format='(A)', file
   files_scans = [files_scans, strtrim(file, 2)]
endwhile
close, 1
nscans = n_elements(files_scans)

if (keyword_set(offset_ra_as) or keyword_set(offset_dec_as)) then begin
   if (n_elements(offset_ra_as) ne nscans or n_elements(offset_dec_as) ne nscans) then begin
      print, 'The number of astrometric offsets does not match the number of scans: exit.'
      goto, end_label
   endif
endif


if (not keyword_set(files_scans_frame)) then files_scans_frame = files_scans(0:min([1, nscans-1]))
if keyword_set(frame_fromallscans) then files_scans_frame = files_scans

dir_out_default = dir_in
file_out_default = strmid(files_scans(0), 0, strlen(files_scans(0)) - 4)

arrays_default = ['PSW', 'PMW', 'PLW']
orient_default = 'scan'
pix_as_default = fwhm_as / 4.
nzadata_default = 'yes'

;;; restore previous parameter choices:
if (not keyword_set(batch)) then begin
   restore, dir_scanmap + file_params
   dir_out_prev = dir_out
   file_out_prev = file_out
   arrays_prev = arrays
   orient_prev = orient
   pix_as_prev = tab_pix_as
   nzadata_prev = nzadata
endif


crs = string(13b) + string(10b)

for i = 0, nscans - 1 do begin
   test = file_test(dir_in + files_scans(i), /read)
   if test eq 0 then begin
      print, crs + 'input file ' + chain(i+1) + ' inexistent or non-readable: exit'
      goto, end_label
   endif
endfor
print, crs + chain(nscans) + ' scan(s) to be combined'

if (keyword_set(jumps_pacs) and (not keyword_set(visu))) then $
   print, crs + 'warning: /jumps_pacs calls a beta-release module in INTERACTIVE MODE to check detections'



if keyword_set(batch) then begin
   dir_out = dir_in
   if inst eq 'spire' then arrays = ['PSW', 'PMW', 'PLW'] else arrays = ['PLW']
   if (not keyword_set(orient)) then orient = orient_default
   if keyword_set(pixsize) then begin
      if n_elements(pixsize) eq 1 then pixsize = pixsize(0) + fltarr(narr)
      tab_pix_as = pixsize
   endif else tab_pix_as = pix_as_default
   if (not keyword_set(nzadata)) then nzadata = nzadata_default
endif else begin

   dir_out = ''
   read, dir_out, prompt = crs + 'output directory:' + crs + '   default = ' + dir_out_default + ' [] ;' $
      + crs + '   previous = ' + dir_out_prev + ' [p] ;' + crs + '      choice = '
   if not keyword_set(dir_out) then dir_out = dir_out_default $
      else if strlowcase(dir_out) eq 'p' then dir_out = dir_out_prev $
      else if strmid(dir_out, strlen(dir_out) - 1, 1) ne '/' then dir_out = dir_out + '/'

   file_out = ''
   read, file_out, prompt = crs + 'output FITS file root:' + crs + '   default = ' + file_out_default + ' [] ;' $
      + crs + '   previous = ' + file_out_prev + ' [p] ;' + crs + '      choice = '
   if not keyword_set(file_out) then file_out = file_out_default $
      else if strlowcase(file_out) eq 'p' then file_out = file_out_prev
endelse

test = file_test(dir_out, /directory, /write)
if test eq 0 then begin
   print, crs + 'output directory non-existent or non-writable: exit'
   goto, end_label
endif


if (not keyword_set(batch)) then begin
   arrays = ''
   if inst eq 'spire' then arr_def = arrays_default else arr_def = ['PLW']
   read, arrays, prompt = crs + 'detector arrays:'  + $
      crs + '(note: [PSW, PMW, PLW] = [70, 100, 160] for PACS ; [250, 350, 500] for SPIRE)' $
      + crs + '   default = ' + strjoin(arr_def, ', ') + ' [] ;' $
      + crs + '   previous = ' + strjoin(arrays_prev, ', ') + ' [p] ;' + crs + '      choice = '
   if not keyword_set(arrays) then arrays = arr_def $
      else if strlowcase(arrays) eq 'p' then arrays = arrays_prev $
      else arrays = find_all_regex(strupcase(arrays), 'P(S|M|L)W')
endif

ind_arrays = intarr(narr)
for detarr = 0, narr - 1 do begin
   w = where(arrays eq arrays_default(detarr), cw)
   if cw ge 1 then ind_arrays(detarr) = 1
endfor


if (not keyword_set(batch)) then begin
   orient = ''
   read, orient, prompt = crs + 'map orientation:' + crs + $
      'either along the scan or in the standard astronomical frame (N up, E left)' $
      + crs + '   default = ' + orient_default + ' [] ;' $
      + crs + '   previous = ' + orient_prev + ' [p] ;' + crs + '      choice = [ /p/scan/astro] '
   if not keyword_set(orient) then orient = orient_default $
      else if strlowcase(orient) eq 'p' then orient = orient_prev $
      else if strlowcase(strmid(orient, 0, 1)) eq 's' then orient = 'scan' $
      else orient = 'astro'

   choice = ''
   read, choice, prompt = crs + 'pixel size (arcsec):' $
      + crs + '   default (FWHM/4) = ' + strjoin(chain(pix_as_default, 2), ' ; ') + ' [] ;' $
      + crs + '   previous = ' + strjoin(chain(pix_as_prev, 2), ' ; ') + ' [p] ;' $
      + crs + '      choice = [ /p/new] '
   if not keyword_set(choice) then tab_pix_as = pix_as_default $
      else if strlowcase(choice) eq 'p' then tab_pix_as = pix_as_prev $
      else begin
         tab_pix_as = pix_as_default
         pas = 0.
         for i = 0, narr - 1 do begin
            w = where(arrays eq arrays_default(i), cw)
            if cw ge 1 then begin
               read, pas, prompt = 'for ' + arrays_default(i) + ': '
               tab_pix_as(i) = float(pas)
            endif
         endfor
      endelse

   nzadata = ''
   read, nzadata, prompt = crs + 'include data taken while acceleration is non-zero ? ' + $
      crs + '(note: always used for drifts determination,' + crs + 'but can be rejected before final projection)' $
      + crs + '   default = ' + nzadata_default + ' [] ;' $
      + crs + '   previous = ' + nzadata_prev + ' [p] ;' + crs + '      choice = [ /p/yes/no] '
   if not keyword_set(nzadata) then nzadata = nzadata_default $
      else if strlowcase(nzadata) eq 'p' then nzadata = nzadata_prev $
      else if strlowcase(strmid(nzadata, 0, 1)) eq 'y' then nzadata = 'yes' $
      else nzadata = 'no'
   print, crs + crs
endif


;;; for PACS, select the detector array corresponding to the data:
if inst eq 'pacs' then begin
   restore, dir_in + files_scans(0)
   tags = tag_names(scan)
   for detarr = 0, narr - 1 do begin
      pos = stregex(tags, 'signal_' + arrays_default(detarr), /fold_case)
      w = where(pos ge 0, cw)
      if cw ge 1 then if ind_arrays(detarr) lt 1 then begin
         if (not keyword_set(batch)) then begin
            print, crs + '****************************************************'
            print, crs + 'The selected detector array does not match the data:'
            print, 'Selecting ' + arrays_default(detarr) + ' instead.'
            print, crs + '****************************************************'
            print, crs
         endif
         ind_arrays = intarr(narr)
         ind_arrays(detarr) = 1
         arrays = [arrays_default(detarr)]
      endif
   endfor
   scan = 0
endif


;;; save new parameter choices:
save, file=dir_scanmap+file_params, dir_out, file_out, arrays, orient, tab_pix_as, nzadata, /xdr



dir_tempo = dir_in
file_scalars = dir_tempo + 'save_scalars.xdr'
file_times = dir_tempo + 'save_time_vectors.xdr'
file_time_coarse = dir_tempo + 'save_time_coarse.xdr'
file_projcoarse = dir_tempo + 'save_projcoarse.xdr'
file_projfine = dir_tempo + 'save_projfine.xdr'
file_pixcoords = dir_tempo + 'save_pixcoords.xdr'
file_vscan = dir_tempo + 'save_vscan.xdr'
file_ntmin_pass = dir_tempo + 'save_ntmin_pass.xdr'
file_maps = dir_tempo + 'save_maps.xdr'
file_XYscan = dir_tempo + 'save_XYscan.xdr'

root_signal = dir_tempo + 'save_signal_'
root_xyas = dir_tempo + 'save_xyas_'



;;; slightly different criterion to reject non-uniform emission for the blue and red arrays of PACS
;;; (red array is noisier), and for shallow observations:
if (inst eq 'pacs' and max(ind_arrays(0:1)) eq 1) then fact_npass_cen = 0.78
if (inst eq 'pacs' and ind_arrays(2) eq 1 and nscans lt 4) then fact_npass_cen = 0.7
if (inst eq 'pacs' and max(ind_arrays(0:1)) eq 1 and nscans lt 4) then fact_npass_cen = 0.76


;;; smaller sampling rate for the blue array of PACS in parallel mode:
if (keyword_set(parallel) and inst eq 'pacs' and max(ind_arrays(0:1)) eq 1) then begin
   logf_inf = 0.2
   logf_sup = 0.4
   logf_high_inf = -0.2
   logf_high_sup = 0.2
endif


;;; pixel grids used for the processing (different from pixel grid for final projection):
;;;   coarse grid to compute drift differences: FWHM
;;;   finer grid to project intermediate maps: FWHM / fact_scales
;;; Both scales are multiplied by fact_fwhm: 1. by default,
;;;   can be larger to improve statistics within stability length for shallow observations
;;;  (high scan speed with low sampling rate).


;;; pixel size for final projection:
tab_pix_as_final = tab_pix_as


;;; optionally come back here to increase stability length
;;; and improve sample statistics for PACS or SPIRE in parallel mode:

fact_fwhm_label:

tab_pix_as = tab_pix_as_final

;;; fine grid pixel size:
w = where(abs(1. - fact_scales * tab_pix_as / (fact_fwhm * fwhm_as)) gt 0.1, cw)
if cw ge 1 then tab_pix_as(w) = fact_fwhm(w) * fwhm_as(w) / fact_scales(w)


;;; diameter of the bolometer imprint on the sky map (assumed to be a uniform disk):
if inst eq 'spire' then tab_diam_as_imprint = fwhm_as / 2. / sqrt(!pi) $
   else tab_diam_as_imprint = 2. * pfov / sqrt(!pi)


;;; time gap between successive scans (crossing time of at least 10 FWHM(PLW) for v_scan >= 10"/s):
timegap = fact_fwhm(narr-1) * fwhm_as(narr-1)


if (nscans gt 1 and nscans lt 4) then clean = 1 else clean = 0



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; initial data readout to determine coordinates frame
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


file_param_blocks = dir_tempo + 'save_param_blocks.xdr'
prepare_blocks,  file_param_blocks, ind_arrays_eff, orient, nt, nblocks, xdeg_blocks, ydeg_blocks, $
   narr=narr, nscans=nscans, nbolo=nbolo, boloref=boloref, arrays_default=arrays_default, $
   max_n_samples=max_n_samples, ind_arrays=ind_arrays, debug=debug, $
   dir_in=dir_in, files_scans_frame=files_scans_frame, hdr_ref=hdr_ref, file_XYscan=file_XYscan

restore, file_param_blocks


;;; common grid for all arrays ?

w = where(ind_arrays eq 1)
if max(abs(tab_pix_as_final(w) - tab_pix_as_final(w(0)))) eq 0 then begin
   common_grid = 1
   ind = max(w)
endif else begin
   w = where(ind_arrays_eff eq 1)
   ind = max(w)
endelse

if keyword_set(hdr_ref) then off_map_as = 0. else off_map_as = 4. * fwhm_as(ind)

if keyword_set(common_grid) then begin
   dimxas = float(max_xdeg(ind) * 3600.D0) + 2. * off_map_as
   dimyas = float(max_ydeg(ind) * 3600.D0) + 2. * off_map_as
   if keyword_set(hdr_ref) then begin
      dimx_final = sxpar(hdr_ref, 'NAXIS1')
      dimy_final = sxpar(hdr_ref, 'NAXIS2')
      crpix1_final = sxpar(hdr_ref, 'CRPIX1') - 0.5
      crpix2_final = sxpar(hdr_ref, 'CRPIX2') - 0.5
   endif else begin
      dimx_final = ceil(dimxas / tab_pix_as_final(ind))
      dimy_final = ceil(dimyas / tab_pix_as_final(ind))
      crpix1_final = (float(off_cen_Xra * 3600.D0) + off_map_as) / tab_pix_as_final(ind)
      crpix2_final = (float(off_cen_Ydec * 3600.D0) + off_map_as) / tab_pix_as_final(ind)
   endelse
endif


;;; common grid for multiple blocks:

if nblocks gt 1 then begin
   dimxas = float(max_xdeg(ind) * 3600.D0) + 2. * off_map_as
   dimyas = float(max_ydeg(ind) * 3600.D0) + 2. * off_map_as
   dimx = ceil(dimxas / tab_pix_as(ind))
   dimy = ceil(dimyas / tab_pix_as(ind))
   crpix1 = (float(off_cen_Xra * 3600.D0) + off_map_as) / tab_pix_as(ind)
   crpix2 = (float(off_cen_Ydec * 3600.D0) + off_map_as) / tab_pix_as(ind)
endif



;;;;;;;;;;;;;;;;;;;;
;;; data block loop
;;;;;;;;;;;;;;;;;;;;


if (not keyword_set(block_start)) then block_start = 0L else block_start = long(block_start)
for block = block_start, nblocks - 1L do begin

   ;;; initialization of the maximum time gap within a subscan:
   maxtimediff = 3.

   ingest_data, file_param_blocks, ind_scans, ind_subscans, nsubscans, vscan_as_s, timestep, $
                narr=narr, nscans=nscans, nt=nt, nblocks=nblocks, block=block, $
                inst=inst, nbolo=nbolo, boloref=boloref, ind_arrays_eff=ind_arrays_eff, $
                arrays_default=arrays_default, xdeg_blocks=xdeg_blocks, ydeg_blocks=ydeg_blocks, $
                orient=orient, timegap=timegap, dir_in=dir_in, files_scans=files_scans, $
                hdr_ref=hdr_ref, file_XYscan=file_XYscan, file_times=file_times, file_vscan=file_vscan, $
                root_signal=root_signal, root_xyas=root_xyas, maxtimediff=maxtimediff, $
                visu=visu, vis_traject=vis_traject, frame_fromallscans=frame_fromallscans, $
                galactic=galactic, minimap=minimap, offset_ra_as=offset_ra_as, offset_dec_as=offset_dec_as

   ;;; temporarily enlarge the bolometer imprint if the redundancy is very limited:
   if (inst eq 'spire' and keyword_set(parallel) and vscan_as_s gt 45.) $
      then tab_diam_as_imprint = 2. * fwhm_as / fact_scales / sqrt(!pi)

   ;;; deactivate average drift subtraction if the map is very small:
   if nsubscans lt 3 * nscans then nothermal = 1



;;;;;;;;;;;;;;;
;;; array loop
;;;;;;;;;;;;;;;


for detarr = 0, narr - 1 do if ind_arrays_eff(detarr) eq 1 then begin
   if nblocks eq 1 then print, crs + crs + 'building ' + arrays_default(detarr) + ' array map'


   ;;; if number of coarse time steps > max_nt_coarse,
   ;;; then the minimum timescale of the average drift will be increased:
   ;;; square of long(4.6D+4) ~= long integer limit
   if !version.memory_bits ge 64 then max_nt_coarse = long64(1.D+5) else max_nt_coarse = long(4.6D+4)


   ;;;initialize processing time for array:
   t_proc = 0.D0
   time_1 = systime(/seconds)


   nb = nbolo(detarr)
   bref = boloref(detarr)
   if !version.memory_bits ge 64 then begin
      nb = long64(nb)
      bref = long64(bref)
   endif

   pix_as = tab_pix_as(detarr)
   if keyword_set(hdr_ref) then begin
      cdelt1 = sxpar(hdr_ref, 'CDELT1', count=cc)
      if cc gt 0 then pix_as_final = float(3600. * sqrt(abs(cdelt1) * sxpar(hdr_ref, 'CDELT2'))) $
         else pix_as_final = float(3600. * sqrt((sxpar(hdr_ref, 'CD1_1'))^2. + (sxpar(hdr_ref, 'CD1_2'))^2.))
   endif else pix_as_final = tab_pix_as_final(detarr)

   nstep_imprint = 10L    ;;; must be even and sufficiently large to ensure accurate imprint projection
   nstep_imprint_final = 10L
   diam_as_imprint = tab_diam_as_imprint(detarr)
   npass_crit = tab_npass_crit(detarr)

   file_signal = root_signal + arrays_default(detarr) + '.xdr'
   file_xyas = root_xyas + arrays_default(detarr) + '.xdr'
   file_xyas_final = root_xyas + 'final_' + arrays_default(detarr) + '.xdr'
   file_imprint = dir_cal + arrays_default(detarr) + '_imprintgrid_' + chain(pix_as, 2) + 'as_' $
      + chain(nstep_imprint) + '.xdr'
   file_imprint_final = dir_cal + arrays_default(detarr) + '_imprintgrid_' + chain(pix_as_final, 2) + 'as_' $
      + chain(nstep_imprint_final) + '.xdr'

   ;;; optionally allow one more iteration for the average drift for PACS:
   if (inst eq 'pacs' and nt * nb lt 5.e+7) then iter_aver_max = 2


   ;;; correct for relative gains (SPIRE):

   if (inst eq 'spire' and (not keyword_set(nogains))) then begin
      restore, dir_cal + 'relative_gains_' + arrays_default(detarr) + '.xdr'
      restore, file_signal
      signal /= (relgains ## (fltarr(nt) + 1.))
      save, file=file_signal, signal, /xdr
      relgains = 0
      signal = 0
   endif


   ;;; exclude first subscan of each PACS observation from destriping fits
   ;;; (strong transient from PCAL flash):
   cut_start = intarr(nscans)
   if (inst eq 'pacs' and nblocks eq 1 and nsubscans ge 3 * nscans) then begin
      if (not keyword_set(nobs)) then read, nobs, prompt='number of distinct observations ? '
      while (not keyword_set(nobs)) do read, nobs
      nobs = round(nobs)
      nrep = nscans / nobs
      for i = 0, nobs - 1, 2 do cut_start(i*nrep) = 1
      for i = 1, nobs - 1, 2 do cut_start(1+(i-1)*nrep) = 1
   endif


   ;;; scale factor between coarse and fine grids:
   fact = ceil(fact_fwhm(detarr) * fwhm_as(detarr) / pix_as)
   pix_as_coarse = pix_as * fact



   restore, file_xyas_final

   ;;; map grid:
   if nblocks eq 1 then begin
      dimxas = float(max(xas) - min(xas)) + 2. * off_map_as
      dimyas = float(max(yas) - min(yas)) + 2. * off_map_as
      dimx = ceil(dimxas / pix_as)
      dimy = ceil(dimyas / pix_as)
      crpix1 = (float(off_cen_Xra * 3600.D0) + off_map_as) / pix_as
      crpix2 = (float(off_cen_Ydec * 3600.D0) + off_map_as) / pix_as
   endif

   if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
      maprot_deg = p_maprot_deg
      restore, file_xyas
      p_dimxas = float(max(xas) - min(xas)) + 2. * off_map_as
      p_dimyas = float(max(yas) - min(yas)) + 2. * off_map_as
      dimx = ceil(p_dimxas / pix_as)
      dimy = ceil(p_dimyas / pix_as)
   endif

   if !version.memory_bits ge 64 then begin
      dimx = long64(dimx)
      dimy = long64(dimy)
   endif


   ;;; pixel coordinates:
   x_bolos = float((off_map_as + xas) / pix_as)
   xas = 0
   y_bolos = float((off_map_as + yas) / pix_as)
   yas = 0
   save, file=file_pixcoords, x_bolos, y_bolos, /xdr
   x_bolos = 0
   y_bolos = 0



   ;;; number of time steps required to cross the FWHM:
   nt_cross = floor(pix_as_coarse / vscan_as_s / timestep)
   t_cross_fwhm = pix_as_coarse / vscan_as_s
   maxtimediff = 5 * nt_cross * timestep

   if nt_cross lt min_nsamples(detarr) then begin
      fact_fwhm = ceil(2. * min_nsamples * vscan_as_s * timestep / fwhm_as) / 2.
      w = where(fact_fwhm ge 5. and ind_arrays_eff gt 0, cw)
      if cw ge 1 then begin
         min_nsamples(w) = 4
         fact_fwhm = ceil(2. * min_nsamples * vscan_as_s * timestep / fwhm_as) / 2.
      endif
      w = where(fact_fwhm gt 2. and ind_arrays_eff gt 0, cw)
      if cw ge 1 then fact_scales(w) = fact_fwhm(w)
      w = where(ind_arrays_eff gt 0)
      print, crs + 'Too few samples per FWHM: increasing the stability length to'
      for k = 0, narr - 1 do if ind_arrays_eff(k) gt 0 then $
         print, '     ' + chain(fact_fwhm(k), 1) + ' FWHM for ' + arrays_default(k)
      print, '   and going back to recompute grids.' + crs
      vis_traject = 0
      goto, fact_fwhm_label
   endif


   ;;; dimensions for the coarse grid used in iterations:
   dimx_iter = dimx / fact
   if dimx mod fact gt 0 then dimx_iter += 1
   dimy_iter = dimy / fact
   if dimy mod fact gt 0 then dimy_iter += 1


   ;;; minimum number of samples per bolometer crossing:
   ntmin_pass = 999 + intarr(nt)
   restore, file_vscan
   w = where(vscan gt 0.)
   ntmin_pass(w) = floor(pix_as_coarse / vscan(w) / timestep)
   w = 0
   vscan = 0
   save, file=file_ntmin_pass, ntmin_pass, /xdr
   ntmin_pass = 0


   ;;; step of the coarse time grid:
   tc_min = fact_fwhm(detarr) * fwhm_as(detarr) / vscan_as_s
   tc_drift_s = max([tc_drift_s, tc_min])
   nt_aver = round(tc_drift_s / timestep)
   tc_drift_s = float(nt_aver * timestep)
   print, crs + 'Assuming for drifts a minimum timescale of ' + chain(tc_drift_s, 2) + ' s'
   print, 'Number of samples to cross a pixel of the coarse grid: ' + chain(nt_cross) + crs

   flag_grid_aver = 1
   tc_drift_s_ini = tc_drift_s
   tab_nt_indiv = nt_aver + lonarr(iter_max)
   n_lt_indiv = 0
   if (inst eq 'pacs' and nsubscans ge 3 * nscans and (not keyword_set(minimap))) then begin
      flag_grid_aver = 2
      restore, file_times
      dt_sub = median(time_s(ind_subscans(1:nsubscans)) - time_s(ind_subscans(0:nsubscans-1)))
      flag_nza = 0
      time_s = 0
      f_max = dt_sub / 2.2 / tc_drift_s
      if f_max ge 27. then begin
         tab_nt_indiv = nt_aver * [27L, 9L, 3L, 1L + lonarr(iter_max)]
         n_lt_indiv = 3
      endif else if f_max ge 9. then begin
         tab_nt_indiv = nt_aver * [9L, 3L, 1L + lonarr(iter_max)]
         n_lt_indiv = 2
      endif else if f_max ge 3. then begin
         tab_nt_indiv = nt_aver * [3L, 1L + lonarr(iter_max)]
         n_lt_indiv = 1
      endif else begin
         tab_nt_indiv = nt_aver * [1L + lonarr(iter_max)]
         n_lt_indiv = 0
      endelse
   endif

   if (nt / nt_aver ge max_nt_coarse and block eq block_start and (not keyword_set(nothermal))) then begin
      flag_grid_aver = 2
      nt_aver = ceil(float(nt) / float(max_nt_coarse))
      tc_drift_s_ini = float(nt_aver * timestep)
      print, crs + 'Increasing the minimum timescale for the average drift'
      print, 'in order to keep the number of coarse time steps below "max_nt_coarse".'
      print, '-> Assuming a new timescale of ' + chain(tc_drift_s_ini, 2) + ' s'
   endif


   ;;; coarse time grid made irregular in order not to mix edges of different scans:
   restore, file_times
   ;;; time_s, flag_nza
   flag_nza = 0

   ind_sc_coarse = [0L]
   w = where(time_s - time_s(0) le tc_drift_s_ini)
   ind_aver = [max(w) + 1]
   ws = 0
   repeat begin
      w = where(time_s(max(ind_aver):*) - time_s(max(ind_aver)) le tc_drift_s_ini)
      ind_aver_next = max(ind_aver) + max(w) + 1
      if ind_aver_next lt ind_scans(ws+1) then ind_aver = [ind_aver, ind_aver_next] $
         else begin
            ind_aver = [ind_aver, ind_scans(ws+1)]
            ind_sc_coarse = [ind_sc_coarse, n_elements(ind_aver)]
            ws += 1
         endelse
   endrep until (ws ge nscans)
   nt_save = n_elements(ind_aver)
   if !version.memory_bits ge 64 then nt_save = long64(nt_save)
   time_save = float((time_s([0, ind_aver(0:nt_save-2)]) + time_s(ind_aver-1)) / 2.)
   time_s = 0
   save, file=file_time_coarse, time_save, /xdr
   time_save = 0



   dim_imprint = 5L   ;;; dimension of the imprint image (must be odd)
   xcmin_imprint = dim_imprint / 2


   save, file=file_scalars, file_times, file_time_coarse, file_signal, file_pixcoords, $
      file_projcoarse, file_projfine, file_imprint, file_maps, file_vscan, file_ntmin_pass, $
      inst, nt, nt_save, nt_aver, timestep, maxtimediff, nb, bref, nscans, nsubscans, $
      logf_inf, logf_sup, logf_high_inf, logf_high_sup, $
      vscan_as_s, dimx, dimy, dimx_iter, dimy_iter, fact, $
      pix_as, diam_as_imprint, dim_imprint, nstep_imprint, xcmin_imprint, $
      npass_crit, fact_npass_cen, visu, debug, /xdr



   ;;; bolometer imprint matrix:
   if block eq block_start then compute_imprint, file_scalars=file_scalars



   ;;; estimate the noise level to be able to detect significant flux gradients:
   update_mask = 1
   if (keyword_set(visu) or keyword_set(jumps_pacs)) then window, 0, xs=1200
   compute_wnoise, file_scalars=file_scalars, maxnoise, highfnoise, update_mask=update_mask, $
      ind_scans=ind_scans, ind_subscans=ind_subscans, /initial, f_fwhm=fact_fwhm(detarr), $
      time_1=time_1, t_proc=t_proc
   print, crs + '"white noise" statistics (from spectral densities):'
   for i = 0, min([nscans - 1, 19]) do print, 'scan ' + chain(i) + ':   min = ' $
      + chain(min(maxnoise(i, *), /nan), 5) + '   max = ' + chain(max(maxnoise(i, *), /nan), 5)

   if (inst eq 'pacs' and keyword_set(debug)) then begin
      compute_diginoise, file_scalars=file_scalars, diginoise, maxnoise=maxnoise, $
         ind_scans=ind_scans, time_1=time_1, t_proc=t_proc
      print, crs + 'median of digitization noise / (< ' + chain(10.^logf_sup, 1) + ' Hz) noise: ' $
         + chain(median(diginoise / sqrt(3.) / maxnoise), 2)
   endif

   print, crs + '"threshold noise":'
   for i = 0, min([nscans - 1, 19]) do print, 'scan ' + chain(i) + ':   min = ' $
      + chain(min(highfnoise(i, *), /nan), 5) + '   max = ' + chain(max(highfnoise(i, *), /nan), 5)




   ;;; vectors for projection on fine grid:
   ;;; (to be done again every time missing samples are added)
   prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans


   ;;; initial maps (with and without flux cal. offsets):
   proj_map, map_simple, file_scalars=file_scalars, maxnoise=maxnoise, ind_scans=ind_scans, /todisk

   map_init = proj_map_init(tab_offsets, file_scalars=file_scalars, maxnoise=maxnoise, $
      ind_scans=ind_scans)



   ;;; vectors for projection on coarse grid:
   ;;; (to be done again every time missing samples are added)
   prepare_proj_coarse, file_scalars=file_scalars



   if keyword_set(visu) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      print, crs + 'initial signal spectral density (on scan length)' + crs
      compute_spd_signal, freq_comp, spd_comp_aver, file_scalars=file_scalars, $
         maxnoise=maxnoise, ind_scans=ind_scans, ind_subscans=ind_subscans
      time_1 = systime(/seconds)
   endif



   ;;;;;;;;;;;;;;;;;;;
   ;;; pre-processing
   ;;;;;;;;;;;;;;;;;;;


   ;;; initialize total drifts:
   drifts_all = fltarr(nt_save, nb)



   ;;; subtraction of baselines:

   ;;; to remove offsets due to inaccurate flux calibration,
   ;;; thermal drift to first order,
   ;;; and drifts with timescale similar to subscan duration

   print, crs + 'pre-processing: baseline subtraction (using simple fits)'
   baselines, drifts_all, map_back, smallfield, size_deg, maprot_deg=maprot_deg, $
      tab_offsets=tab_offsets, file_scalars=file_scalars, ind_sc_coarse=ind_sc_coarse, $
      ind_aver=ind_aver, ind_scans=ind_scans, ind_subscans=ind_subscans, maxnoise=maxnoise, $
      cut_start=cut_start, galactic=galactic, minimap=minimap, nocross=nocross, $
      time_1=time_1, t_proc=t_proc

   if keyword_set(clean) then proj_map_clean, file_scalars=file_scalars, $
      maxnoise=maxnoise, ind_scans=ind_scans $
   else proj_map, file_scalars=file_scalars, maxnoise=maxnoise, ind_scans=ind_scans
   restore, file_maps
   if keyword_set(visu) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      if keyword_set(clean) then map_signal = map_signal_uncl
      disp_ima, map_signal, win=1, weightmap=map_weight, title='after simple baseline subtraction'
      disp_ima, map_simple, win=2, weightmap=map_weight, title='raw data'
      disp_ima, map_init - map_signal, win=3, weightmap=map_weight, $
         title='baselines (linear fits, without flux cal. offsets)'
      res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
      time_1 = systime(/seconds)
   endif
   map_prev = map_signal
   map_signal = 0
   map_weight = 0



   ;;; correction for brightness discontinuities (PACS):
   if keyword_set(jumps_pacs) then corr_jumps_pacs, file_scalars=file_scalars, $
      ind_scans=ind_scans, ind_subscans=ind_subscans, maxnoise=maxnoise, f_fwhm=fact_fwhm(detarr), $
      time_1=time_1, t_proc=t_proc



   ;;; destriping to improve simple baselines (order 1 on subscan length):

   if (nscans ge 2 and (not keyword_set(nocross)) and (not keyword_set(minimap))) then begin
      print, crs + 'pre-processing: destriping (residual baseline subtraction using the redundancy)' + crs
      destripe, drifts_all, flag_valid_scans, map_back=map_back, smallfield=smallfield, $
         size_deg=size_deg, file_scalars=file_scalars, ind_sc_coarse=ind_sc_coarse, $
         ind_scans=ind_scans, ind_subscans=ind_subscans, maxnoise=maxnoise, $
         cut_start=cut_start, galactic=galactic, time_1=time_1, t_proc=t_proc

      if keyword_set(clean) then proj_map_clean, file_scalars=file_scalars, $
         maxnoise=maxnoise, ind_scans=ind_scans $
      else proj_map, file_scalars=file_scalars, maxnoise=maxnoise, ind_scans=ind_scans
      restore, file_maps
      if keyword_set(visu) then begin
         time_2 = systime(/seconds)
         t_proc += (time_2 - time_1)
         if keyword_set(clean) then map_signal = map_signal_uncl
         disp_ima, map_signal, win=1, weightmap=map_weight, title='after full baseline subtraction'
         disp_ima, map_prev - map_signal, win=2, weightmap=map_weight, $
            title='residual baselines (from redundancy)'
         print, 'note: The exclusion of bolometers during the destriping'
         print, '      may create false artefacts that are present'
         print, '      only in the difference map display (not in the signal).' + crs
         res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
         time_1 = systime(/seconds)
      endif
   endif else flag_valid_scans = 1 + intarr(nscans)
   map_prev = 0
   map_signal = 0
   map_weight = 0

   map_back = 0



   ;;; update white noise:
   if (not keyword_set(noglitch)) then mask_gl = 1 else mask_gl = 0
   compute_wnoise, file_scalars=file_scalars, maxnoise, highfnoise, ind_scans=ind_scans, $
      ind_subscans=ind_subscans, mask_gl=mask_gl, f_fwhm=fact_fwhm(detarr), cut_start=cut_start, $
      time_1=time_1, t_proc=t_proc
   print, crs + 'updated "white noise" statistics:'
   for i = 0, min([nscans - 1, 19]) do print, 'scan ' + chain(i) + ':   min = ' $
      + chain(min(maxnoise(i, *), /nan), 5) + '   max = ' + chain(max(maxnoise(i, *), /nan), 5)
   print, crs


   if keyword_set(visu) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      print, crs + 'signal spectral density after baseline subtraction (on scan length)' + crs
      compute_spd_signal, freq_comp, spd_comp_aver, file_scalars=file_scalars, $
         maxnoise=maxnoise, ind_scans=ind_scans, ind_subscans=ind_subscans
      time_1 = systime(/seconds)
   endif



   ;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ;;; drift subtraction loop
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;

   iter = 0
   iter_indiv = 0
   indiv_completed = 0
   amp_drifts = 999. * maxnoise

   if keyword_set(nothermal) then aver_completed = 1 else aver_completed = 0


   repeat begin

      if (aver_completed eq 1 and iter_indiv eq 0) then begin
         drifts_all_aver = drifts_all
         restore, file_time_coarse
         time_save_aver = time_save
         time_save = 0
         drifts_all = fltarr(nt_save, nb)
      endif

      ;;; update coarse time grid after average drift subtraction:
      if (aver_completed eq 1 and flag_grid_aver eq 2) then begin
         restore, file_scalars
         nt_aver_old = nt_aver
         restore, file_time_coarse
         time_save_old = time_save
         drifts_all_old = drifts_all
         nt_aver = tab_nt_indiv(iter_indiv)
         tc_drift_s = float(nt_aver * timestep)

         restore, file_times
         flag_nza = 0
         ind_sc_coarse = [0L]
         w = where(time_s - time_s(0) le tc_drift_s)
         ind_aver = [max(w) + 1]
         ws = 0
         repeat begin
            w = where(time_s(max(ind_aver):*) - time_s(max(ind_aver)) le tc_drift_s)
            ind_aver_next = max(ind_aver) + max(w) + 1
            if ind_aver_next lt ind_scans(ws+1) then ind_aver = [ind_aver, ind_aver_next] $
               else begin
               ind_aver = [ind_aver, ind_scans(ws+1)]
               ind_sc_coarse = [ind_sc_coarse, n_elements(ind_aver)]
               ws += 1
            endelse
         endrep until (ws ge nscans)
         nt_save = n_elements(ind_aver)
         if !version.memory_bits ge 64 then nt_save = long64(nt_save)
         time_save = float((time_s([0, ind_aver(0:nt_save-2)]) + time_s(ind_aver-1)) / 2.)
         time_s = 0
         save, file=file_time_coarse, time_save, /xdr

         drifts_all = fltarr(nt_save, nb)
         for b = 0L, nb - 1 do drifts_all(*, b) = interpol(drifts_all_old(*, b), time_save_old, time_save)
         time_save_old = 0
         drifts_all_old = 0
         time_save = 0

         if (inst eq 'spire' or iter_indiv eq n_lt_indiv) then flag_grid_aver = 1
         save, file=file_scalars, file_times, file_time_coarse, file_signal, file_pixcoords, $
            file_projcoarse, file_projfine, file_imprint, file_maps, file_vscan, file_ntmin_pass, $
            inst, nt, nt_save, nt_aver, timestep, maxtimediff, nb, bref, nscans, nsubscans, $
            logf_inf, logf_sup, logf_high_inf, logf_high_sup, $
            vscan_as_s, dimx, dimy, dimx_iter, dimy_iter, fact, $
            pix_as, diam_as_imprint, dim_imprint, nstep_imprint, xcmin_imprint, $
            npass_crit, fact_npass_cen, visu, debug, /xdr
      endif


      ;;; update white noise and/or mask glitches:
      mask_gl = 0
      if (not keyword_set(wp_transient)) then wp_transient = [-9L]
      if (inst eq 'spire' and iter eq 1) then begin
         if (not keyword_set(noglitch)) then mask_gl = 1 else mask_gl = 0
         compute_wnoise, file_scalars=file_scalars, maxnoise, highfnoise, ind_scans=ind_scans, $
            ind_subscans=ind_subscans, mask_gl=mask_gl, f_fwhm=fact_fwhm(detarr), $
            wp_transient=wp_transient, time_1=time_1, t_proc=t_proc
         print, crs + 'updated "white noise" statistics:'
         for i = 0, min([nscans - 1, 19]) do print, 'scan ' + chain(i) + ':   min = ' $
            + chain(min(maxnoise(i, *), /nan), 5) + '   max = ' + chain(max(maxnoise(i, *), /nan), 5)
         print, crs
      endif
      if (inst eq 'pacs' and iter eq 1 and (not keyword_set(noglitch))) then begin
         ;;; no longer exclude first subscan after PCAL from glitch detection
         mask_gl = 1
         compute_wnoise, file_scalars=file_scalars, maxnoise_dump, highfnoise_dump, ind_scans=ind_scans, $
            ind_subscans=ind_subscans, /initial, /mask_gl, f_fwhm=fact_fwhm(detarr), cut_start=0, $
            wp_transient=wp_transient, time_1=time_1, t_proc=t_proc
         maxnoise_dump = 0
         highfnoise_dump = 0
      endif


      if (keyword_set(debug) and iter gt 0) then begin
         time_2 = systime(/seconds)
         t_proc += (time_2 - time_1)
         print, crs + 'signal spectral density after iteration ' + chain(iter-1) + crs
         compute_spd_signal, freq_comp, spd_comp_aver, file_scalars=file_scalars, $
            maxnoise=maxnoise, ind_scans=ind_scans, ind_subscans=ind_subscans
         time_1 = systime(/seconds)
      endif


      if aver_completed eq 0 then mode_drifts = 'average drift' else mode_drifts = 'individual drifts'
      min_nsamp = min_nsamples(detarr)



      ;;; new projection:
      if keyword_set(clean) then proj_map_clean, file_scalars=file_scalars, $
         maxnoise=maxnoise, ind_scans=ind_scans $
      else proj_map, file_scalars=file_scalars, maxnoise=maxnoise, ind_scans=ind_scans
      if keyword_set(visu) then begin
         time_2 = systime(/seconds)
         t_proc += (time_2 - time_1)
         restore, file_maps
         if keyword_set(clean) then map_signal = map_signal_uncl
         disp_ima, map_signal, win=1, title='iteration '+chain(iter)+' : input map', weightmap=map_weight
         res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
         map_signal = 0
         map_weight = 0
         time_1 = systime(/seconds)
      endif




      ;;; drift subtraction:

      print, 'iteration ' + chain(iter) + ' : ' + mode_drifts
      print, '              time resolution = ' + chain(tc_drift_s, 2) + ' s'

      if ((iter eq 0 or (iter eq 1 and inst eq 'spire')) and (not keyword_set(noglitch)) $
         and fact_fwhm(detarr) le 4.) then detect_gl = 1 else detect_gl = 0


      if mode_drifts eq 'average drift' then begin

         title = 'iteration ' + chain(iter) + ' : drifts (all-bolometers average)'
         if inst eq 'pacs' then dr_bolo = p_arrhw(detarr) * pfov(detarr) / pix_as
         amp_drifts = subtract_aver_drift(drifts_all, flag_valid_scans=flag_valid_scans, $
            maxnoise=maxnoise, highfnoise=highfnoise, ind_sc_coarse=ind_sc_coarse, ind_scans=ind_scans, $
            ind_subscans=ind_subscans, ind_aver=ind_aver, file_scalars=file_scalars, title=title, $
            min_nsamp=min_nsamp, detect_gl=detect_gl, t_cross_fwhm=t_cross_fwhm, f_fwhm=fact_fwhm(detarr), $
            parallel=parallel, nocross=nocross, iter=iter, cut_start=cut_start, dr_bolo=dr_bolo, $
            clean=clean, time_1=time_1, t_proc=t_proc)

      endif else begin

         title = 'iteration ' + chain(iter) + ' : drifts (bolometer ' + chain(bref) + ')'
         amp_drifts = subtract_indiv_drifts(drifts_all, stopcorr, maxnoise=maxnoise, highfnoise=highfnoise, $
            ind_sc_coarse=ind_sc_coarse, ind_scans=ind_scans, ind_subscans=ind_subscans, ind_aver=ind_aver, $
            file_scalars=file_scalars, title=title, min_nsamp=min_nsamp, detect_gl=detect_gl, $
            t_cross_fwhm=t_cross_fwhm, f_fwhm=fact_fwhm(detarr), iter_tot=iter, iter_indiv=iter_indiv, $
            clean=clean, time_1=time_1, t_proc=t_proc, n_lt_indiv=n_lt_indiv)

      endelse




      ;;; prepare next iteration:
      iter += 1
      if (mode_drifts eq 'individual drifts' and max(amp_drifts) gt 0.) then iter_indiv += 1

      if inst eq 'spire' then begin
         if (max(amp_drifts / maxnoise, /nan) lt 1. and median(amp_drifts / maxnoise, /even) le 0.5) then begin
            if (mode_drifts eq 'individual drifts' and (iter gt 1 or max(amp_drifts) eq 0.)) $
               then indiv_completed = 1
            if mode_drifts eq 'average drift' then aver_completed = 1
         endif
      endif else begin
         wdef = where(finite(maxnoise) gt 0, cdef)
         wconv = where(amp_drifts(wdef) / maxnoise(wdef) lt 1., cconv)
         if (float(cconv) / float(cdef) ge 0.97 and (median(amp_drifts / maxnoise, /even) le 0.5 $
          or max(amp_drifts / maxnoise, /nan) lt 1.)) then begin
            if (mode_drifts eq 'individual drifts' and (iter gt 1 or max(amp_drifts) eq 0.) $
               and iter_indiv gt n_lt_indiv) then indiv_completed = 1
            if mode_drifts eq 'average drift' then aver_completed = 1
         endif
      endelse
      if (aver_completed lt 1 and iter ge iter_aver_max) then aver_completed = 1
      if (keyword_set(stopcorr) and mode_drifts eq 'individual drifts') then indiv_completed = 1

   endrep until (iter_indiv ge iter_max or (indiv_completed eq 1 and aver_completed eq 1))



   
   time_2 = systime(/seconds)
   t_proc += (time_2 - time_1)
   print, crs + 'data processed in ' + chain(t_proc) + ' s'
   print, 'observation duration = ' + chain(nt * timestep) + ' s'



   restore, file_time_coarse
   for b = 0L, nb - 1L do drifts_all(*, b) += interpol(drifts_all_aver(*, b), time_save_aver, time_save)
   drifts_all_aver = 0
   time_save_aver = 0
   time_save = 0



   if keyword_set(visu) then begin
      print, crs + 'signal spectral density after full correction (on scan length)'
      compute_spd_signal, freq_comp, spd_comp_aver, file_scalars=file_scalars, $
         maxnoise=maxnoise, ind_scans=ind_scans, ind_subscans=ind_subscans
      freq_comp = 0
      spd_comp = 0
      spd_comp_aver = 0

      ;;; compute spectral density of removed low-frequency noise:
      print, 'subtracted drifts spectral density (on coarse time grid)'
      compute_nspd, file_scalars=file_scalars, drifts_sum=drifts_all, maxnoise=maxnoise, $
         ind_scans=ind_scans, ind_subscans=ind_subscans, nblocks=nblocks, $
         title='total drifts spectral density'
   endif



   ;;; respective amplitudes of thermal drift and individual drifts:

   drift_thermal = fltarr(nt_save)
   for i = 0, nscans - 1 do begin
      wb = where(finite(maxnoise(i, *)) gt 0, cb)
      drifts_scan = drifts_all(ind_sc_coarse(i):ind_sc_coarse(i+1)-1, *)
      drift_thermal(ind_sc_coarse(i)) = total(drifts_scan(*, wb) $
         / rebin(maxnoise(i, wb)^2., ind_sc_coarse(i+1)-ind_sc_coarse(i), cb), 2) $
         / total(1./ maxnoise(i, wb)^2.)
   endfor

   if keyword_set(debug) then begin
      wset, 0
      restore, file_time_coarse
      otherplot, time_save, drift_thermal, xtitle='time (s)', ytitle='thermal drift'
      res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   endif


   med = mean(drift_thermal)
   amp_thermal = 3. * sqrt(mean((drift_thermal - med)^2.))

   restore, file_times
   restore, file_time_coarse

   drifts_uncorr = drifts_all - rebin(drift_thermal, nt_save, nb)
   for i = 0, nscans - 1 do begin
      for b = 0L, nb - 1 do if finite(maxnoise(i, b)) gt 0 then $
         drifts_uncorr(ind_sc_coarse(i):ind_sc_coarse(i+1)-1, b) -= $
         mean(drifts_uncorr(ind_sc_coarse(i):ind_sc_coarse(i+1)-1, b))
   endfor
   wb = where(finite(total(maxnoise, 1)) gt 0)
   amp_uncorr = 3. * sqrt(mean(drifts_uncorr(*, wb)^2.))

   time_s = 0
   flag_nza = 0
   time_save = 0

   print, crs + 'drift amplitudes in units of median white noise and in Jy/bolometer:'
   print, '     thermal: ' + chain(amp_thermal / median(maxnoise), 2) + ' WN or ' $
      + chain(amp_thermal, 5) + ' Jy/b'
   print, '     uncorrelated: ' + chain(amp_uncorr / median(maxnoise), 2) + ' WN or ' $
      + chain(amp_uncorr, 5) + ' Jy/b'


   if keyword_set(visu) then begin
      if (not keyword_set(nothermal)) then compute_nspd, file_scalars=file_scalars, $
         drifts_sum=rebin(drift_thermal, nt_save, nb), maxnoise=maxnoise, ind_scans=ind_scans, $
         ind_subscans=ind_subscans, nblocks=nblocks, title='average drift spectral density'
      compute_nspd, file_scalars=file_scalars, drifts_sum=drifts_uncorr, maxnoise=maxnoise, $
         ind_scans=ind_scans, ind_subscans=ind_subscans, nblocks=nblocks, $
         title='uncorrelated drifts spectral density'
   endif




   ;;;;;;;;;;;;;;;;;;;;;
   ;;; final projection
   ;;;;;;;;;;;;;;;;;;;;;

   if (pix_as_final ne pix_as or orient eq 'astro' or keyword_set(hdr_ref)) then begin

      restore, file_xyas_final

      if (not keyword_set(common_grid)) then begin
         if keyword_set(hdr_ref) then begin
            dimx_final = sxpar(hdr_ref, 'NAXIS1')
            dimy_final = sxpar(hdr_ref, 'NAXIS2')
            crpix1_final = sxpar(hdr_ref, 'CRPIX1') - 0.5
            crpix2_final = sxpar(hdr_ref, 'CRPIX2') - 0.5
         endif else begin
            if orient eq 'astro' then begin
               dimx_final = ceil((float(max(xas) - min(xas)) + 2. * off_map_as) / pix_as_final)
               dimy_final = ceil((float(max(yas) - min(yas)) + 2. * off_map_as) / pix_as_final)
               crpix1_final = (float(off_cen_Xra * 3600.D0) + off_map_as) / pix_as_final
               crpix2_final = (float(off_cen_Ydec * 3600.D0) + off_map_as) / pix_as_final
            endif else begin
               dimx_final = ceil(dimxas / pix_as_final)
               dimy_final = ceil(dimyas / pix_as_final)
               crpix1_final = (float(off_cen_Xra * 3600.D0) + off_map_as) / pix_as_final
               crpix2_final = (float(off_cen_Ydec * 3600.D0) + off_map_as) / pix_as_final
            endelse
         endelse
      endif

      if keyword_set(hdr_ref) then begin
         x_bolos = float(xas / pix_as_final)
         xas = 0
         y_bolos = float(yas / pix_as_final)
         yas = 0
      endif else begin
         x_bolos = float((off_map_as + xas) / pix_as_final)
         xas = 0
         y_bolos = float((off_map_as + yas) / pix_as_final)
         yas = 0
      endelse
      save, file=file_pixcoords, x_bolos, y_bolos, /xdr
      x_bolos = 0
      y_bolos = 0

      if inst eq 'spire' then diameter = min([fwhm_as(detarr) / 2. / sqrt(!pi), $
         2. * pix_as_final / sqrt(!pi)])
      if inst eq 'pacs' then diameter = 2. * sqrt(0.8) * pfov(detarr) / sqrt(!pi)
      if block eq block_start then compute_imprint, file_scalars=file_scalars, pixsize=pix_as_final, $
         diameter=diameter, nstep_imprint_final=nstep_imprint_final, file_imprint_final=file_imprint_final

      if nzadata ne 'no' then prepare_proj_fine, file_scalars=file_scalars, $
         xdim=dimx_final, ydim=dimy_final, ind_scans=ind_scans

   endif else begin
      restore, file_scalars
      dimx_final = dimx
      dimy_final = dimy
      crpix1_final = crpix1
      crpix2_final = crpix2
   endelse


   ;;; include or not non-zero acceleration data:
   if nzadata eq 'no' then prepare_proj_fine, file_scalars=file_scalars, $
      xdim=dimx_final, ydim=dimy_final, ind_scans=ind_scans, nzadata=nzadata


   print, crs + 'final projection of drift-corrected time series'

   restore, file_times
   flag_nza = 0
   restore, file_time_coarse
   drifts = fltarr(nt, nb)
   for b = 0L, nb - 1 do drifts(*, b) = interpol(drifts_all(*, b), time_save, float(time_s))
   time_s = 0
   time_save = 0

   if keyword_set(clean) then proj_map_clean, file_scalars=file_scalars, $
      nstep_imprint_final=nstep_imprint_final, file_imprint_final=file_imprint_final, $
      dimx_proj=dimx_final, dimy_proj=dimy_final, maxnoise=maxnoise, ind_scans=ind_scans, $
      /comp_errors, lowfnoise=drifts $
     else proj_map, file_scalars=file_scalars, nstep_imprint_final=nstep_imprint_final, $
      file_imprint_final=file_imprint_final, dimx_proj=dimx_final, dimy_proj=dimy_final, $
      maxnoise=maxnoise, ind_scans=ind_scans, /comp_errors, lowfnoise=drifts
   drifts = 0


   restore, file_maps
   if keyword_set(clean) then begin
      map_signal_cl = map_signal
      map_signal = map_signal_uncl
   endif


   ;;; set approximately the zero level of the map:
   if (not keyword_set(galactic)) then stat_map, med, sig, map=map_signal, weightmap=map_weight, noise_level=median(maxnoise) $
    else begin
       w = where(map_weight ge 3.)
       histo = histogram(map_signal(w), locations=sb, binsize=median(maxnoise), /nan)
       w = where(total(histo, /cumulative) ge 30)
       med = sb(w(0))
    endelse
   w = where(map_weight gt 0.)
   map_signal(w) -= med
   if keyword_set(clean) then map_signal_cl(w) -= med


   if keyword_set(visu) then begin
      disp_ima, map_signal, win=1, title='final map', weightmap=map_weight
      disp_ima, map_weight, win=2, title='weight map'
      disp_ima, map_error, win=3, title='error map', weightmap=map_weight, noise_level=median(maxnoise)/10.
      disp_ima, map_lowfnoise, win=4, title='total drifts (excluding flux cal. offsets)', $
        weightmap=map_weight, noise_level=10.*median(maxnoise)
      res = dialog_message('Write output ?', /center, /information, title='Pause')
   endif


   if n_elements(wp_transient) gt 1 then begin
      wp_transient = wp_transient(1:*)
      restore, file_signal
      signal_save = signal
      signal = !values.f_nan + fltarr(nt, nb)
      signal(wp_transient) = signal_save(wp_transient) - med
      save, file=file_signal, signal, /xdr
      prepare_proj_fine, file_scalars=file_scalars, xdim=dimx_final, ydim=dimy_final, ind_scans=ind_scans
      proj_map, map_transient, file_scalars=file_scalars, nstep_imprint_final=nstep_imprint_final, $
         file_imprint_final=file_imprint_final, dimx_proj=dimx_final, dimy_proj=dimy_final, $
         maxnoise=maxnoise, ind_scans=ind_scans
      w = where(map_weight le 0.)
      map_transient(w) = !values.f_nan
      if keyword_set(visu) then begin
         disp_ima, map_transient, win=5, title='transient signal map'
         res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
      endif
   endif




;;;;;;;;;;;;;;;;;;;
;;; output written
;;;;;;;;;;;;;;;;;;;


   if keyword_set(clean) then nplanes = 5 else nplanes = 4
   cube = fltarr(dimx_final, dimy_final, nplanes)
   cube(*, * ,0) = map_signal
   cube(*, * ,1) = map_error
   cube(*, *, 2) = map_lowfnoise
   cube(*, *, 3) = map_weight
   if keyword_set(clean) then cube(*, *, 4) = map_signal_cl
   w = where(map_weight le 0.)
   for i = 0, nplanes - 1 do if i ne 3 then cube(w+dimx_final*dimy_final*i) = !values.f_nan

   ;;; brightness conversion:
   if inst eq 'pacs' then begin
      cube(*, *, 0:2) *= (pix_as_final / pfov(detarr))^2.
      if keyword_set(clean) then cube(*, *, 4) *= (pix_as_final / pfov(detarr))^2.
      bunit = 'Jy/pixel'
   endif else bunit = 'Jy/beam'

   key_planes = 'PLANE' + chain(indgen(nplanes))
   str_planes = ['signal map', 'error map', 'total drifts (excluding flux calibration offsets)', 'weight map']
   if keyword_set(clean) then str_planes = [str_planes, 'signal map weighted to exclude noisy scans']

   mkhdr, hdr, cube
   sxaddpar, hdr, 'TELESCOP', 'Herschel'
   sxaddpar, hdr, 'INSTRUME', strupcase(inst)
   sxaddpar, hdr, 'FILTER', arrays_names(detarr)
   sxaddpar, hdr, 'OBSMODE', 'SCAN'
   sxaddpar, hdr, 'NSCANS', nscans, 'number of combined scans'
   sxaddpar, hdr, 'NBLOCKS', nblocks, 'number of spatial blocks for processing'
   sxaddpar, hdr, 'BLOCK', 'TOTAL', 'block number of this map / total map'
   sxaddpar, hdr, 'BUNIT', bunit
   sxaddpar, hdr, 'PFOV', pix_as_final, 'pixel size (arcseconds)'
   sxaddpar, hdr, 'CTYPE1', 'RA---TAN'
   sxaddpar, hdr, 'CTYPE2', 'DEC--TAN'
   sxaddpar, hdr, 'EQUINOX', 2000.
   sxaddpar, hdr, 'CRPIX1', crpix1_final + 0.5, 'reference pixel in X coordinate'
   sxaddpar, hdr, 'CRPIX2', crpix2_final + 0.5, 'reference pixel in Y coordinate'
   sxaddpar, hdr, 'CRVAL1', crval1, 'RA of reference position (degrees)'
   sxaddpar, hdr, 'CRVAL2', crval2, 'DEC of reference position (degrees)'
   cdelt = pix_as_final / 3600.
   sxaddpar, hdr, 'CDELT1', -cdelt
   sxaddpar, hdr, 'CDELT2', cdelt
   if orient eq 'astro' then crota2 = 0. else crota2 = -scanrot_deg
   sxaddpar, hdr, 'CROTA2', crota2, 'angle from Y axis to North axis (degrees)'
   for i = 0, nplanes - 1 do sxaddpar, hdr, key_planes(i), str_planes(i)
   get_version, parentdir=root, /silent, vers=vers
   sxaddpar, hdr, 'SOFTWARE', 'processed from level 1 with Scanamorphos '+vers+' (Roussel 2011)'
   if keyword_set(nothermal) then options1 = 'nothermal' else options1 = ''
   if keyword_set(noglitch) then if options1 eq '' then options1 = 'noglitch' $
      else options1 = options1 + ', noglitch'
   if keyword_set(nogains) then if options1 eq '' then options1 = 'nogains' $
      else options1 = options1 + ', nogains'
   if keyword_set(parallel) then options2 = 'parallel' else options2 = ''
   if keyword_set(galactic) then if options2 eq '' then options2 = 'galactic' $
      else options2 = options2 + ', galactic'
   if keyword_set(minimap) then if options2 eq '' then options2 = 'minimap' $
      else options2 = options2 + ', minimap'
   if keyword_set(nocross) then if options2 eq '' then options2 = 'nocross' $
      else options2 = options2 + ', nocross'
   if keyword_set(jumps_pacs) then if options2 eq '' then options2 = 'jumps_pacs' $
      else options2 = options2 + ', jumps_pacs'
   if options2 ne '' then key1 = 'OPTIONS1' else key1 = 'OPTIONS'
   if options1 ne '' then key2 = 'OPTIONS2' else key2 = 'OPTIONS'
   if options1 ne '' then sxaddpar, hdr, key1, options1
   if options2 ne '' then sxaddpar, hdr, key2, options2
   if options1 eq '' and options2 eq '' then sxaddpar, hdr, 'OPTIONS', 'none'


   hdr_block = hdr
   if nblocks gt 1 then sxaddpar, hdr_block, 'BLOCK', block, 'block number of this map / total map'

   if (nblocks eq 1 and keyword_set(one_plane_fits)) then begin
      hdr_ima = hdr_block
      sxaddpar, hdr_ima, 'NAXIS', 2
      sxdelpar, hdr_ima, 'NAXIS3'
      sxdelpar, hdr_ima, key_planes
      suffix = ['signal', 'error', 'drifts', 'weight', 'clmap']
      for i = 0, nplanes - 1 do begin
         hdr_plane = hdr_ima
         sxaddpar, hdr_plane, key_planes(i), str_planes(i)
         writefits, dir_out + file_out + '_' + inst + arrays_names(detarr) + '_' + suffix(i) + '.fits', $
            cube(*, *, i), hdr_plane
      endfor
   endif else writefits, dir_out + file_out + '_' + inst + arrays_names(detarr) + '_' + chain(block) + '.fits', $
      cube, hdr_block

   if keyword_set(map_transient) then begin
      if inst eq 'pacs' then map_transient *= (pix_as_final / pfov(detarr))^2.
      hdr_ima = hdr_block
      sxaddpar, hdr_ima, 'NAXIS', 2
      sxdelpar, hdr_ima, 'NAXIS3'
      sxdelpar, hdr_ima, key_planes
      sxaddpar, hdr_ima, key_planes(0), 'transient signal map'
      print, crs + 'saving a transient signal map (asteroids for example),'
      print, '   mapping only the scan leg(s) containing the signal,'
      print, '   with the same background level as in the plane-0 map'
      writefits, dir_out + file_out + '_' + inst + arrays_names(detarr) + '_transient_' $
         + chain(block) + '.fits', map_transient, hdr_ima
   endif


   if keyword_set(visu) then begin
      print, crs + 'time for last checks'
      print, 'Now is the time to save the corrected time series, and/or other variables'
      print, '(move the files "save_signal_[detector].xdr" to a subdirectory).'
      res = dialog_message('   Finish ?   ', /center, /information, title='Pause')
   endif

   while !d.window ge 0 do wdelete, !d.window

   cube = 0
   map_signal = 0
   map_error = 0
   map_lowfnoise = 0
   map_weight = 0

endif

endfor


;;; clean up intermediate files written to disk:
file_delete, file_scalars, file_times, file_time_coarse, file_projcoarse, file_projfine, $
   file_pixcoords, file_vscan, file_ntmin_pass, file_maps, $
   file_XYscan, file_param_blocks, /noexpand_path, /quiet
for detarr = 0, narr - 1 do if ind_arrays_eff(detarr) eq 1 then begin
   suffix = arrays_default(detarr) + '.xdr'
   file_delete, root_signal + suffix, root_xyas + suffix, root_xyas + 'final_' + suffix, $
      /noexpand_path, /quiet
endif




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; blocks stitched together 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

if nblocks gt 1 then begin
   print, crs + 'Combining the ' + chain(nblocks) + ' spatial blocks into a single map.' + crs

   w = where(ind_arrays_eff eq 1)
   ind = w(0)
   root_out = dir_out + file_out + '_' + inst + arrays_names(ind)
   stitch_blocks, dimx_final=dimx_final, dimy_final=dimy_final, nplanes=nplanes, nblocks=nblocks, $
      plane_weight=3, root_out=root_out, hdr=hdr
endif


while !d.window ge 0 do wdelete, !d.window

print, crs + 'Bye' + crs


end_label:
end


;;;;;;;;;
;;;;;;;;;
;;;;;;;;;
