pro stitch_blocks, dimx_final=dimx_final, dimy_final=dimy_final, nplanes=nplanes, nblocks=nblocks, $
    plane_weight=plane_weight, root_out=root_out, hdr=hdr


;;; for observations sliced into several blocks (sub-fields overlapping at the edges):
;;; combine all blocks into a simgle map

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



dimxy = dimx_final * dimy_final

cube_combined = fltarr(dimx_final, dimy_final, nplanes)


for block = 0L, nblocks - 1 do begin
   cube = readfits(root_out + '_' + chain(block) + '.fits', /silent)

   ;;; determine region of new block to be combined with others (sufficient weight):
   map_weight = cube(*, *, plane_weight)
   map_weight_combined = cube_combined(*, *, plane_weight)

   ;;; remove possible bias in signal plane between different blocks:
   tab_off = fltarr(nplanes)
   if block gt 0 then begin
      w_over = where(map_weight * map_weight_combined gt 0.)
      m = min([median(map_weight(w_over)), median(map_weight_combined(w_over))])
      w_over = where(map_weight gt m / 2. and map_weight_combined gt m / 2.)
      m = min([median(map_weight(w_over)), median(map_weight_combined(w_over))])
      w_over = where(map_weight gt 0.7 * m and map_weight_combined gt 0.7 * m $
         and cube(*, *, 1) lt 2. * cube_combined(*, *, 1) and cube_combined(*, *, 1) lt 2. * cube(*, *, 1))
      w_comb = where((map_weight ge 0.7 * m and map_weight_combined lt 0.7 * m) $
         or map_weight_combined le (Machar()).eps $
         or (map_weight ge 1.5 * map_weight_combined and cube(*, *, 1) lt 2. * cube_combined(*, *, 1)) $
         or (cube(*, *, 1) le 0.7 * cube_combined(*, *, 1) and map_weight gt 0.5 * map_weight_combined))
      off = median(cube(w_over) - cube_combined(w_over))
      print, 'offset of block ' + chain(block) + ' wrt preceding block(s) = ' + chain(off, 5)
      tab_off(0) = off
      if nplanes gt 4 then tab_off(4) = off
   endif else w_comb = where(map_weight gt 0.)

   ;;; add signal plane(s) with offset, weight and error planes with no bias:
   for k = 0, nplanes - 1 do cube_combined(w_comb+k*dimxy) = cube(w_comb+k*dimxy) - tab_off(k)
endfor

w = where(cube_combined(*, *, plane_weight) le 0.)
for i = 0, nplanes - 1 do if i ne plane_weight then cube_combined(w+dimx_final*dimy_final*i) = !values.f_nan

writefits, root_out + '_combined.fits', cube_combined, hdr


end
